#include <sys/types.h>
#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "animation.h"


				/* Chance table... relative to each other */
				/* Can be any size/total */
int act_chance[NUM_ACTS] = {
    500,			/* Move/enter obj/leave object */
    100,			/* Open/close obj */
    300,			/* Saying things */
    100,			/* Taking things */
    10,				/* Droping things */
    40,				/* Locking things */
    10,				/* Unlocking things */
     5,				/* Shouting things */
    40,				/* Giving things */
    40,				/* Stealing things */
    50				/* Attacking things */
};

/****************************************************************************\
  Move an object around, or let it open and close things that are its
  current location. (Open/Closing not done yet!).
\****************************************************************************/

void animate_move ( ind )
    int ind;
{
    int exit_list[MAXDIRECTIONS],
        num_exits = 0;
    int dirn,			/* Direction to go in */
        where,			/* Where the item is (real_loc) */
        playerno,		/* Player who holds item */
        newloc;			/* Items new position */

    where = real_loc ( ind );	/* Where is the object */

				/* If not held and under pit frogging */
				/* aint moving! */
    if ( where == UPITLOC )
	return;
				/* Can it get out to its real_loc ? */
    if ( ! accessout ( ind, where ) ) {
				/* Perhaps it should try to open what */
				/* its in? If actions allow. */
	return;
    }
				/* Chance of closing object it is in */
				/* If actions allow... */

				/* Is it held by a player? */
    if ( ( playerno = whoholdsitem ( ind ) ) != -1 ) {
				/* Is it in something? */
	if ( obj[ind].location < 0 )
	    say ( form ( "The %s decides to leave the %s!\n",
			obj[ind].desc,
			obj[-obj[ind].location].desc ) );
	else
	    say ( form ( "The %s wriggles from your grasp!\n",
			obj[ind].desc ) );
				/* Tell other people here */
	speak ( form ( "The %s has escaped from %s!\n",
		      obj[ind].desc, player[playerno].name ),
	       playerno, player[playerno].where );

				/* Move it out */
	obj[ind].location = player[playerno].where;
	    
    } else {
	int i;

	if ( obj[ind].location < 0 ) {
	    speak ( form ( "The %s decides to leave the %s!\n",
			obj[ind].desc,
			obj[-obj[ind].location].desc ), NO_ONE, where );
	    obj[ind].location = where;
	} else {
	
				/* Its not held - move out of room */
				/* Pick a direction from location */

				/* Or perhaps - we should pick an object */
				/* to enter... perhaps opening it first */

	    for ( i = 0; i < no_directions; i++ )
		if ( loc[where].exits[i] != 0 )
		    exit_list[num_exits++] = i;

				/* Is there a way out? */
	    if ( num_exits == 0 )
		return;
	    
				/* Pick an exit */
	    dirn = exit_list[(rand () % num_exits)];

				/* Is there an exit */
	    if ( ( newloc = loc[where].exits[dirn] ) != 0 ) {
				/* Is the exit to the upit? */
		if ( loc_pitp(where) ) {
		    newloc = UPITLOC;
		    shout ( form ( "\
!!! The %s has thrown itself into the pit!!!\n", obj[ind].desc ),
			   NO_ONE, MESG_GLOBAL );
		}
		else
		    speak ( form ( "The %s has gone %s.\n",
				  obj[ind].desc, 
				  dir[dirn].desc ),
			   NO_ONE, where );
				/* Move it out man! */
		if ( newloc == -1 )
		    newloc = 0;
		newloc = abs ( newloc );
		obj[ind].location = newloc;
				/* Tell people in the room its arrived! */
		speak ( form ( "The %s has arrived.\n",
			      obj[ind].desc ), NO_ONE, newloc );
	    }
	}
    }
}

/****************************************************************************\
  Not implemented yet... NULL action
\****************************************************************************/

void animate_open ( ind )
    int ind;
{
}

/****************************************************************************\
  Objects says something random.
\****************************************************************************/

void animate_say ( ind )
    int ind;
{
    int sayno = (random () % no_sayings);
    int where = where_item ( ind );

    if ( ! accessout ( ind, where ) )
	speak ( form ( "You hear a muffled noise coming from the %s.\n",
		      obj[base_item (ind)].desc ), NO_ONE, where );
    else
	speak ( form ( "The %s says; \"%s\".\n", 
		      obj[ind].desc,
		      sayings[sayno] ), NO_ONE, where );
}

/*--------------------------------------------------------------------------*/

void animate_take ( ind )
    int ind;
{
}

void animate_drop ( ind )
    int ind;
{
}

void animate_lock ( ind )
    int ind;
{
}

void animate_unlock ( ind )
    int ind;
{
}

/****************************************************************************\
  Shout - make objects SHOUT!
\****************************************************************************/

void animate_shout ( ind )
    int ind;
{
    int shout = (random () % no_shouts);
    int where = where_item ( ind );

    if ( ! accessout ( ind, where ) )
	speak ( form ( "You hear a noise coming from the %s.\n",
		      obj[base_item (ind)].desc ), NO_ONE, where );
    else {
	shmortal ( form ( "Someone somewhere shouts \"%s\".\n", 
			 shouts[shout] ),
		  NO_ONE, MESG_SHOUT );
	shwiz ( form ( "The %s (in %s (%s)) shouts \"%s\".\n",
		      obj[ind].desc,
		      loc[where].name,
		      loc[where].code,
		      shouts[shout] ),
	       NO_ONE, MESG_SHOUT );
    }
}

void animate_give ( ind )
    int ind;
{
}

void animate_steal ( ind )
    int ind;
{
}

void animate_attack ( ind )
    int ind;
{
}

/****************************************************************************\
  Make object (ind) DO something, first of all pick what action to perform..
  then call the respective handling routine.

  The object can actually pick an action which it cannot perform... in this
  case we simply return. This means that objects with only one flag set
  do not do that action more, than say an object with all its flags set.
\****************************************************************************/

void animate_object ( ind )
    int ind;
{
    static int total = 0;	/* Total chance of all avail actions */
    int i,
        roll,			/* Roll on chance */
        sum;			/* Sum */

    if ( ! total ) {		/* Initialise total if necassary */
	for ( i = 0; i < NUM_ACTS; i++ )
	    total += act_chance[i];
    }

    roll = (random() % total);
    sum = 0;

    for ( i = 0; i < NUM_ACTS; i++ ) {
	sum += act_chance[i];
	if ( roll < sum )
	    break;
    }

				/* Can object perform this action? */
    if ( (obj[ind].actions & (1<<i)) == 0 )
	return;

    switch ( i ) {
    case 0:			/* Movement */
	animate_move ( ind );
	break;

    case 1:			/* Open/Close */
	animate_open ( ind );
	break;

    case 2:			/* Say something */
	animate_say ( ind );	
	break;

    case 3:			/* Take something */
	animate_take ( ind );
	break;

    case 4:			/* Drop something */
	animate_drop ( ind );
	break;

    case 5:			/* Lock something */
	animate_lock ( ind );
	break;

    case 6:			/* Unlock something */
	animate_unlock ( ind );
	break;

    case 7:			/* Shout something */
	animate_shout ( ind );
	break;

    case 8:			/* Give something */
	animate_give ( ind );
	break;

    case 9:			/* Steal something */
	animate_steal ( ind );
	break;

    case 10:			/* Attack something */
	animate_attack ( ind );
	break;

    }
}


/****************************************************************************\
  The animate command (spell) - makes an object do some action;

  animate <obj> [area]   - Animate object in area - default next unused area.
  
\****************************************************************************/
void animate_cmd ( playerno, command )
{
    say ( "That spell seems not to be implemented in the MundiVerse!\n" );
}

/****************************************************************************\
  Similar to animate...
\****************************************************************************/

void deanimate_cmd ( playerno, command )
{
    say ( "That spell seems not to be implemented in the MundiVerse!\n" );
}

/****************************************************************************\
  Increase speed of object
\****************************************************************************/

void haste_cmd ( playerno, command )
{
    say ( "That spell seems not to be implemented in the MundiVerse!\n" );
}

/****************************************************************************\
  Decrease speed of object
\****************************************************************************/
void slow_cmd ( playerno, command )
{
    say ( "That spell seems not to be implemented in the MundiVerse!\n" );
}
