/***************************************************************************\
|*									   *|
|*	Daemon.c:  Server program for MUNDI.				   *|
|*	                This version started on or about 18th May 1988.	   *|
|*			Oh, it's so frogging fabulous!			   *|
|*									   *|
|*	Exit status meanings:						   *|
|*									   *|
|*	        0: Daemon successfully booted (running in background).	   *|
|*	                (also: daemon dies when all clients die)	   *|
|*	        1: Socket() creation failed.				   *|
|*	        2: Bind() failed.					   *|
|*		3: Fcntl() failed to make accept non-blocking (unused)	   *|
|*	        4: Listen() failed.					   *|
|*	        5: Accept() failed.					   *|
|*	        6: Select() failed.					   *|
|*	        7: Read() from socket failed.				   *|
|*	        8: Write() to socket failed.				   *|
|*	        9: Fork() failed.					   *|
|*	        10: Execl() (not execlp() :-) failed.			   *|
|*	        11: Signal() handling failed.				   *|
|*	        12: Malloc() failed.					   *|
|*	        13: Wait() failed.					   *|
|*	        14: Gethostname() failed.				   *|
|*	        15: Wrong host.						   *|
|*		16: Daemon crashed on purpose.				   *|
|*		17: Couldn't open an initialisation file.		   *|
|*		18: Number of data exceeded #defined limit.		   *|
|*		19: Setsockopt() failed.				   *|
|*		20: Couldn't open terminal file.			   *|
|*		21: Ioctl() failed to give back pty.			   *|
|*		22: Freopen() couldn't redirect stdout.			   *|
|*		23: Couldn't open/create/whatever the password file.	   *|
|*		24: Couldn't stat the password file.			   *|
|*		25: Couldn't read it this time.  Frogging thing.           *|
|*		26: Could and did read it, but there wasn't enough of it.  *|
|*		27: Couldn't allocate memory to store it all.  Cheese.     *|
|*		28: Couldn't allocate more memory for new persona.	   *|
|*		29: Couldn't lseek() on password file, A?		   *|
|*		30: Couldn't write persona to file.			   *|
|*									   *|
\***************************************************************************/
/***************************************************************************\
|*									   *|
|*  The #definition of debug is here in daemon.c (as opposed to .h)	   *|
|*  since it is the only module where DEBUG is used, and thus this	   *|
|*  re-positioning saves a lot of re-compiling when I change it.	   *|
|*									   *|
\***************************************************************************/

#define	DEBUG			/* Run in foreground; dump core at exit */
				/* 2001: was #undef but I don't trust BGing */

/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/wait.h>
#define BSD_COMP		/* 2001: prod ioctl.h into defining FIONREAD */
#include <sys/ioctl.h>
#undef BSD_COMP
/*#include <sys/filio.h>*/	/* 2001: Not available on Linux */
#include <fcntl.h>
#include <varargs.h>

#include "global.h"
#include "daemon.h"
#include "handle.h"
#include "initialise.h"
#include "look.h"
#include "persona.h"
#include "utils.h"
#include "events.h"
#include "schedule.h"

/*--------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *malloc ();
extern char *rindex ();
extern time_t time ();

/*--------------------------------------------------------------------------*/

char *argv[MAXARGS];
int argc, no_directions, no_locations, no_objects, no_personas, no_users = 0;
int no_hows = 0;
fd_set fdsinuse;
char *IFS = " ";		/* This is a kludge, purely to keep the */
				/* compiler happy about the fact that */
				/* utils.c raves on and on about IFS, */
				/* but the daemon knows nothing of it. */
				/* Nasty, but kinda neat in its way. */
				/* Actually, it might be kinda neat to */
				/* be able to use HOOPY_ISSPACE & co. */
/*--------------------------------------------------------------------------*/
/* SIGPIPE handling */

int writing = -1;		/* -1 if not writing, otherwise playerno */
				/* of connection we are writing to... */
int cleanup = -1;		/* -1 if no clean up needed, playerno */
				/* of persona to clean up on otherwise */
/*--------------------------------------------------------------------------*/

void logerror (va_alist)
  va_dcl			/* Resist the tempation to add a ';'! */
{
  va_list pvar;			/* Parameter-traversing variable */
  static char temp1[LINELEN];	/* Variable to hold date(1)-like bit */
  static char temp2[LINELEN];	/* Variable to hold printf(3)-like bit */
  char *format;			/* Variable to hold sprintf format */
  time_t timeofday;		/* Long-int-thing variable for time */

  (void) time (&timeofday);
  (void) strcpy (temp1, ctime (&timeofday));
  temp1[strlen (temp1)-1] = '\0';

  va_start (pvar);
  format = va_arg (pvar, char*);
  (void) vsprintf (temp2, format, pvar);
  va_end (pvar);

  (void) printf ("%s: %s", temp1, temp2);
  fflush ( stdout );
}

/*--------------------------------------------------------------------------*/

/*
   Finds the index to the string array av, of ac items, which starts with
   the string arg.

   Usefull for checking for arguments....
*/

int findarg ( ac, av, arg )
int ac;
char **av;
char *arg;
{
    int i, l = strlen ( arg );

    for ( i = 1; i < ac; i++ )
	if ( strncmp ( av[i], arg, l ) == 0 )
	    return ( i );
    return ( 0 );
}

/*--------------------------------------------------------------------------*/


void quit (value)
  int value;
{
  (void) fflush (stdout);
#ifdef xDEBUG			/* 2001: added "x" prefix: this is silly */
  abort ();
#else /* DEBUG */
  exit (value);
#endif /* DEBUG */
}

/*--------------------------------------------------------------------------*/

void die (value)
  int value;
{
  logerror ("Exit %d, function %d: %s\n", errno, value, sys_errlist[errno]);
  quit (value);
}

/*--------------------------------------------------------------------------*/

void killed (sig,code,scp,addr)
    int sig;
    int code;
    struct sigcontext *scp;
    char *addr;
{
  logerror ("Killed: Signal %d, code %d.\n", sig, code );
  quit (6);
}

/*--------------------------------------------------------------------------*/

void sigpipe ()
{
  signal ( SIGPIPE, SIG_IGN );
  if ( cleanup != -1 )
      logerror ( "SIGPIPE: Cheese! Client#%d should have got killed.\n", 
		cleanup );
  cleanup = writing;
  signal ( SIGPIPE, sigpipe );
}

/***************************************************************************\
|*                                                                         *|
|*  BE CAREFUL: remember this will read from several different             *|
|*  filedes'.  Therefore, always exhaust an input line from within a       *|
|*  single player's handle() routine before reading the next line.         *|
|*                                                                         *|
\***************************************************************************/

/****************************************************************************\
  I changed this cos you tend to get massive chunks of input when someone
  does a source... basically you gotta free the return of this sometime!
\****************************************************************************/

char *saferead (fildes)
    int fildes;
{
    int retval;
    long size;
    static char *message = 0;

    if ( ! message )
	message = (char *) malloc ( 1 ); /* 2001: was malloc(0) */
    
				/* This way we get ALL input */
    if ( ioctl ( fildes, FIONREAD, &size ) == -1 )
      { if (errno == ECONNRESET) { *message = 0; return message; }
	else { die (21); } }	/* 2001: deal with aborting client */
    if ( ( message = (char *) realloc ( message, size + 1 ) ) == 0 )
	die (12);

    if ( ( retval = read ( fildes, message, size ) ) == -1 )
	if ( errno == ECONNRESET )
	    return ( 0 );
	else
	    die (7);
				/* Make sure its NULL terminated */
    message[retval] = 0;
    return ( message );

#if 0
  if ((retval = read (fildes, message, LINELEN)) == -1)
    if (errno == ECONNRESET)	/* If other end of socket has died ... */
      retval = 0;		/* Then we have reached eof, so say so. */
  else				/* Any other error means ... */
    die (7);			/* Genuine tea, so crash out. */

  message[retval] = '\0';	/* Chop off the string at its end (?!) */
  return (retval);		/* This is 0 if EOF, otherwise != 0 */
#endif
}

/****************************************************************************\
  This next function is used to send messages to a player before he is in the
  game, Mirk used to use say(), and safewrite() but this could produce
  some very nasty bugs.
\****************************************************************************/

void send_reply ( playerno, str )
int playerno;
char *str;
{
    writing = playerno;
    if ( player[playerno].fd != -1 )
	if ( write ( player[playerno].fd, str, strlen ( str ) ) == -1 )
	    if ( errno != EPIPE )
		die (8);
    writing = -1;
}

/***************************************************************************\
|*                                                                         *|
|*	NOTE: in mundi.c and associated modules (ie. the frontend)         *|
|*	the function safewrite takes arguments (fildes, message)           *|
|*      In daemon.c and associated modules, it takes arguments             *|
|*      (user number, message).  DO NOT GET CONFUSED!                      *|
|*                                                                         *|
|*      Conjecture: this function throws away any part of the string       *|
|*      	occurring after the last '\n' without printing it.         *|
|*      	Do we want this?  Test it when everything is working.      *|
|*                                                                         *|
\***************************************************************************/

void safewrite (playerno, message)
  int playerno;
  char *message;
{
  int i;
  char temp[DESCLEN];
  char snoopline[LINELEN];
  char *line, *rest, c;
				/* This wasnt here and tee'd up the
				   game sometimes! */
  if ( player[playerno].clientmode < PLAYINGGAME )
      return;

  (void) strcpy (temp, message);
  rest = temp;

  while (rest != NULL) {
    line = rest;
    if ((rest = index (rest, '\n')) != NULL) {
      c = *(++rest);
      *rest = '\0';
      send_reply ( playerno, line );
				/* No snooping on escape - codes */
      if ( line[0] != ESCAPE ) {
	  (void) sprintf (snoopline, "| %s", line);
	  for (i = 0; i < MAXUSERS; i++)
	      if (player[i].snoop == playerno)
		  safewrite (i, snoopline);
      }

      *rest = c;
    }
  }
}

/*--------------------------------------------------------------------------*/
int do_shout ( playerno, type )
    int playerno;
    mesgtype type;
{
    switch ( type ) {
    case MESG_SHOUT:
	if ( IGNORESHOUTP(playerno) )
	    return 0;
	break;
    case MESG_GLOBAL:
	if ( IGNOREGLOBALP(playerno) )
	    return 0;
	break;
    default:
	break;
    }
    return ( 1 );
}
/*--------------------------------------------------------------------------*/

void shout (message, notme, type)
  char *message;
  int notme;
  mesgtype type;		/* Type of message */
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((i != notme) && (ISACTIVE (i)) && (player[i].fd != -1)) {
				/* Shouldn't need last test, since */
				/* fd can only be != -1 if the mode */
				/* is active, but what the cheese? */
	if ( do_shout ( i, type ) )
	    safewrite (i, message);
    }
}

/***************************************************************************\
|*									   *|
|*  Great Quotations of Our Time: #1 in a Series			   *|
|*  ============================================			   *|
|*									   *|
|*  "I'm not a referential transparency man, and you can quote me on that" *|
|*		-- Meurig Beynon, November 1988.			   *|
|*									   *|
\***************************************************************************/

void shmortal (message, notme, type)
  char *message;
  int notme;
  mesgtype type;
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((player[i].fd != -1) && (i != notme) && (!WIZMODEP (i)))
	if ( do_shout ( i, type ) )
	    safewrite (i, message);
}

/*--------------------------------------------------------------------------*/

void shwiz (message, notme, type)
  char *message;
  int notme;
  mesgtype type;
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((player[i].fd != -1) && (i != notme) && (WIZMODEP (i)))
	if ( do_shout ( i, type ) )
	    safewrite (i, message);
}

/*--------------------------------------------------------------------------*/

void speak (message, notme, here)
  char *message;
  int notme;
  int here;
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((player[i].fd != -1) && (player[i].where == here) && (i != notme))
      safewrite (i, message);
}

/*--------------------------------------------------------------------------*/

void speak2 (message, playerno, nothim )
  char *message;
  int playerno;
  int nothim;
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((player[i].fd != -1) && (player[i].where == player[playerno].where) && (i != playerno) && ( i != nothim ))
      safewrite (i, message);
}

/*-------------------------------------------------------------------------*/

main (tmpc, tmpv)
  int tmpc;
  char **tmpv;
{
  int i, socketfd, len, newfd, newuser = 0, one = 1;
				/* Hideous kludge.  Turning option on. */
  char temp[LINELEN];
  char outmessage[LINELEN];
  struct sockaddr_in s_in, from;
  fd_set fdsready;
  int maxfds = getdtablesize ();
  /* 2001: Remove nonsensical #ifndef DEBUG bracket on next line */
  int status = 0;

				/* Fish I might use both... who cares! */
  srandom (time(0));
  srand (time(0));
	   
/***************************************************************************\
|* 								 	   *|
|*  Simple initialisation - Check that we are running on the right	   *|
|*  machine, and die if we are not.  Otherwise, set up the static	   *|
|*  external variables argc and argv, from tmpc and tmpv.  The catch	   *|
|*  HUP, INT (Ctrl-C), TSTP (Ctrl-Z) and throw them away.  Catch	   *|
|*  SIGTERM (software kill) and make it tidy up the daemon and die.	   *|
|*  									   *|
\***************************************************************************/

#if 0				/* 2001: snipped (it's a stupid thing to do) */
  if (gethostname (temp, LINELEN) == -1)
    die (14);
  else
    if (strcmp (temp, HOSTMACHINE)) {
      (void) printf ("MUNDI daemon can only be booted on %s.\n", HOSTMACHINE);
      quit (15);
    }
#endif /*0*/

  argc = tmpc;
  for (i = 0; i < argc; i++) {
    if ((argv[i] = malloc ((unsigned) strlen (tmpv[i])+1)) == NULL)
      die (12);
    (void) strcpy (argv[i], tmpv[i]);
  }
  (void) strcpy (argv[0], basename (argv[0]));

  if ((signal (SIGHUP,  SIG_IGN) == BADSIG) ||
#ifndef DEBUG
      (signal (SIGINT,  SIG_IGN) == BADSIG) ||
      (signal (SIGQUIT, SIG_IGN) == BADSIG) ||
      (signal (SIGTSTP, SIG_IGN) == BADSIG) ||
#endif /* DEBUG */
      (signal (SIGTERM, killed)  == BADSIG) ||
      (signal (SIGPIPE, sigpipe) == BADSIG))
    die (11);

/***************************************************************************\
|*  									   *|
|*  Magic.  Frog about with INET sockets.  Open the socket, and select	   *|
|*  the reuse-address option so that when a daemon has died, another	   *|
|*  can instantly be booted using the same INET port number.  Set the	   *|
|*  structure s_in to be a string of null bytes, the fill in its	   *|
|*  members.  Bind it to the socket.  Tell the socket to keep a		   *|
|*  backlog of up to 5 attempted logins.				   *|
|* 								 	   *|
\***************************************************************************/

  if ((socketfd = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    die (1);			/* Bring the socket into existance */

  if (setsockopt (socketfd, SOL_SOCKET, SO_REUSEADDR, (char*) &one,
		  sizeof (int)) == -1)
    die (19);			/* Make the address reusable when we die */

  bzero ((char*) &s_in, sizeof (s_in));
				/* Blank out the INET socket structure */
  s_in.sin_family = AF_INET;	/* Tell it that is INET domain, not UNIX */
  s_in.sin_addr.s_addr = INADDR_ANY;
				/* Tell it to accept any incoming calls? */
  /* 2001: added htons() invocation in next line.  How did THAT get missed? */
  s_in.sin_port = htons(INETPORT); /* Tell it to connect to the right INET no. */
  if (bind (socketfd, (struct sockaddr*) &s_in, sizeof (s_in)) == -1)
    die (2);			/* Actually do the above stuff */

  if (listen (socketfd, MAXBACKLOG))
    die (4);			/* Keep biggest backlog of calls we can */

/***************************************************************************\
|*  									   *|
|*  Turn the program into a daemon - print message and get data,	   *|
|*  disociate from process group, disociate from terminal, close all	   *|
|*  open files except socket and standard output, redirect standard	   *|
|*  output to logfile, print login message to logfile, and put process	   *|
|*  in background.							   *|
|* 								 	   *|
\***************************************************************************/

  (void) printf ("Booting MUNDI daemon, version %s\n", VERSION);

#ifndef DEBUG
  (void) setpgrp (0, getpid ());
  (void) setuid ( geteuid () );

  if ((i = open ("/dev/tty", O_RDWR)) == -1)
    /*die (20)*/;		/* 2001: ENOTTY on Solaris 2.6; maybe OK? */
  else
    if (ioctl (i, TIOCNOTTY) == -1)
      die (21);

  for (i = maxfds-1; i >= 0; i--)
    if ((i != 1) && (i != socketfd))
      (void) close (i);

  (void) sprintf (temp, "%s/%s", MUNDI_DIR, LOGFILE);
  if (freopen (temp, "a", stdout) == NULL)
    die (22);
#endif /* DEBUG */

  logerror ("MUNDI daemon (version %s) booted\n", VERSION);

  if (
#ifndef DEBUG			/* 2001: added this wrapping */
      ! findarg ( argc, argv, "-fork" ) 
#else /*DEBUG*/
      0
#endif /*!DEBUG*/
      ) {
      if ((status = fork ()) == -1)
	  die (9);
      else
	  if (status)
	      exit (0);
  } else {
      logerror ( "WARNING: Daemon is not forking.\n" );
  }

  init_events ();		/* Initialise event handler */
  block_events ();		/* Not yet! */

  initialise ();

/****************************************************************************\
  Any initial events should be placed in ev_startevents ("events.c",
  and declared in "events.h")
\****************************************************************************/
  add_event ( 0, 0, ev_startevents, "" ); 

/***************************************************************************\
|*  									   *|
|*  Clear all bits in the mask of which file-descriptors are ready to	   *|
|*  be read, and set all the users' file-descriptor fields to -1	   *|
|*  (marking "unused").  Then enter main program loop.			   *|
|*  									   *|
\***************************************************************************/

  FD_ZERO (&fdsinuse);
  for (i = 0; i < MAXUSERS; i++) {
    player[i].fd = -1;
    player[i].snoop = -1;
    player[i].wizstuff = 0;
  }
  
  while (1) {
    (void) fflush (stdout);	/* Write any outstanding messages. */
    fdsready = fdsinuse;	/* Set all client fd's ready to trigger */
    FD_SET (socketfd, &fdsready); /* Set parent-socket ready to trigger */
    unblock_events ();		/* Yo - events get down and boogey! */
    if (select (maxfds, &fdsready, (fd_set*) 0, (fd_set*) 0,
		(struct timeval*) 0) == -1)
      if (errno == EINTR) {	/* Call was interrupted by, for instance */
				/* commencement of tracing (trace -p) */
	check_events ();	/* Check to see if some events are due */
	continue;		/* so don't worry about it. */
      } else			/* Any other errno means genuine tea, so die */
	die (6);
    
    check_events ();		/* Check yer events here */
    block_events ();		/* Sorry no events now... */
    if (FD_ISSET (socketfd, &fdsready)) {
      len = sizeof (from);	/* You have to do this - don't ask me why! */
      if ((newfd = accept (socketfd, (struct sockaddr*) &from, &len)) == -1)
        die (5);
      else {
	newuser = 0;
	while ((newuser < MAXUSERS) && (player[newuser].fd != -1))
	  newuser++;
	if (newuser == MAXUSERS) {
	  logerror ("Had to reject a client - too full\n");
	  (void) sprintf (outmessage, "%s\n", NO_CONNECTION);
	  if (write (newfd, outmessage, strlen (outmessage)) == -1)
	      if (errno != EPIPE )
		  die (8);	/* Can't use safewrite (snoop code) */
	  (void) close (newfd);
	}
	else {
	  player[newuser].fd = newfd;
	  player[newuser].clientmode = AWAITINGNAME;
	  FD_SET (newfd, &fdsinuse);
	  (void) sprintf (outmessage, "%s\n", CONNECTION);
	  if (write (newfd, outmessage, strlen (outmessage)) == -1)
	    die (8);		/* Can't use safewrite (snoop code) */
	}
      }
    }

    for (i = 0; i < MAXUSERS; i++) {
      unblock_events ();	/* Unblock for a minute */
      check_events ();		/* Check some events man! */
      block_events ();		/* Block em up for a minute */
      if (cleanup == i) {	/* Handle SIGPIPE as low down as possible */
	logerror ( "SIGPIPE: Cleaning up client#%d\n", i );
        tidyup (i);
        cleanup = -1;
      }
      if (player[i].fd != -1 && FD_ISSET (player[i].fd, &fdsready))
	handletext (i);
    }
  }
}

/*--------------------------------------------------------------------------*/





