/***************************************************************************\
|*								  	   *|
|*  Daemon.h:       Server program for what will eventually be MUNDI.	   *|
|*                  This version started on or about 18th May 1988.	   *|
|*  		NB.  This is a header file - NOT a source file!		   *|
|*								  	   *|
\***************************************************************************/

#define	VERSION "(5.1)"
#define LOGFILE	"lib/system/log"
#define SAVE_DIR "lib/save"
#define OBJECTS_SAVE "objects.sav"
#define LOCATIONS_SAVE "locations.sav"

/*-------------------------------------------------------------------------*/

typedef int
  exitlist[MAXDIRECTIONS];	/* Baaaad stuff, typedef syntax! */

struct synonym {		/* All component words are equivalent */
  int wordc;			/* How many synonyms are there? */
  char wordv[SYNONYMS][WORDLEN]; /* What actually ARE they? */
};

struct direction {
  char desc[LINELEN];		/* Capital letter, full length */
  char name[LINELEN];		/* No capital - used by player */
  char code[WORDLEN];		/* Codename, used as abbreviation. */
};

struct location {
  char *desc;			/* Must be malloc(3)d to right size */
  char name[LINELEN];		/* Short name, used in brief mode */
  char code[WORDLEN];		/* Codename, used for teleport etc. */
  exitlist exits;	        /* Locations reached by given exits */	
  int  flags;			/* What sort of location is it ? */
};

struct object {
  char desc[LINELEN];		/* Description used by program */
  struct synonym name;		/* Synonym names used by player */
  int location;			/* Position, index into loc[] */
  int how;			/* How it is at location (in, on etc) */
  int value;			/* Value in terms of being t */
  int weight;			/* Weight in grammes */
  int size;			/* User by container routines */
  int flag;			/* Can it be taken, worn, etc? */
  int container;		/* In what sense is it a container? */
  int seen;			/* For "goto" commands etc. */
  int damage;			/* Damage done when killing */
  int key;			/* Object number to unlock this object */
				/* Bit 4 of container must be set */
  int speed;			/* percentage chance of action */
				/* One object is checked for action */
				/* every ten seconds */
  int actions;			/* Actions that object may perform */
};

struct howstruct {
    char name[LINELEN];		/* How the object is located */
    int  whatto;		/* What it is located to loc/obj (0/1) */
};

typedef void (*PFV)();		/* Pointer to void returning func */

typedef enum {			/* These next are the ONLY legal types */
				/* of command */
    NO_ARG,			/* void */
    P_ARG,			/* int playerno */
    PC_ARG,			/* int playerno, char *command */
    PC1_ARG,			/* int playerno, char *command, */
				/* char *opt_arg1 */		
    PC12_ARG			/* int playerno, char *command, */
				/* char *opt_arg1, char *opt_arg2 */
} argtype;
	
typedef struct {
    char     *command;
    argtype   type;		/* Type of arguments - or if enabled! */
    char     *opt_arg1,
             *opt_arg2;
    PFV       handler;
    int	      flags;		/* Bit 0 - 1 - If cannot be disabled */
				/* Bit 1 - 1 - If disabled */
} COMMAND;

  

#define CANNOTDISABLE	1
#define DISABLED	2

#define ENABLEDP(ind)     (!(comms[ind].flags & DISABLED))
#define CANDISABLEP(ind)  (!(comms[ind].flags & CANNOTDISABLE))

#define WHAT_LOCATION   0
#define WHAT_PERSONA    1
#define WHAT_OBJECT     2

extern char *argv[MAXARGS];
extern int argc, no_directions, no_locations, no_objects, no_personas,
  no_users, no_shouts, no_sayings, no_hows;

extern int anim_list[MAXOBJECTS]; /* List of animate objects */
extern int num_anim;		/* Number of animate objects */

extern struct direction dir[MAXDIRECTIONS];
extern struct location loc[MAXLOCATIONS];
extern struct object obj[MAXOBJECTS];
extern struct howstruct how[MAXHOWS];
extern COMMAND comms[MAXCOMMANDS];
				/* Opening times of mundi */
extern int times[DAYSINWEEK][HOURSINDAY];
extern char days[DAYSINWEEK][MAXDAYLEN];

extern char sayings[MAXSAYINGS][LINELEN];
extern char shouts[MAXSHOUTS][LINELEN];

/*-------------------------------------------------------------------------*/

extern void quit ();
extern void die ();
extern char *saferead ();
extern void send_reply ();
extern void safewrite ();
extern void shout ();
extern void shmortal ();
extern void shwiz ();
extern void speak ();
extern void logerror ();
extern int safesystem ();
extern char *basename ();
extern void speak2 ();

/*-------------------------------------------------------------------------*/


#define moveablep(ind)    (obj[ind].flag & (1<<0))
#define wearablep(ind)    (obj[ind].flag & (1<<1))
#define drinkablep(ind)	  (obj[ind].flag & (1<<2))
#define poisonp(ind)	  (drinkablep(ind) && obj[ind].damage < 0)
#define ediblep(ind)	  (!drinkablep(ind) && \
			   obj[ind].size && obj[ind].weight)
#define fireablep(ind)	  (obj[ind].flag & (1<<3))
#define singleshotp(ind)  (fireablep(ind) && (!(obj[ind].flag & (1<<4))))
#define multishotp(ind)	  (fireablep(ind) && (!(singleshotp(ind))))
#define nuclearp(ind)	  (obj[ind].flag & (1<<5))
#define nuclearsavep(ind) (obj[ind].flag & (1<<6))
#define weaponp(ind)	  (!(obj[ind].flag & (1<<7)) && obj[ind].damage > 0)
#define cursedweaponp(ind) (!(obj[ind].flag & (1<<7)) && obj[ind].damage < 0)
#define explodep(ind)	  (obj[ind].flag & (1<<8))


#define treasurep(ind)	  (obj[ind].value > 0)
#define antitreasurep(ind) (obj[ind].value < 0)

#define transparentp(ind) (obj[ind].container & (1<<0))
#define openp(ind)	  (obj[ind].container & (1<<1))
#define openablep(ind)	  (obj[ind].container & (1<<2))
#define surfacep(ind)	  (obj[ind].container & (1<<3))
#define containerp(ind)   ((openp(ind) || openablep(ind)) && !surfacep(ind))
#define lockablep(ind)	  (obj[ind].container & (1<<4))
#define lockedp(ind)	  (obj[ind].container & (1<<5))
#define unlockedp(ind)	  (!lockedp(ind))
#define objectkey(ind)	  (lockablep(ind) ? obj[ind].key : -1)
#define animatep(ind)     ((obj[ind].speed > 0) && (obj[ind].actions != 0))

#define act_movep(ind)    (obj[ind].actions & (1<<0))
#define act_openp(ind)	  (obj[ind].actions & (1<<1))
#define act_sayp(ind)     (obj[ind].actions & (1<<2))
#define act_takep(ind)    (obj[ind].actions & (1<<3))
#define act_dropp(ind)    (obj[ind].actions & (1<<4))
#define act_lockp(ind)    (obj[ind].actions & (1<<5))
#define act_unlockp(ind)  (obj[ind].actions & (1<<6))
#define act_shoutp(ind)   (obj[ind].actions & (1<<7))
#define act_givep(ind)    (obj[ind].actions & (1<<8))
#define act_stealp(ind)   (obj[ind].actions & (1<<9))
#define act_attackp(ind)  (obj[ind].actions & (1<<10))
#define NUM_ACTS	  11	/* How many different flags? */

/* write */
#define loc_pitp(ind)        (loc[ind].flags & (1<<0))
#define loc_noteletop(ind)   (loc[ind].flags & (1<<1))
#define loc_notelefrp(ind)   (loc[ind].flags & (1<<2))
#define loc_echop(ind)       (loc[ind].flags & (1<<3))
#define loc_safeforcep(ind)  (loc[ind].flags & (1<<4))
#define loc_safestealp(ind)  (loc[ind].flags & (1<<5))
#define loc_safekillp(ind)   (loc[ind].flags & (1<<6))
#define loc_safesnoopp(ind)  (loc[ind].flags & (1<<7))
#define loc_safesummonp(ind) (loc[ind].flags & (1<<8))

#define exit_name_w(x)  ( ( x < -1 ) ? "-" : "" ),  ( ( x == 0 ) ? "NO_WHERE" : ( ( x == -1 ) ? "LIMBO" : ( ( x < 0 ) ? loc[-x].code : loc[x].code ) ) )

#define exit_name(x)  ( ( x < -1 ) ? "-" : "" ),  ( ( x == 0 ) ? "     " : ( ( x == -1 ) ? "LIMBO" : ( ( x < 0 ) ? loc[-x].code : loc[x].code ) ) )

/*--------------------------------------------------------------------------*/

typedef enum {
    MESG_WARN,			/* Warning (unblockable) */
    MESG_SPEAK,			/* Spoken mesg */
    MESG_SHOUT,			/* Shouted mesg */
    MESG_GLOBAL			/* Information mesg */
} mesgtype;

