/***************************************************************************\
|*                                                                         *|
|*  Handle.c:	Part of the server program for MUNDI - handle input lines. *|
|*		This version started on or about 26th May 1988.		   *|
|*                                                                         *|
\***************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/file.h>

#include "global.h"
#include "daemon.h"
#include "look.h"
#include "handle.h"
#include "persona.h"
#include "utils.h"
#include "magic.h"
#include "verbs.h"
#include "verbs2.h"
#include "schedule.h"
#include "events.h"
#include "animation.h"

/*--------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *malloc ();
extern char *realloc ();
extern long random ();

void disable_cmd ();

/****************************************************************************\
  Returns the direction number equivalent to string, or -1 if none
\****************************************************************************/

int strtodir ( dirstr )
int dirstr;
{
    int i, d = -1;
    for ( i = 0; i < no_directions; i++ )
	if ( samestring ( dir[i].name, dirstr ) ) {
	    d = i;
	    break;
	}
    return ( d );
}

/****************************************************************************\
  Returns the new location number, or a defined constant if invalid.
  Also changes direction for pit movemt.
\****************************************************************************/

int getnewloc ( playerno, direction )
int playerno;
int *direction;
{
    int newloc, i;

    if ( *direction == -1 )
	return ( DIR_NOEXIT );

    if ((newloc = loc[player[playerno].where].exits[*direction]) == 0) {
	if (*direction != 0)
	    return ( DIR_NOEXIT ); /* No exit */
	else if (player[playerno].where == PITLOC)
	    return ( DIR_ATPIT ); /* We are at the pit */
	else 
	    return ( DIR_NOPITDIR ); /* There is no pit exit */
    }

    if ( *direction == 0)
	for (i = 1; i < no_directions; i++)
	    if (loc[player[playerno].where].exits[i] ==
		loc[player[playerno].where].exits[0])
		*direction = i;
    
    if ((newloc == -1) || (newloc >= no_locations))
	newloc = 0;
    if (newloc == UPITLOC && !ARCHMODEP (playerno))
	newloc = 0;

    return ( abs ( newloc ) );
}

/*--------------------------------------------------------------------------*/

int strtoplayer (victimstr)
  char *victimstr;
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((ISACTIVE (i)) && (samestring (player[i].name, victimstr)))
      return (i);
  
  return (-1);
}

/*--------------------------------------------------------------------------*/

int strtopersona (victimstr)
  char *victimstr;
{
  int i;
  
  for (i = 1; i < no_personas; i++)
    if (samestring (persona_list[i].name, victimstr))
      return (i);

  return (-1);
}

/****************************************************************************\
  Must supply exact matching names for both of these, used in arch
  wizard commands that need to be preciese.
\****************************************************************************/

int strmatchplayer (victimstr)
  char *victimstr;
{
  int i;

  for (i = 0; i < MAXUSERS; i++)
    if ((ISACTIVE (i)) && (samenstring (player[i].name, victimstr)))
      return (i);
  
  return (-1);
}

/*--------------------------------------------------------------------------*/

int strmatchpersona (victimstr)
  char *victimstr;
{
  int i;
  
  for (i = 1; i < no_personas; i++)
    if (samenstring (persona_list[i].name, victimstr))
      return (i);

  return (-1);
}

/****************************************************************************\
  Max changed this function, which now tries to match the full name, then
  recursively calls for any containers held, then for sub components of the
  word (using synonyms).

  It's a pile of old carp -- Mirk (Sat Feb 24 14:29:24 GMT 1990)

\****************************************************************************/

int find_object ( name, objv, objc, here )
char *name;
char **objv;
int objc;
int here;
{
    int i, j, k;
				/* Check inside any container here */
    for (i = 0; i < no_objects; i++) {
	if ((obj[i].location == here) && (((obj[i].container) & 10) != 0)) {
	    if ( ( j = find_object ( name, objv, objc, -i) ) != -1 )
		return ( j );
	}
    }
				/* Try matching full name */
    for (i = 0; i < no_objects; i++) {
	if ( samenstring ( obj[i].desc, name ) )
	    return(i);
    }
				/* Try matching one component of string */
				/* to each of the synonyms */
    for (i = 0; i < no_objects; i++) {
	for (j = 0; j < obj[i].name.wordc; j++) {
	    for (k = 0; k < objc; k++ ) {
		if ( samestring (obj[i].name.wordv[j], objv[k]))
		    return (i);
	    }
	}
    }
    return (-1);		/* Couldnt find the item */
}

/*--------------------------------------------------------------------------*/

int strtonoun (string, here)
  char *string;
  int here;
{
#define MAXOBJW 100		/* Number of words in an object (maximum) */
				/* Ps. this is horrible code */
  static char objv[MAXOBJW][LINELEN];
  static char *objvp[MAXOBJW];
  static char name[LINELEN];
  static char temp[LINELEN];
  int i = 0, ret;

  strcpy ( name, string );
  strcpy ( temp, string );

				/* Preprocess string into components */
  while ( i < MAXOBJW ) {
      if ( *string == '\0' )
	  break;
      getnextword ( objv[i], temp );
      objvp[i] = objv[i++];
  }
				/* Try and find the object */
  if ( (ret = find_object ( name, objvp, i, here )) == -1 )
      return ( -1 );
  strcpy ( string, obj[ret].desc );
  return ( ret );
}
/*--------------------------------------------------------------------------*/

int strtoroomcode ( place )
    char *place;
{
    int i; 
    for ( i = 0; i < no_locations; i++)
	if (samestring (loc[i].code, place))
	    return ( i );
    return ( -1 );
}

/*--------------------------------------------------------------------------*/

void tidyup (playerno)
{
  extern fd_set fdsinuse;	/* It's really a bit naughty to include */
				/* this here when in a sense it belongs */
				/* in daemon.h, but it is silly to have */
				/* to #include <sys/types.h> before     */
				/* daemon.h in every file that needs it */
  if ( player[playerno].fd != -1 ) {
      (void) close (player[playerno].fd);
      FD_CLR (player[playerno].fd, &fdsinuse);
      player[playerno].fd = -1;
  }
  player[playerno].record_no = 0;
  player[playerno].clientmode = AVAILABLE;
  *player[playerno].name = '\0';
}

/***************************************************************************\
|*                                                                         *|
|*  When a player signs off, is exorcised, has his process killed, or in   *|
|*  any other way leaves the game, this routine is called.  It does the    *|
|*  following:								   *|
|*                                                                         *|
|*	(1)  Make the dying player drop all his objects, with a one-in-    *|
|*		three chance that each object will magically vanish.       *|
|*	(2)  Terminate any snoop sessions on the dying player.             *|
|*	(3)  Close the dying player's socket (if not already closed)       *|
|*	(4)  Clear the bit on the fd_set which is select(2)d on.	   *|
|*	(5)  Reset his notional socket-fd to -1				   *|
|*	(6)  Reset clientmode to AVAILABLE.				   *|
|*	(7)  Clear his name to prevent erroneous PERSONABUSY messages.     *|
|*	(8)  Check if everyone is dead; if so, and there was a command-    *|
|*		line argument in the daemon's invocation, then die.        *|
|*                                                                         *|
|*  Cheese!  I've had to move numbers 3 to 7 above into the previous       *|
|*  function, tidyup(), so that this can be called without doing stuff     *|
|*  like logerror() and no_players-- when someone fails to login at all.   *|
|*                                                                         *|
\***************************************************************************/

void killoff (playerno)
  int playerno;
{
  int i;
  char temp[LINELEN];

  if ( player[playerno].fd == -1 ) {
      tidyup ( playerno );
      return;
  }

				/* Set totaltime */
  player[playerno].totaltime += (time(0) - player[playerno].logtime);

  save_persona (playerno, "");
      
  for (i = 0; i < no_objects; i++)
      if (obj[i].location == player[playerno].invloc)
	  if (((random () % 5) == 0) && (obj[i].value == 0)) {
	      obj[i].location = UPITLOC;
	      (void) sprintf (temp, "%s has to drop,, the %s, which vanishes!\n",
			      player[playerno].name, obj[i].desc);
	      speak (temp, playerno, player[playerno].where);
	  }
	  else {
	      obj[i].location = player[playerno].where;
	      (void) sprintf (temp, "%s has to drop the %s!\n",
			      player[playerno].name, obj[i].desc);
	      speak (temp, playerno, player[playerno].where);
	  }
  
  (void) sprintf (temp, "You stop snooping on %s.\n", player[playerno].name);
  for (i = 0; i < MAXUSERS; i++)
      if ((player[i].fd != -1) && (player[i].snoop == playerno)) {
	  safewrite (i, temp);
	  player[i].snoop = -1;
      }
  
  logerror ("Logout: client #%d, name \"%s\".\n", playerno,
	    player[playerno].name);

  tidyup (playerno);
  
  if ((--no_users == 0) && (argc > 1)) {
      logerror ("Clients have all died, so I am dying too\n");
      quit (0);
  }
}

/*--------------------------------------------------------------------------*/

void visikilloff (playerno)
  int playerno;
{
  char temp[LINELEN];

  (void) sprintf (temp, "*** %s has just signed off\n",
		  player[playerno].name);
  killoff (playerno);
  shout (temp, playerno, MESG_GLOBAL);
}

/*--------------------------------------------------------------------------*/

void quit_game (playerno)
  int playerno;
{
  say (form ("%s\n", QUIT_MUNDI));
  visikilloff (playerno);
}

/*--------------------------------------------------------------------------*/

void scorepoints (playerno, value, incrhp, show )
  int playerno;
  int value;
  int incrhp;			/* 1 - If to increase hp/maxhps */
  int show;			/* 1 - if to show points got */
				/* 2 - Spell cost */
				/* 3 - Show points lost */
{
  char outmessage[LINELEN];
  int  current, sc_level, delta = 0;
  char *reply;

  player[playerno].newscore += value;

  current = player[playerno].newscore + player[playerno].oldscore;

  sc_level = compute_level ( sc_levels, current, SC_LEVELS );

  if ( incrhp ) {
      if ( value <= -100 )
	  delta = value/100 - 3;
      else if ( value > -100 && value <= -60 )
	  delta = -3;
      else if ( value > -60 && value <= -30 )
	  delta = -2;
      else if ( value > -30 && value <= -5 )
	  delta = -1;
      else if ( value > -5 && value <= 5 )
	  delta = 0;
      else if ( value > 5 && value <= 30 )
	  delta = 1;
      else if ( value > 30 && value <= 60 )
	  delta = 2;
      else if ( value > 60 && value <= 100 )
	  delta = 3;
      else if ( value > 100 )
	  delta = value/100 + 3;

      if ( value >= 0 )
	  player[playerno].maxhp += delta;
      else player[playerno].hp += delta;
      
      if ( player[playerno].maxhp < 10 )
	  player[playerno].maxhp = 10;

      reply = heal_player ( playerno, delta, 0, 1 );
      if ( reply )
	  say ( reply );
  }

  if ( show == 1 )
      (void) sprintf (outmessage, "*** You score %d point%s!\n", 
		      (value < 0) ?
		      (WIZMODEP(playerno) ? value : -value) :
		      value, PLURAL (value));
  else if ( show == 2 )
      (void) sprintf (outmessage, "That spell cost you %d points!\n",
		      -value );
  else if ( show == 3 )
      (void) sprintf (outmessage, "*** You loose %d point%s!\n",
		      ((value < 0) ? -value : value), PLURAL(value) );

  if ( show )
      say (outmessage);

  if ( player[playerno].level != sc_level ) {

      int i, old_level = player[playerno].level;

      if ( sc_level > old_level ) {
	  if ( sc_level >= MIN_WIZLEVEL && !CANWIZP(playerno)) {
				/* Now a wiz */
	      (void) sprintf ( outmessage,
			      "*!* Frog me you're a wizard mate! *!*\n" );
	      shout ( form ( "*!* %s has attained Wizard-ship *!*\n",
			    player[playerno].name ), NO_ONE, MESG_GLOBAL );
	      player[playerno].wizstuff |= CANWIZ;
	  } else
	      (void) sprintf ( outmessage, "\
*** Welcome to level %d, %s! ***\n",
			      sc_level, sc_levels[sc_level].name );

	  calculate_stats( playerno, sc_level);
	  
      } else {
	  if ( sc_level < MIN_WIZLEVEL 
	      && CANWIZP(playerno)
	      && !CANARCHP(playerno)
	      && !CANMETAP(playerno)) {
	      player[playerno].wizstuff = 0;
	      (void) sprintf ( outmessage, "\
*!* Your Wizard status has been revoked! *!*\n" );
	      shout ( form ( "*!* %s has lost Wiz-status! *!*\n",
			    player[playerno].name ), NO_ONE, MESG_GLOBAL );
	  } else
	      (void) sprintf ( outmessage, "\
*** You have been demoted to level %d, %s! *** \n",
			      sc_level, sc_levels[sc_level].name );

	  calculate_stats( playerno, sc_level );
      }
      say (outmessage);

  }
}

/*--------------------------------------------------------------------------*/

void error (playerno, verb, rest)
  int playerno;
  char *verb;
  char *rest;
{
  if (*rest == '\0')
    say (form ("I don't know how to \"%s\".\n", verb));
  else
    say (form ("I don't know how to \"%s\" something.\n", verb));
}

/*--------------------------------------------------------------------------*/
/****************************************************************************\
  Max changed the way commands are parsed from the old style (which is
  #ifdef'ed out later), to a command table. This of course makes life harder
  since some functions required differing arguments.

  Ok so what does the table mean? Easy;

  {
    char    *command;		Name of the command
    argtype  type;			Arguments to pass;
				P_ARG    - pass playerno
				PC_ARG   - pass playerno, command
				PC1_ARG  - pass playerno, command, opt_arg1
				PC1_ARG  - pass playerno, command, opt_arg[12]
    char    *opt_arg1,		Optional string argument 1
            *opt_arg2;		Optional string argument 2
    PFV      handler;		Function which carries out the action
  }

\****************************************************************************/

COMMAND comms[MAXCOMMANDS] = {
    { "look",		P_ARG,	NULL,	NULL,	look		},
    { "who",		P_ARG,	NULL,   NULL,   who     	},
    { "quit",		P_ARG,	NULL,   NULL,   quit_game	},
    { "crash",		P_ARG,	NULL,   NULL,   crash     	},
    { "score",  	P_ARG,	NULL,	NULL,	printscore	},
    { "inventory",	P_ARG,	NULL,	NULL,	inventory	},
    { "get",		PC_ARG, NULL,	NULL,	getstring	},
    { "take",		PC_ARG, NULL,	NULL,	getstring	},
    { "drop",		PC_ARG, NULL,	NULL,	dropstring	},
    { "leave",		PC_ARG, NULL,   NULL,   dropstring	},
    { "open",		PC_ARG, NULL,	NULL,	openstring	},
    { "close",		PC_ARG,	NULL,	NULL,	closestring	},
    { "put",		PC_ARG, NULL,   NULL,   putstring       },
    { "place",		PC_ARG, NULL,   NULL,   putstring       },
    { "lock",		PC1_ARG,
	  "lock",	NULL,   lockstring      },
    { "unlock",		PC1_ARG,
	  "unlock",	NULL,   lockstring      },
    { "shout",		PC_ARG,	NULL,	NULL,	shoutmess	},
    { "exorcise",	PC_ARG,	NULL,	NULL,	exorcise	},
    { "summon",		PC_ARG,	NULL,	NULL,	summon		},
    { "snoop",		PC_ARG,	NULL,	NULL,	snoopon		},
    { "unsnoop",	P_ARG,	NULL,	NULL,	snoopoff	},
    { "superex",	PC_ARG,	NULL,	NULL,	superex		},
    { "where",		PC_ARG,	NULL,	NULL,	wherestring	},
    { "wizmode",	PC_ARG,	NULL,	NULL,	gowiz		},
    { "nowizmode",	PC_ARG,	NULL,	NULL,	nowiz		},
    { "archmode",	PC_ARG,	NULL,	NULL,	goarch, 1	},
    { "noarchmode",	PC_ARG,	NULL,	NULL,	noarch, 1	},
    { "metamode",	PC_ARG,	NULL,	NULL,	gometa, 1	},
    { "nometamode",	PC_ARG,	NULL,	NULL,	nometa, 1	},
    { "verbose",	P_ARG,	NULL,	NULL,	verbose		},
    { "brief",		P_ARG,	NULL,	NULL,	brief		},
    { "exits",		P_ARG,	NULL,	NULL,	exits		},
    { "noexits",	P_ARG,	NULL,	NULL,	noexits		},
    { "objects",	P_ARG,	NULL,	NULL,	objects		},
    { "noobjects",	P_ARG,	NULL,	NULL,	noobjects	},
    { "allexits",	P_ARG,	NULL,	NULL,	allexits	},
    { "uniqueexits",	P_ARG,	NULL,	NULL,	uniqueexits	},
    { "numbers",	P_ARG,	NULL,	NULL,	numbers		},
    { "nonumbers",	P_ARG,	NULL,	NULL,	nonumbers	},
    { "ignore",		PC_ARG, NULL,   NULL,   ignore		},
    { "notice",		PC_ARG, NULL,   NULL,   notice		},
    { "teleport",	PC_ARG,	NULL,	NULL,	teleport	},
    { "force",		PC_ARG,	NULL,	NULL,	force		},
    { "animate",	PC_ARG, NULL,   NULL,   animate_cmd	},
    { "deanimate",	PC_ARG, NULL,   NULL,   deanimate_cmd	},
    { "haste",		PC_ARG, NULL,   NULL,   haste_cmd	},
    { "slow",		PC_ARG, NULL,   NULL,   slow_cmd	},
    { "wizact",		PC_ARG,	NULL,	NULL,	wizact		},
    { "archact",	PC_ARG,	NULL,	NULL,	archact		},
    { "stat",		PC_ARG,	NULL,	NULL,	statstring	},
    { "save",		PC_ARG,	NULL,	NULL,	save_persona	},
    { "delete",		PC_ARG,	NULL,	NULL,	delete_string	},
    { "sync",		P_ARG,	NULL,	NULL,	write_pwfile	},
    { "plist",		PC_ARG,	NULL,	NULL,	plist		},
    { "match",		PC_ARG, NULL,   NULL,   match		},
    { "frog",		PC_ARG,	NULL,	NULL,	frog		},
    { "tea",		PC_ARG, NULL,   NULL,   tea		},
    { "decorate",	PC_ARG, NULL,   NULL,   decorate	},
    { "mkobject",	P_ARG,  NULL,   NULL,   mkobject        },
    { "mkroom",		P_ARG,  NULL,   NULL,   mkroom          },
    { "saveobj",	PC_ARG, NULL,	NULL,   saveobj		},
    { "saveroom",	PC_ARG, NULL,	NULL,	saveroom	},
    { "events",		P_ARG,  NULL,   NULL,   show_events     },
    { "delay",		PC_ARG, NULL,   NULL,   delay_cmd       },
    { "disable",	PC1_ARG,
	  "disable",   NULL,   disable_cmd     },
    { "enable",		PC1_ARG,
	  "enable",    NULL,   disable_cmd      },
    { "kludge",		PC_ARG, NULL,   NULL,   kludge_events   },
    { "shoot",		PC_ARG, NULL,   NULL,   fire  		},
    { "fire",		PC_ARG, NULL,   NULL,   fire  		},
    { "kill",		PC_ARG, NULL,   NULL,   killstring	},
    { "quaff",		PC_ARG, NULL,   NULL,   quaff		},
    { "drink",          PC_ARG, NULL,   NULL,   quaff           },
    { "levels",		P_ARG,  NULL,   NULL,   levels          },
    { "steal",		PC_ARG, NULL,   NULL,   steal		},
    { "give",		PC_ARG, NULL,	NULL,	give		},
    { "throw",          PC_ARG, NULL,   NULL,   throw           },
    { "examine",	PC_ARG, NULL,   NULL,   examine		},
    { "sayto",		PC_ARG, NULL,   NULL,   sayto		},
    { "purity",		P_ARG,  NULL,   NULL,   purity          },
    { "fatness",	P_ARG,  NULL,   NULL,   fatness         },
    { "commands",	P_ARG,  NULL,   NULL,   print_commands  },
    { "superget",	PC_ARG, NULL,	NULL,	sgetstring	},
    { "sg",		PC_ARG, NULL,	NULL,	sgetstring	},
    { "eat",		PC_ARG, NULL,   NULL,   eat		},
    { "win",		P_ARG,  NULL,   NULL,   win		},
    { "kiss",		PC12_ARG,
			"kiss",
			"kisses",
			kiss					},
    { "hug",		PC12_ARG,
			"hug",
			"hugs",
			kiss					},
    { "cuddle",		PC12_ARG,
			"cuddle",
			"cuddles",
			kiss					},
    { "stroke",		PC12_ARG,
			"stroke",
			"strokes",
			kiss					},
    { "punch",		PC12_ARG,
			"punch",
			"punches",
			kiss					},
    { "kick",		PC12_ARG,
			"kick",
			"kicks",
			kiss					},
    { "hit",		PC12_ARG,
			"hit",
			"hits",
			kiss					},
    { "slap",		PC12_ARG,
			"slap",
			"slaps",
			kiss					},
    { "tickle",		PC12_ARG,
			"tickle",
			"tickles",
			kiss					},
    { "phlegm",		PC12_ARG,
			"phlegm",
			"phlegms",
			kiss					},
    { "poke",		PC12_ARG,
			"poke",
			"pokes",
			kiss					},
    { "forgive",	PC12_ARG,
			"forgive",
			"forgives",
			kiss					},
    { "jump",		PC12_ARG,
			"jump",
			"Wheeeee!\n",
			jump					},
    { "vomit",		PC12_ARG,
			"vomit",
			"Euurgh!\n",
			jump					},
    { "grin",		PC12_ARG,
			"grin",
			":->\n",
			jump					},
    { "smile",		PC12_ARG,
			"smile",
			":-)\n",
			jump					},
    { "smirk",		PC12_ARG,
			"smirk",
			"A?A?A?\n",
			jump					},
    { "laugh",		PC12_ARG,
			"laugh",
			"Hahaha!\n",
			jump					},
    { "giggle",		PC12_ARG,
			"giggle",
			"Hehehe!\n",
			jump					},
    { "sulk",		PC12_ARG,
			"sulk",
			":-(\n",
			jump					},
    { "blush",		PC12_ARG,
			"blush",
			"Oh! :-}\n",
			jump					},
    { "dance",		PC12_ARG,
			"dance",
			"Fairy!\n",
			jump					},
    { "sing",		PC12_ARG,
			"sing",
			"La la la\n",
			jump					},
    { "cry",		PC12_ARG,
			"cry",
			"Boo hoo!\n",
			jump					},
    { "pray",		PC12_ARG,
			"pray",
			"You immediately feel much better.\n",
			jump					},
    { "snigger",	PC12_ARG,
			"snigger",
			"Heh heh\n",
			jump					},
    { "apologise",	PC12_ARG,
			"apologise",
	  		"I hope you meant it!\n",
			jump					},
    { "enid",		PC12_ARG,
			"EnidBlyton",
			"o.\n",
			jump					},
    { "pout",		PC12_ARG,
			"pout",
			":-I\n",
			jump					},
    { "juggle",		PC12_ARG,
			"juggle",
			"You juggle nearly as well as Marv!\n",
			jump					},
    { "panic",		PC12_ARG,               
			"panic",
			"Argh!\n",
			jump					},
    { "winge",		PC12_ARG,
			"winge",
			"Pain!\n",
			jump					},
    { "whimper",	PC12_ARG,
			"whimper",
			"Wimp!\n",
			jump					},
    { "be",		PC12_ARG,
			"BE",
			"Yeah!\n",
			jump					},
    { "grovel",		PC12_ARG,
			"grovel",
			"Lick lick!\n",
			jump					},
    { "mince",		PC12_ARG,
			"mince",
			"Oooh, you sexy beast...\n",
			jump					},
    { "slime",		PC12_ARG,
			"slime",
	  		"slimes",
			kiss					},
    { "slidekick",	PC12_ARG,
			"sliding kick",
			"Wow.. witty!\n",
			jump					},
    { "cower",		PC12_ARG,
			"cower",
			"Mooooooooo...\n",
			jump					},
    { "moo",		PC12_ARG,
			"moo",
			"Fish.\n",
			jump					},
    { "quack",		PC12_ARG,
			"quack",
			"Togart would be proud of you.\n",
			jump					},
    { "cluck",		PC12_ARG,
			"cluck",
			"Pwawk pwawk pwaaawk!\n",
			jump					}, 
    { "squeeze",	PC12_ARG,
			"squeeze",
			"Mmmmm...\n",
			jump					},
    { "cook",		PC12_ARG,
			"cook",
			"Sizzle.\n",
			jump					},
    { "jumpkick",	PC12_ARG,
			"jumping-kick",
			"SPLAT 'em!\n",
			jump					},    
    { "freak",		PC12_ARG,
			"freak",
			"Whoop whoop!\n",
			jump					},
    { "mellow",		PC12_ARG,
			"mellow",
			"Yeah... listen to some Floyd man...\n",
			jump					},
    { "three",		PC12_ARG,
	  		"three",
			"Threeing three Blackadder!\n",
	  		jump					},
    { "five",		PC12_ARG,
	  		"five",
			"Five - like is cool!\n",
	  		jump					},
    { "bucket",		PC12_ARG,
	  		"bucket",
			"Woah - to many buckets spoil the game.\n",
	  		jump					},
    { NULL } /* 2001: removed extraneous elements in this last initialiser */
};


/*--------------------------------------------------------------------------*/

void commands (playerno, command)
  int playerno;
  char *command;
{
  int  direction = -1;
  char verb [LINELEN];
  char mesg [LINELEN];

  getnextword (verb, command);

				/* Handle inventory first */
  if (strcmp (verb, "i") && strcmp (verb, "I"))
      direction = strtodir ( verb );
				/* Look up command */

  if (direction != -1)             move        (playerno, direction);
  else {
      int i = 0, found = 0;

      while ( ! found ) {
				/* Check for end of table */
	  if ( comms[i].command == NULL )
	      break;
				/* Check for verb is command */
	  if ( VERBIS ( comms[i].command ) ) {
	      found = 1;
	      if ( ! ENABLEDP ( i ) && ! ARCHMODEP ( playerno ) ) {
		  say ( form ( "The %s command has been disabled!\n",
			      comms[i].command ) );
	      } else {
		  switch ( comms[i].type ) {
		  case NO_ARG:
				/* Yeah they can do that */
		      comms[i].handler ();
		      break;

		  case P_ARG:
				/* And that - BUT THATS IT */
		      comms[i].handler ( playerno );
		      break;

		  case PC_ARG:
		      comms[i].handler ( playerno, command );
		      break;

		  case PC1_ARG:
		      comms[i].handler ( playerno, command,
					comms[i].opt_arg1 );
		      break;

		  case PC12_ARG:
		      comms[i].handler ( playerno, command,
					comms[i].opt_arg1,
					comms[i].opt_arg2 );
		      break;

		  default:
		      error (playerno, verb, command); 
		      break;
		  }
	      }
	  } else
	      i++;
      }
      if ( ! found ) {
	  if ( strtoplayer ( verb ) != -1 ) {
				/* Cant use form here.... */
	      (void) sprintf ( mesg, "%s %s", verb, command );
	      sayto ( playerno, mesg );
	  }
	  else
	      error (playerno, verb, command); 
      }
  }
}

/***************************************************************************\
|*                                                                         *|
|*  handle() likes to be fed strings without a terminating '\n'.  This is  *|
|*  so that it can successfully compare against the PseudoSignals.  Each   *|
|*  line passed to handle() is a complete command or PseudoSignal -- all   *|
|*  line-breaks etc. have already been taken care of.			   *|
|*                                                                         *|
\***************************************************************************/

void handle (playerno, string)
  int playerno;
  char *string;
{
  int i;
  static char temp[LINELEN];

  (void) sprintf (temp, "| > %s\n", string);
  for (i = 0; i < MAXUSERS; i++)
    if ((player[i].fd != -1) && (player[i].snoop == playerno))
      safewrite (i, temp);

  switch (*string) {

  case '\0':
    say ("Stop that!  You *know* it's impossible.\n");
    break;

  case '"':
  case '\'':
  case '`':
    speak (form ("%s says \"%s\".\n", player[playerno].name, string+1),
	   playerno, player[playerno].where);

    if ( loc_echop(player[playerno].where) ) {
	add_event ( 5, 0, ev_saytoroom, "dsss", player[playerno].where,
		   "The words \"", string+1, "\" drift across the room...\n" );
	add_event ( 300, 0, ev_saytoroom, "dsss", player[playerno].where,
		   "The faint echos of \"", string+1,
		   "\" drift across the room...\n" );
    }
    break;


  case ESCAPE:
    if (!strcmp (string, SLEEPING)) {
      shout (form ("*** %s is taking a break.\n", player[playerno].name),
	     playerno, MESG_GLOBAL);
      player[playerno].clientmode = ASLEEP;
    }

    else if (!strcmp (string, WOKEN)) {
      shout (form ("*** %s is back again.\n", player[playerno].name),
	     playerno, MESG_GLOBAL);
      player[playerno].clientmode = PLAYINGGAME;
    }

    else if (!strcmp (string, QUIT_MUNDI))
      visikilloff (playerno);

    else
      shout ("Some weird thing dun happened!\n", NO_ONE, MESG_WARN);

    break;
    
  default:
    commands (playerno, string);
    break;
  }

  player[playerno].newmoves++;
}

/*--------------------------------------------------------------------------*/

void handleline (playerno, line)
  int playerno;
  char *line;
{
  char *cp;
  char *rest = line;		/* Rest of line starts as the whole line */

  while ((cp = index (rest, '\n')) != NULL) {
				/* Make cp point at the first '\n' */
    *cp = '\0';			/* Overwrite it with a NULL (terminator) */
    handle (playerno,rest);	/* Handle the message up to 1st '\n' */

				/* There was a bug here - that caused the  */
				/* daemon to die without saying why... */
				/* NO CHECK was made to see if the client */
				/* had died before going on... */
    if ( player[playerno].clientmode == AVAILABLE ||
	 player[playerno].fd == -1 ) {
				/* Player has died! Ignore rest of line */
	break;
    }

    rest = cp+1;		/* Move along to the next bit (if any) */
  }

  if (*rest != '\0')		/* There must be nothing after last '\n' */
    (void) printf ("Protocol breach: \"%s\".\n", rest);
}

/*-------------------------------------------------------------------------*/

void handletext (playerno)
  int playerno;
{
  char *line;

  if ( ( line = saferead (player[playerno].fd) ) != 0 )
    switch (player[playerno].clientmode) {

    case AWAITINGNAME:		/* Client connected but name not yet sent */
      if (verify (playerno, line) == -1)
	tidyup (playerno);
      break;

    case AWAITINGSIGN:		/* Login information has been recieved */
    case AWAITINGINFO:		/* Login information has been recieved */
      if (signon (playerno, line) != -1)
	no_users++;		/* If signon is successful, note the fact */
      else			/* Otherwise, tidy up the phlegm. */
	tidyup (playerno);	/* This will be silent, but deadly :-) */
      break;

    case PLAYINGGAME:		/* A command-line has been recieved */
    case ASLEEP:		/* from an active client */
      handleline (playerno, line); /* Parse and handle that command-line */
      break;

				/* "This can never happen" */
    default:			/*  -- famous last words */
      (void) puts ("Recieved message from client in undefined mode.");
      break;
    }

  else				/* Player has frogged out */
    visikilloff (playerno);
}

/***************************************************************************\
|*									   *|
|*  The previous four functions work as follows: When the daemon selects   *|
|*  on the file-descriptor of a client, it invokes handletext(), which	   *|
|*  switches on the mode of the client in question.  It then invokes a	   *|
|*  function to handle whatever stage of the login process is		   *|
|*  appropriate, or, if the client is already fully logged-in, it calls	   *|
|*  handleline().  This appallingly badly named function in fact breaks	   *|
|*  up the string passd to it into a number of lines, breaking at the	   *|
|*  '\n's, and passes the results, one by one, to handle().  This looks	   *|
|*  after passing on the command-line to anyone who happens to be	   *|
|*  snooping on the player whose command-lines are being handled, then	   *|
|*  checks to see if the client has sent any pseudosignals, which are	   *|
|*  acted on as appropriate.  If none have been sent, then the		   *|
|*  command-line is passed to commands() which checks it agianst each of   *|
|*  the hard-wired fundamental commands.  After this, it will eventually   *|
|*  check against data-defined verbs from the ADL, but that can wait.	   *|
|*									   *|
\***************************************************************************/

/*--------------------------------------------------------------------------*/

void disable_cmd ( playerno, command, action )
    int playerno;
    char *command;
    char *action;
{
    int i = 0,
        found = 0,
        list = 0;
				/* No arguments shows disabled commands */
    if ( *command == '\0' ) {
	say ( "The following commands are disabled;\n" );
	list = 1;
    } else {
	if ( ! ARCHMODEP ( playerno ) ) {
	    say ( "Frog off - you aint no Arch-wiz type person.\n" );
	    return;
	}
    }

    while ( comms[i].command != NULL ) {
	if ( list ) {
	    if ( ! ENABLEDP ( i ) )
		say ( form ( "%s\n", comms[i].command ) );
	} else {
	    if ( samenstring ( command, comms[i].command ) ) {
		found = 1;
		break;
	    } 
	}
	i++;
    }

    if ( list )
	return;

    if ( ! found ) {
	say ( form ( "The command \"%s\" does not exist.\n", command ) );
	return;
    }

    if ( samenstring ( action, "disable" ) ) {
	if ( ! CANDISABLEP ( i ) ) {
	    say ( form ( "The %s command cannot be disabled.\n", command ) );
	    return;
	}
	if ( ! ENABLEDP ( i ) ) {
	    say ( form ( "The %s command is already disabled.\n", command ) );
	    return;
	}
	comms[i].flags |= DISABLED;
    } else {
	if ( ENABLEDP ( i ) ) {
	    say ( form ( "The %s command is already enabled.\n", command ) );
	    return;
	}
	comms[i].flags &= ~DISABLED;
    }

    shout ( form ( "*** The %s command has been %sd by %s. ***\n",
		  command, action, player[playerno].name ),
	   NO_ONE, MESG_WARN );
}

