/***************************************************************************\
|*  									   *|
|*  Initialise.c:   Part of the Mundi-daemon - read info from files.	   *|
|*                  This version started on or about 8th June 1988.	   *|
|*  									   *|
\***************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>

#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "initialise.h"
#include "schedule.h"
#include "events.h"

/*--------------------------------------------------------------------------*/

char *malloc ();
/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */

/*--------------------------------------------------------------------------*/

int no_directions;
int no_locations;
int no_sayings;
int no_shouts;

/*--------------------------------------------------------------------------*/

int anim_list[MAXOBJECTS];
int num_anim = 0;

struct direction dir[MAXDIRECTIONS];
struct location loc[MAXLOCATIONS];
struct object obj[MAXOBJECTS];
struct howstruct how[MAXHOWS];

int times[DAYSINWEEK][HOURSINDAY];
char days[DAYSINWEEK][MAXDAYLEN];

char sayings[MAXSAYINGS][LINELEN];
char shouts[MAXSHOUTS][LINELEN];

/*--------------------------------------------------------------------------*/

int getnumber (fp)
  FILE *fp;
{
  int retval = 0;
  (void) fscanf (fp, "%d", &retval);
  return (retval);
}

/*--------------------------------------------------------------------------*/

void getdescription (description, fp)
  char description[DESCLEN];
  FILE *fp;
{
  char line[LINELEN];

  *description = '\0';
  (void) fgets (line, LINELEN, fp);
  while (*line != '@') {
    if (*line != COMMENT_CHAR)
      (void) strcat (description, line);
    (void) fgets (line, LINELEN, fp);
  }
}

/*--------------------------------------------------------------------------*/

void getline (name, fp)
  char *name;
  FILE *fp;
{
  *name = COMMENT_CHAR;
  while ((*name == COMMENT_CHAR) && (!feof (fp)))
    (void) fgets (name, LINELEN, fp);
  name[strlen (name) -1] = '\0'; /* Kill unwanted "\n" character */
}

/*--------------------------------------------------------------------------*/

int getmeaning (meaning, fp)
  char *meaning;
  FILE *fp;
{
  char ch;
  int ind = 0;

  while (isalpha (ch = getc (fp)))
    meaning[ind++] = ch;
  meaning[ind] = '\0';
  return (ch == ' ');
}

/*--------------------------------------------------------------------------*/

int getexits (exits, fp)
  exitlist exits;
  FILE *fp;
{
  int i;

  for (i = 0; i < no_directions; i++)
    exits[i] = getnumber (fp);

  return (i == no_directions);
}

/*--------------------------------------------------------------------------*/

int getdirections ()
{
  char temp[LINELEN];
  char filename[LINELEN];
  FILE *fp;
  int i = 0;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  DIRECTIONS_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  getline (temp, fp);
  while ((!feof (fp)) && (i < MAXDIRECTIONS)) {
    (void) strcpy (dir[i].desc, temp);
    *temp = tolower (*temp);
    (void) strcpy (dir[i].name, temp);
    temp[1] = '\0';
    (void) strcpy (dir[i].code, temp);
    if (++i == MAXDIRECTIONS)
	die(18);
    getline (temp, fp);
  }

  fclose ( fp );
  return (i);
}

/*--------------------------------------------------------------------------*/

int getlocations ()
{
  char temp[DESCLEN];
  char filename[LINELEN];
  FILE *fp;
  int i = 0;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  LOCATIONS_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  while ((!feof (fp)) && (i < MAXLOCATIONS)) {
    getdescription (temp, fp);
    if ((loc[i].desc = malloc ((unsigned) strlen (temp)+1)) == NULL)
      die (12);
    (void) strcpy (loc[i].desc, temp);
    getline (temp, fp);
    (void) strcpy (loc[i].name, temp);
    getline (temp, fp);
    (void) strcpy (loc[i].code, temp);
    if ( getexits (loc[i].exits, fp) )
	loc[i].flags = getnumber (fp);
    
    (void) fscanf (fp, "%*s%*c");

    if (++i == MAXLOCATIONS)
	die (18);
  }


  fclose ( fp );
  return (i);
}

/*--------------------------------------------------------------------------*/

int getobjects ()
{
  char temp[LINELEN];
  char filename[LINELEN];
  FILE *fp;
  int i = 0, j, eoln;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  OBJECTS_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  while ((!feof (fp)) && (i < MAXOBJECTS)) {
    getline (temp, fp);
    (void) strcpy (obj[i].desc, temp);

    j = 0;
    do {
      eoln = !getmeaning (temp, fp);
      (void) strcpy (obj[i].name.wordv[j], temp);
      j++;
    } while (!eoln);
    obj[i].name.wordc = j;

    obj[i].location  = getnumber (fp);
    obj[i].value     = getnumber (fp);
    obj[i].weight    = getnumber (fp);
    obj[i].size      = getnumber (fp);
    obj[i].flag      = getnumber (fp);
    obj[i].container = getnumber (fp);
    obj[i].seen      = getnumber (fp);
    obj[i].damage    = getnumber (fp);
    obj[i].key	     = getnumber (fp);
    obj[i].speed     = getnumber (fp);
    obj[i].actions   = getnumber (fp);
    obj[i].how       = getnumber (fp); 


    (void) fscanf (fp, "%*s%*c");

    if (++i == MAXOBJECTS)
	die (18);
  }


  fclose ( fp );
  return (i);
}

/*--------------------------------------------------------------------------*/

void gettimes ()
{
  char temp[LINELEN];
  char filename[LINELEN];
  FILE *fp;
  int i = 0, j, eoln;
  char *ptr;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  TIMES_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  getline (temp, fp);		/* Hours! Ignore */

  while ((!feof (fp)) && (i < DAYSINWEEK)) {
      getline ( temp, fp );	/* Blank line */
      getline ( temp, fp );	/* Times of opening */

      strncpy ( days[i], temp, MAXDAYLEN-1 );

      for ( j = 0; j < HOURSINDAY; j++ ) {
	  switch ( temp[(j*3)+MAXDAYLEN] ) {
	  case '+':
	      times[i][j] = 1;
	      break;
	  case 'X':
	      times[i][j] = 2;
	      break;
	  default:
	      times[i][j] = 0;
	      break;
	  }
      }
      i++;
  }
  
  fclose ( fp );		/* 2001: moved this before the condition */
  if ( mundi_open() ) {
      int day, hour, secs;

      hour = next_close_time ( &day );
				/* 1/2 Hour before close */
      if ( ( secs = (compute_seconds_till ( day, hour ) - WARN1_TIME) ) < 0 )
	  return;		/* 2001: was "secs = 0" */
      logerror ( "Be closing in %d seconds (%s %d:00)\n",
		secs, days[day], hour );
      add_event ( secs, 0, ev_close_mundi, "" );
  } else {
      int day, hour, secs;

      hour = next_open_time ( &day );
      if ( ( secs = compute_seconds_till ( day, hour ) ) < 0 )
	  secs = 0;
      logerror ( "Be opening in %d seconds (%s %d:00)\n",
		secs, days[day], hour );
      add_event ( secs, 0, ev_open_mundi, "" );
  }
}

/*--------------------------------------------------------------------------*/
				/* Is mundi currently open */
int mundi_open ()
{
    return ( times[get_day()][get_hour ()] );
}
	
/*--------------------------------------------------------------------------*/
				/* Return next time of opening */
int next_open_time ( day )
    int *day;
{
    int d = get_day (), i = d,
        h = get_hour (), j = h;

    while ( 1 ) {
	if ( ++j == HOURSINDAY ) {
	    j = 0;
	    if ( ++i == DAYSINWEEK )
		i = 0;
	}
	if ( i == d && j == h ) {
	    j = -1;
	    break;
	}
	if ( times[i][j] )
	    break;
    }   
    if ( day )
	*day = i;
    return ( j );		/* Return hour */
}
/*--------------------------------------------------------------------------*/
				/* Return next time of closing */
int next_close_time ( day )
    int *day;
{
    int d = get_day (), i = d,
        h = get_hour (), j = h;

    while ( 1 ) {
	if ( ++j == HOURSINDAY ) {
	    j = 0;
	    if ( ++i == DAYSINWEEK )
		i = 0;
	}
	if ( i == d && j == h ) {
	    j == -1;
	    break;
	}
	if ( times[i][j] == 0 )
	    break;
    }   
    if ( day )
	*day = i;
    return ( j );		/* Return hour */
}
/*--------------------------------------------------------------------------*/

int getsayings ()
{
  char filename[LINELEN];
  FILE *fp;
  int i = 0, j, eoln;
  char *ptr;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  SAYINGS_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  while ( !feof ( fp ) & i < MAXSAYINGS )
      getline (sayings[i++], fp);
      
  fclose ( fp );
  return ( i );
}

/*--------------------------------------------------------------------------*/

int getshouts ()
{
  char filename[LINELEN];
  FILE *fp;
  int i = 0, j, eoln;
  char *ptr;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  SHOUTS_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  while ( !feof ( fp ) & i < MAXSHOUTS )
      getline (shouts[i++], fp);
      
  fclose ( fp );
  return ( i );
}

/*--------------------------------------------------------------------------*/

int gethows ()
{
  char temp[LINELEN];
  char filename[LINELEN];
  FILE *fp;
  int i = 0;

  (void) sprintf (filename, "%s/%s/%s", MUNDI_DIR, INITIALISE_DIR,
		  HOWS_FILE);
  if ((fp = fopen (filename, "r")) == NULL)
    die (17);

  while ((!feof (fp)) && (i < MAXHOWS)) {
    how[i].whatto = getnumber (fp);
    getline (temp, fp);
    (void) strcpy (how[i].name, temp);
    if (++i == MAXHOWS)
	die(18);
  }

  fclose ( fp );
  return (i);
}

/*--------------------------------------------------------------------------*/

void initialise ()
{
  no_directions = getdirections ();
  no_locations  = getlocations  ();
  no_objects    = getobjects    ();
  no_personas   = getpersonas   (); /* Code for this is in persona.c */
  no_sayings    = getsayings    ();
  no_shouts     = getshouts     ();
  no_hows       = gethows       ();
  gettimes ();
}

/*--------------------------------------------------------------------------*/

void write_object ( fp, item )
    FILE *fp;
    int item;
{

    int i;
    
    fprintf ( fp, "#\n# Object #%d\n%s\n", item, obj[item].desc );
    for ( i = 0; i < obj[item].name.wordc; i++ ) {
	fputs ( obj[item].name.wordv[i], fp );
	fputs ( " ", fp );
    }
    fputs ( "\n", fp );
    fprintf ( fp, "%s_%s %d %d %d %d %d %d %d %d\n",
	     exit_name_w(obj[item].location),
	     obj[item].value,
	     obj[item].weight,
	     obj[item].size,
	     obj[item].flag,
	     obj[item].container,
	     obj[item].seen,
	     obj[item].damage,
	     obj[item].key );
}

/*--------------------------------------------------------------------------*/

void write_room ( fp, room )
    FILE *fp;
    int room;
{
    int i;
    fprintf ( fp, "#\n# Location #%d\n#\n%s%s", room, loc[room].desc );
    fprintf ( fp, "@\n%s\n%s\n", loc[room].name, loc[room].code );
    for ( i = 0; i < no_directions; i++ )
	fprintf ( fp, "%s_%s ", exit_name_w(loc[room].exits[i]) );
    fputs ( "\n", fp );
}




