/***************************************************************************\
|*                                                                         *|
|*  Look.c:	Module to print location descriptions, inventory etc.	   *|
|*		This version started about 8th June 1988.		   *|
|*                                                                         *|
\***************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>

#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "look.h"
#include "utils.h"

/*-------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */

/*-------------------------------------------------------------------------*/

void describe_text (playerno, here)
  int playerno;
  int here;
{
  char outmessage[LINELEN];

  if (player[playerno].displaymode & PRNUMBERS) {
    (void) sprintf (outmessage, "(Location number %d, %s)\n", here,
		    loc[here].code);
    say (outmessage);
  }

  if (player[playerno].displaymode & PRVERBOSE)
    say (loc[here].desc);
  else {
    (void) sprintf (outmessage, "%s.\n", loc[here].name);
    say (outmessage);
  }
}

/*-------------------------------------------------------------------------*/

int nonempty (here)
  int here;
{
  int i, retval=0;

  for (i = 0; i < no_objects; i++)
    if (obj[i].location == here) {
      retval = 1;
      break;
    }
  return (retval);
}

/*--------------------------------------------------------------------------*/

void printcontents (playerno, ind, level)
    int playerno;
    int ind;
    int level;
{
    int i;
    char outmessage[LINELEN];
				/* Gotta see inside the thing... */
    if ( ! transparentp (ind) && ! openp(ind) && ! surfacep(ind) )
	return;

    (void) sprintf (outmessage, "%*s", level*INDENT, "" );

    if (!nonempty (-ind)) {
	if (openp(ind) && !surfacep(ind))
	    (void) strcat (outmessage, "It is open.\n");
	else 	  
	    (void) strcat (outmessage, ".\n");
	say (outmessage);
	return;
    }

    if (surfacep(ind))
	(void) strcat (outmessage, "On it you can see:\n");
    else
	if (openp(ind))
	    if (openablep(ind))
		(void) strcat (outmessage, 
			       form ( "It is %sopen and contains:\n",
				     lockedp(ind) ? "locked " : "" ) );
	    else
		(void) strcat (outmessage, "It contains:\n");
				/* It must be transparent! */
	else
	    (void) strcat (outmessage, 
			   form ( "It is %sclosed and contains:\n",
				 lockedp(ind) ? "locked " : "" ) );
    say (outmessage);
    for (i = 0; i < no_objects; i++)
	if (obj[i].location == -ind) {
	    printobject (playerno, i, level+1);
	}
}

/*--------------------------------------------------------------------------*/

void printobject (playerno, ind, level)
    int playerno;
    int ind;
    int level;
{
    int i;
    char outmessage[DESCLEN];

    (void) sprintf (outmessage, "%*sA %s", level*INDENT, "", obj[ind].desc);
    
    if (! transparentp (ind) && ! openp(ind) && ! surfacep(ind) ) {
	(void) strcat ( outmessage, ".\n" );
	say ( outmessage );
	return;
    }

    if (!nonempty (-ind)) {
	if (openp(ind) && !surfacep(ind))
	    (void) strcat (outmessage, " which is open.\n");
	else 	  
	    (void) strcat (outmessage, ".\n");
	say (outmessage);
	return;
    }

    if (surfacep(ind))
	(void) strcat (outmessage, " on which you can see:\n");
    else
	if (openp(ind))
	    if (openablep(ind))
		(void) strcat (outmessage, 
			       form (" which is %sopen and contains:\n",
				     lockedp(ind) ? "locked " : "" ) );
	    else
		(void) strcat (outmessage, " which contains:\n");
	else
	    (void) strcat (outmessage, 
			   form ( " which is %sclosed and contains:\n",
				 lockedp(ind) ? "locked " : "" ) );
    say (outmessage);
    for (i = 0; i < no_objects; i++)
	if (obj[i].location == -ind) {
	    printobject (playerno, i, level+1);
	}
}

/*-------------------------------------------------------------------------*/

void describe_objects (playerno, here)
  int playerno;
  int here;
{
  int ind;

  if ((player[playerno].displaymode & PROBJECTS) && (nonempty (here))) {
    say ("You can also see:\n");
    for (ind = 0; ind < no_objects; ind++)
      if (obj[ind].location == here)
        printobject (playerno, ind, 1);
  }
}

/*-------------------------------------------------------------------------*/

int unique (array, here, ind)
  int *array;
  int here;
  int ind;
{
  int i, result = 1;

  for (i = 0; i < ind; i++)
    if (array[i] == loc[here].exits[ind]) {
      result = 0;
      i = ind;
    }
  return (result);
}

/***************************************************************************\
|*                                                                         *|
|*  NB.	In the following function, the loops all start at 1 instead of 0.  *|
|*	This is done in order to prevent mundi from listing "pit" as an	   *|
|*	obvious exit.  Witty, is it not?  It is.			   *|
|*									   *|
\***************************************************************************/

void describe_exits (playerno, here)
  int playerno;
  int here;
{
  int i;
  int no_exits = 0;
  int no_sofar = 0;
  int dest[MAXDIRECTIONS];
  char outmessage [LINELEN];

  if (player[playerno].displaymode & PREXITS) {
    for (i = 1; i < no_directions; i++)
      if (   ((ARCHMODEP(playerno) && (loc[here].exits[i] != 0)) ||
	      ((loc[here].exits[i] > 0) ||
	       (loc[here].exits[i] == -1)))
	  && ((player[playerno].displaymode & PRALLEXITS) ||
	      (unique (dest, here, i)))) {
	no_exits++;
	dest[i] = loc[here].exits[i];
      }
      else
        dest[i] = 0;

    if (no_exits != 0) {
      if (no_exits == 1)
        (void) strcpy (outmessage, "The only obvious exit leads ");
      else
        (void) strcpy (outmessage, "Obvious exits lead ");
      for (i = 1; i < no_directions; i++)
        if ( dest[i] ) {
          no_sofar++; 
          if (no_sofar == no_exits && no_exits > 1)
            (void) strcat (outmessage, "\b\b and ");
          (void) strcat (outmessage, dir[i].desc);
	  (void) strcat (outmessage, ", ");
        }
      (void) strcat (outmessage, "\b\b. \n");
    }
    else
      (void) strcpy (outmessage, "There seems to be no way out of here.\n");

    say (outmessage);
  }
}

/*-------------------------------------------------------------------------*/

void describe_people (playerno, here)
  int playerno;
  int here;
{
  int i;
  int no_people = 0;
  int no_sofar = 0;
  char outmessage[DESCLEN];

  *outmessage = '\0';
  for (i = 0; i < MAXUSERS; i++)
    if ((player[i].fd != -1) && (player[i].where == here) &&
	(i != playerno)	&& (ISACTIVE (i)))
      no_people++;

  if (no_people) {
    for (i = 0; i < MAXUSERS; i++)
      if ((player[i].fd != -1) && (player[i].where == here) &&
	  (i != playerno) && (ISACTIVE (i))) {
	no_sofar++;
	if ((no_people == 1) && (no_sofar == no_people)) {
	  (void) strcat (outmessage, player[i].name);
	  (void) strcat (outmessage, " is here.\n");
	  i = MAXUSERS;
	}
	else
	  if (no_sofar == no_people) {
	    (void) strcat (outmessage, "\b\b and ");
	    (void) strcat (outmessage, player[i].name);
	    (void) strcat (outmessage, " are here.\n");
	    i = MAXUSERS;
	  }
	  else {
	    (void) strcat (outmessage, player[i].name);
	    (void) strcat (outmessage, ", ");
	  }
      }
    say (outmessage);
  }
}

/*-------------------------------------------------------------------------*/

void describe (playerno, here)
  int playerno;
  int here;
{
  describe_text (playerno, here);
  describe_objects (playerno, here);
  describe_exits (playerno, here);
  describe_people (playerno, here);
  say ("\n");
}

/*-------------------------------------------------------------------------*/

void moveto (playerno, newloc)
{
    if ( newloc < -1 )		/* hidden exit */
	newloc = abs (newloc);
				/* -1 implies location zero */
    if ( newloc == -1 )
	newloc = 0;

    player[playerno].where = newloc;
    say (form ("%s%s\n", CHANGE_LOCATION, loc[newloc].code));
    describe (playerno, newloc);
}

/*-------------------------------------------------------------------------*/

int generalhere (item, mask, location)
  int item;
{
  if (obj[item].location == location)
    return (1);
  else
    if ((obj[item].location < 0) &&
	(((obj[-obj[item].location].container) & mask) != 0))
      return (generalhere (-obj[item].location, mask, location));
  return(0);
}

/*--------------------------------------------------------------------------*/
				/* Can item get out to location */
int accessout ( item, where )
    int item;
    int where;
{
				/* Everything has to be open! */
    return ( generalhere ( item, 10, where ) );
}

/*-------------------------------------------------------------------------*/

int reallyheld (item, playerno)
  int item;
  int playerno;
{
  return (generalhere (item, 15, player[playerno].invloc));
}

/*-------------------------------------------------------------------------*/

int reallyhere (item, location)
  int item;
  int location;
{
  return (generalhere (item, 15, location));
}

/*-------------------------------------------------------------------------*/

int accessheld (item, playerno)
  int item;
  int playerno;
{
  return (generalhere (item, 10, player[playerno].invloc));
}

/*-------------------------------------------------------------------------*/

int accesshere (item, playerno)
  int item;
  int playerno;
{
  return (generalhere (item, 10, player[playerno].where));
}

/*-------------------------------------------------------------------------*/

int visiheld (item, playerno)
  int item;
  int playerno;
{
  return (generalhere (item, 11, player[playerno].invloc));
}

/*-------------------------------------------------------------------------*/

int visihere (item, playerno)
  int item;
  int playerno;
{
  return (generalhere (item, 11, player[playerno].where));
}

/*-------------------------------------------------------------------------*/

int accessible (item, playerno)
  int item;
  int playerno;
{
  return (accesshere (item, playerno) || accessheld (item, playerno));
}

/*-------------------------------------------------------------------------*/

int visible (item, playerno)
  int item;
  int playerno;
{
  return (visihere (item, playerno) || visiheld (item, playerno));
}

/*-------------------------------------------------------------------------*/

int totalweight (item)
  int item;
{
  int i;
  int retval = 0;

  retval = obj[item].weight;
  if ((obj[item].container) != 0)
    for (i = 0; i < no_objects; i++)
      if (obj[i].location == -item)
        retval += totalweight (i);
  return (retval);
}

/*--------------------------------------------------------------------------*/

int totalintweight (item)
  int item;
{
  int i;
  int retval = 0;

  if ( explodep(item) )
      retval = 1000000;
  else
      retval = obj[item].weight;

  if ((obj[item].container) != 0)
    for (i = 0; i < no_objects; i++)
      if (obj[i].location == -item)
        retval += totalintweight (i);
  return (retval);
}

/*--------------------------------------------------------------------------*/

int totallocweight (loc)
  int loc;
{
  int i;
  int retval = 0;

  for ( i = 0; i < no_objects; i++ )
      if ( obj[i].location == loc )
	  retval += totalweight(i);
  return(retval);
}

/*-------------------------------------------------------------------------*/

void verbose (playerno)
  int playerno;
{
  player[playerno].displaymode |= PRVERBOSE;
  say ("Maximum verbosity.\n");
}

/*-------------------------------------------------------------------------*/

void brief (playerno)
  int playerno;
{
  player[playerno].displaymode &= ~PRVERBOSE;
  say ("Brief descriptions.\n");
}

/*-------------------------------------------------------------------------*/

void exits (playerno)
  int playerno;
{
  player[playerno].displaymode |= PREXITS;
  say ("Exits will be explicitly listed.\n");
}

/*-------------------------------------------------------------------------*/

void noexits (playerno)
  int playerno;
{
  player[playerno].displaymode &= ~PREXITS;
  say ("Exits will not be explicitly listed.\n");
}

/*-------------------------------------------------------------------------*/

void objects (playerno)
  int playerno;
{
  player[playerno].displaymode |= PROBJECTS;
  say ("Objects will be listed with room descriptions.\n");
}

/*-------------------------------------------------------------------------*/

void noobjects (playerno)
  int playerno;
{
  player[playerno].displaymode &= ~PROBJECTS;
  say ("Objects will not be listed with descriptions.\n");
}

/*-------------------------------------------------------------------------*/

void allexits (playerno)
  int playerno;
{
  player[playerno].displaymode |= PRALLEXITS;
  say ("All exits will be displayed when any are.\n");
}

/*-------------------------------------------------------------------------*/

void uniqueexits (playerno)
  int playerno;
{
  player[playerno].displaymode &= ~PRALLEXITS;
  say ("Only one exit will be listed leading to any given destination.\n");
}

/*--------------------------------------------------------------------------*/

void ignore ( playerno, command )
    int playerno;
    char *command;
{
    if ( samestring ( "shouts", command ) ) {
	player[playerno].ignorestuff |= IGNORESHOUT;
	say ( "Shouted messages will now be ignored.\n" );
	return;
    }
    if ( samestring ( "global", command ) ) {
	player[playerno].ignorestuff |= IGNOREGLOBAL;
	say ( "Global messages will now be ignored.\n" );
	return;
    }
    if ( samestring ( "all", command ) ) {
	player[playerno].ignorestuff |= IGNOREALL;
	say ( "All messages will now be ignored.\n" );
	return;
    }
    say ( "usage: ignore global | shouts | all\n" );
}
	
/*--------------------------------------------------------------------------*/

void notice ( playerno, command )
    int playerno;
    char *command;
{
    if ( samestring ( "shouts", command ) ) {
	player[playerno].ignorestuff &= ~IGNORESHOUT;
	say ( "Shouted messages will now be noticed.\n" );
	return;
    }
    if ( samestring ( "global", command ) ) {
	player[playerno].ignorestuff &= ~IGNOREGLOBAL;
	say ( "Global messages will now be noticed.\n" );
	return;
    }
    if ( samestring ( "all", command ) ) {
	player[playerno].ignorestuff &= ~IGNOREALL;
	say ( "All messages will now be noticed.\n" );
	return;
    }
    say ( "usage: notice global | shouts | all\n" );
}

/*-------------------------------------------------------------------------*/

void numbers (playerno)
  int playerno;
{
  if (!WIZMODEP (playerno))
    say ("Only wizards are allowed to see location numbers, so frog off!\n");
  else {
    player[playerno].displaymode |= PRNUMBERS;
    say ("Location numbers will be diplayed with descriptions.\n");
  }
}

/*-------------------------------------------------------------------------*/

void nonumbers (playerno)
  int playerno;
{
  player[playerno].displaymode &= ~PRNUMBERS;
  say ("Location numbers will not be diplayed with descriptions.\n");
}

/*-------------------------------------------------------------------------*/

int totalvalue (here)
  int here;
{
  int i, total = 0;

  for (i = 0; i < no_objects; i++)
    if ((here < 0) || (reallyhere (i, here)))
      total += obj[i].value;

  return (total);
}

/*-------------------------------------------------------------------------*/

void printscore (playerno)
  int playerno;
{
  char outmessage[DESCLEN];

  (void) sprintf (outmessage,
		  "In %d move%s, you have scored %d point%s this game.\n",
		  player[playerno].newmoves,
		  PLURAL (player[playerno].newmoves),
		  player[playerno].newscore,
		  PLURAL (player[playerno].newscore));
  say (outmessage);

  (void) sprintf (outmessage, "This gives you a total of %d point%s.\n",
		  player[playerno].newscore+player[playerno].oldscore,
		  PLURAL(player[playerno].newscore+player[playerno].oldscore)
		  );
  say (outmessage);
}

/*--------------------------------------------------------------------------*/

void inventory (playerno)
  int playerno;
{
  int i, gotone=0;

  for (i = 0; i < no_objects; i++)
    if (obj[i].location == player[playerno].invloc) {
      if (!gotone)
	say ("You are carrying:\n");
      printobject (playerno, i, 1);
      gotone = 1;
    }
  if (!gotone)
    say ("You are not carrying anything.\n");
}

/*--------------------------------------------------------------------------*/

char *person_sd (person)	/* Person short-description */
  struct persona person;	/* Useable on personas or players */
{
  static char buffer[DESCLEN];	/* DONT USE FORM PLEASE! */
  int pu_level = 0,
      fa_level = 0,
      sc_level = person.level;
				/* Get correct levels */
  pu_level = compute_level ( pu_levels, person.purity, PU_LEVELS );
  fa_level = compute_level ( fa_levels, person.intcont, FA_LEVELS );

  (void) sprintf (buffer, "%s the %s, %s %s.\n",
		  person.name,
		  pu_levels[pu_level].name,
		  fa_levels[fa_level].name,
		  sc_levels[sc_level].name);
  return (buffer);
}

/*-------------------------------------------------------------------------*/
				/* Location/inventory where item lives */
int real_loc (item)
  int item;
{
  if (obj[item].location >= 0)
    return (obj[item].location);
  else
    return (real_loc (-obj[item].location));
}

/*--------------------------------------------------------------------------*/
				/* The item which contains item */
int base_item (item)
  int item;
{
				/* If not contained - then itself */
  if (obj[item].location >= 0)
    return (item);
  else
    return (base_item (-obj[item].location));
}

/*-------------------------------------------------------------------------*/
				/* Which room is item in */
int where_item ( item )
    int item;
{
    int where = whoholdsitem ( item );

    if ( where == -1 )
	where = real_loc ( item );
    else
	where = player[where].where;

    return ( where );
}
