/***************************************************************************\
|*                                                                         *|
|*  Magic.c:	Part of the server program for MUNDI			   *|
|*		This module will handle all the routines for sundry	   *|
|*		types of magic -- both wiz-commands, (teleport etc.)	   *|
|*		and common-or-garden spells, (eg. where)		   *|
|*  Started:	Sat Jun 17 00:10:37 BST 1989				   *|
|*                                                                         *|
\***************************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <strings.h>

#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "handle.h"
#include "look.h"
#include "utils.h"
#include "initialise.h"
#include "verbs.h"
#include "schedule.h"
#include "events.h"

/*-------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *strdup ();
extern char *realloc ();

/*-------------------------------------------------------------------------*/
				/* Can playerno dominate victim as wiz? */
int wizdominatep ( playerno, victim )
    int playerno;
    int victim;
{
    if (  (player[playerno].wizstuff & (WIZMODE|ARCHMODE|METAMODE))
	> (player[victim].wizstuff & (WIZMODE|ARCHMODE|METAMODE)) )
	return ( 1 );
    else
	return ( player[playerno].level > player[victim].level );
}

/*--------------------------------------------------------------------------*/
				/* Return TRUE if castable and */
				/* deducts points */
				/* If FALSE does nothing */
int castablep ( playerno, cost )
    int playerno;
    int cost;
{
    if ( (player[playerno].oldscore + player[playerno].newscore) < cost )
	return ( 0 );
    scorepoints ( playerno, -cost, 1, 2 );
    return ( 1 );
}

/*--------------------------------------------------------------------------*/
				/* Return playerno of holder of item,
				   -1 if not held */
int whoholdsitem ( item )
    int item;
{
    int ploc = real_loc ( item );
				/* In a room */
    if ( ploc < MAXLOCATIONS )
	return ( -1 );

    ploc -= MAXLOCATIONS;
				/* Its not in the game! */
    if ( ploc >= MAXUSERS )
	return ( -1 );
				/* Player isn't actually on game */
    if ( player[ploc].fd == -1 )
	return ( -1 );

    return ( ploc );		/* Held */
}

/*--------------------------------------------------------------------------*/

void gowiz (playerno, command)
  int playerno;
  char *command;
{
  if ((WIZMODEP (playerno)) && (!ARCHMODEP (playerno)))
    say ("You're *already* in wiz mode, you burma!\n");
  else
    if (!CANWIZP (playerno))
      say ("Uh-huh.  No-no.  Wizmode is a trifle restricted.\n");
    else {
      say ("### OK, you are now in wiz mode.  Be careful!\n");
      if ( ! samenstring ( command, "sneakily" ) )
	  shout (form ("### %s is now in wiz mode!  Beware!\n",
		   player[playerno].name), playerno, MESG_GLOBAL);
      player[playerno].wizstuff |= WIZMODE;
    }
}

/*--------------------------------------------------------------------------*/

void nowiz (playerno, command)
  int playerno;
  char *command;
{
  if (!WIZMODEP (playerno))
    say ("You're not in wiz mode *now*, you thinging thing!!\n");
  else {
    say ("### OK, you are no longer in wiz mode.\n");
    if ( ! samenstring ( command, "sneakily" ) )
	shout (form ("### %s has left wiz mode ...  Relax!\n",
		     player[playerno].name), playerno, MESG_GLOBAL);
    player[playerno].wizstuff &= ~WIZMODE;
  }
}

/*-------------------------------------------------------------------------*/

void goarch (playerno, command)
  int playerno;
  char *command;
{
  if ((ARCHMODEP (playerno)) && (!METAMODEP (playerno)))
    say ("You're *already* in arch mode, you burma!\n");
  else
    if (!CANARCHP (playerno))
      say ("Your weedy persona is not powerful enough to enter archmode.\n");
    else {
      say ("### OK, you are now in arch mode!\n");
      if ( *command == '\0' )
	  shout (form ("### %s has entered arch mode!\n",
		       player[playerno].name), playerno, MESG_GLOBAL);
      player[playerno].wizstuff |= ARCHMODE;
    }
}

/*--------------------------------------------------------------------------*/

void noarch (playerno, command)
  int playerno;
  char *command;
{
  if (!ARCHMODEP (playerno))
    say ("You're not in arch mode *now*, chicken-head!\n");
  else {
    say ("### OK, you are no longer in arch mode.\n");
    if ( *command == '\0' )
	shout (form ("### %s has left arch mode.\n", player[playerno].name),
	       playerno, MESG_GLOBAL);
    player[playerno].wizstuff &= ~ARCHMODE;
  }
}

/*-------------------------------------------------------------------------*/

void gometa (playerno, command)
  int playerno;
  char *command;
{
  if (METAMODEP (playerno))
    say ("You're *already* in meta mode, tea-head!\n");
  else
    if (!CANMETAP(playerno))
      say ("Nope.  \"There can be only some.\"\n");
    else {
      say ("### OK, you are now in meta mode.  Be careful!\n");
      if ( *command == '\0' )
	  shout (form ("### %s has entered meta-wiz-mode!\n",
		       player[playerno].name), playerno, MESG_WARN);
      player[playerno].wizstuff |= METAMODE;
    }
}

/*--------------------------------------------------------------------------*/

void nometa (playerno, command)
  int playerno;
  char *command;
{
  if (!METAMODEP (playerno))
    say ("You're not in metamode.  You probably don't even know what it \
is.\n");
  else {
    say ("### OK, you are no longer in meta mode.\n");
    if ( *command == '\0' )
	shout (form ("### %s has left meta mode!\n", player[playerno].name),
	       playerno, MESG_WARN);
    player[playerno].wizstuff &= ~METAMODE;
  }
}

/*-------------------------------------------------------------------------*/

void teleport (playerno, place)
  int playerno;
  char *place;
{
    int i;
    int newloc;
    int cost;

    if (!WIZMODEP (playerno)) {
	say ("Utterly no way you can teleport, mortal p(h)easant!\n");
	return;
    }

    newloc = strtoroomcode ( place );

    if ( newloc > 0 ) 
	if ( !ARCHMODEP (playerno) && loc_noteletop(newloc) ) {
	    say ( "You can't teleport to there!\n" );
	    return;
	}
    if ( !ARCHMODEP (playerno) && loc_notelefrp(player[playerno].where) ) {
	say("You can't teleport from here!\n");
	return;
    }
    if (newloc == -1) {
	for (i = 0; i < MAXUSERS; i++) {
	    if ((player[i].fd != -1) && (samestring (player[i].name, place))) {
		newloc = player[i].where;
		if ( !ARCHMODEP (playerno) && loc_noteletop(newloc) ) {
		    say("You can't teleport to there!\n");
		    return;
		}
		if ( !ARCHMODEP (playerno) && loc_notelefrp(player[playerno].where) ) {
		    say("You can't teleport from here!\n");
		    return;
		}
		say ( form ( "(Going to join %s)\n",
			    player[i].name ) );
		break;
	    }
	}
    }

    if (newloc == -1) {
	say ( form ( "There is no location called %s.\n", place ) );
	return;
    }

    if ( !ARCHMODEP (playerno ) ) {
	cost = totallocweight(player[playerno].invloc) +
	       player[playerno].intcont + 100;
	if ( ! castablep ( playerno, cost ) ) {
	    say ( "Frog off you aint got enough points!\n" );
	    return;
	}
    }
    speak ( form ( "In a flash of light, %s disappeared!\n",
		      player[playerno].name ),
	   playerno, player[playerno].where );
    speak ( form ( "In a vast cloud of smoke, %s just appeared!\n",
		  player[playerno].name ),
	   playerno, newloc );
    say ("You teleport ...\n");
    moveto (playerno, newloc);
}

/*--------------------------------------------------------------------------*/

void summon (playerno, string)
  int playerno;
  char *string;
{
  int victim = -1, cost;
  char item[LINELEN];

  getnextword (item, string);

  if (!WIZMODEP (playerno)) {
      say ("Surely you jest?  You?!  A mere mortal?!!  Summon?!!!  Hahaha!\n");
      return;
  }
  if ((victim = strtoplayer (item)) == -1) {
      say ( form ( "There is no-one called \"%s\" in the game!\n",
		  item ) );
      return;
  }
  if ( !ARCHMODEP(playerno) && loc_safesummonp(player[victim].where) ) {
      say ( "Nope that location is safe from summoning.\n" );
      return;
  }
  if ( ! wizdominatep ( playerno, victim ) ) {
      say ( form ( "Sorry your not powerful enough to summon %s!\n",
		  player[victim].name ) );
      safewrite ( victim, form ( "*** %s tried to summon you.\n",
				player[playerno].name ) );
      return;
  }
  if ( !ARCHMODEP (playerno ) ) {
      cost = totallocweight(player[victim].invloc) +
	  player[victim].intcont + 100;
      if ( ! castablep ( playerno, cost ) ) {
	  say ( "Frog off you aint got enough points!\n" );
	  return;
      }
  }

  speak ( form ( "%s just vanished, as if by magic!\n",
		      player[victim].name ),
	 victim, player[victim].where);
  safewrite ( victim, form ( "\nYou have been magically summoned by %s!\n\n",
			    player[playerno].name ) );
  moveto (victim, player[playerno].where);

  speak ( form ( "%s has just appeared %s.\n",
		player[victim].name,
		( *string == '\0' ) ? "in a shower of cold vomit" : string ),
	 victim, player[victim].where );
}

/*--------------------------------------------------------------------------*/

void whereperson (playerno, ind, level)
  int playerno;
  int ind;
  int level;
{
  char temp[LINELEN];
  if ( ! rollagainst ( playerno, -90, player[playerno].intelligence ) )
      return;
  
  if ((player[ind].where < 0) || (player[ind].where >= no_locations))
    (void) sprintf (temp, "%s is in a very odd place indeed!\n",
		    player[ind].name);
  else {
    (void) sprintf (temp, "%s is at \"%s\"",
		    player[ind].name, loc[player[ind].where].name);
    if (WIZMODEP (playerno))
      (void) sprintf (temp+strlen (temp), " (%s)",
		      loc[player[ind].where].code);
    (void) strcat (temp, "\n");
  }

  say (form ("%*s%s", level*INDENT, "", temp));
}

/*--------------------------------------------------------------------------*/

void whereitem (playerno, ind, level)

  int playerno;
  int ind;
  int level;
{
  char temp[LINELEN];
  int where;

  if ( ! rollagainst ( playerno, -75, player[playerno].intelligence ) )
      return;

  if (obj[ind].location <0) {
    (void) sprintf (temp, "%*sThe %s is %cn the %s.\n",
		    level*INDENT, "", obj[ind].desc,
		    (((obj[-obj[ind].location].container) & 8)? 'o': 'i'),
		    obj[-obj[ind].location].desc);
    say (temp);
    whereitem (playerno, -obj[ind].location, level+1);
  }
  else
    if (obj[ind].location < MAXLOCATIONS) {
      (void) sprintf (temp, "%*sThe %s is at \"%s\"",
		      level*INDENT, "", obj[ind].desc,
		      loc[obj[ind].location].name);
      if (WIZMODEP (playerno))
	(void) sprintf (temp+strlen (temp), " (%s)",
			loc[obj[ind].location].code);
      (void) strcat (temp, "\n");
      say (temp);
    } else {
	if ( ( where = whoholdsitem ( ind ) ) != -1 ) {
	    (void) sprintf (temp, "%*sThe %s is held by %s.\n",
			    level*INDENT, "", obj[ind].desc,
			    player[where].name);
	    say (temp);
	    whereperson (playerno, obj[ind].location-MAXLOCATIONS, level+1);
	} else {
	    (void) sprintf (temp, "%*sFrog knows where the %s is!\n",
			level*INDENT, "", obj[ind].desc);
	    say (temp);
	}
    }
}

/*--------------------------------------------------------------------------*/

void wherestring (playerno, string)
  int playerno;
  char *string;
{
  int number = 0;
  int ind = -1;
  int gotone = 0;
  char item[LINELEN];
  char temp[LINELEN];

  while (*string != '\0') {
    number++;
    getnextobject (item, string);
    if (samestring ("treasure", item)) {
      for (ind = 0; ind < no_objects; ind++)
	if ((obj[ind].value) && (!reallyhere (ind, STORELOC)) &&
				(!reallyhere (ind, UPITLOC))) {
	  whereitem (playerno, ind, 0);
	  gotone = 1;
	}
      if (!gotone)
	say ("All the treasure has been pitted!  \"crash\" to restart.\n");
    }
    else
      if ((ind = strtoplayer (item)) != -1)
	whereperson (playerno, ind, 0);
      else
	if ((ind = strtonoun (item, player[playerno].where)) != -1)
	  whereitem (playerno, ind, 0);
	else {
	  (void) sprintf (temp, "I don't know what a \"%s\" is.\n", item);
	  say (temp);
	}
  }

  if (number == 0) {
    (void) sprintf (temp, "What about a noun, you frogging great cretin?\n");
    say (temp);
  }
}

/*--------------------------------------------------------------------------*/

#define X(attribute) victimstruct.attribute

void statperson (playerno, victimstruct)
  int playerno;
  struct persona victimstruct;
{
  static char temp[DESCLEN];

  if ( player[playerno].level < X(level) ) {
      say ( "Sorry that person is too high a level!\n" );
      return;
  }

  (void) sprintf (temp, "\
Client#: %d\tName: \"%s\"\tPasswd: \"%s\"\n\
Code: \"%s\"\tHost: \"%s\"\tFree: \"%s\"\n\
Logtime: %ld\tFd: %d\tPhysical: \"%s\"\n\
Haircolour: \"%s\"\tEyecolour: \"%s\"\tSex: \"%s\"\n\
Strength: %d\tIntelligence: %d\tAgility: %d\tHoopiness: %d\n\
Intcap: %d\tIntcont: %d\tLuck: %d\t\tPurity: %d\n\
Hp: %d\tMaxHp: %d\tGames: %d\tLevel: %d\n\
OldSc: %d\tNewSc: %d\tOldMv: %d\tNewMv: %d\n\
Wizstuff: %d\tIgnorestuff: %d\tMaimstuff: %d\tDisplaymode: %d\n\
Clientmode: %d\tWhere: %d\tSnoop: %d\tRecord_no: %d\n\
Invloc: %d\tSpare1: %ld\tSpare2: %ld\tSpare3: %ld\n\n",
		  playerno, X(name),
		  "<CENSORED>",
		  X(code), X(host), X(free),
		  0, X(fd), X(physical),
		  X(haircolour), X(eyecolour), X(sex),
		  X(strength), X(intelligence), X(agility), X(hoopiness),
		  X(intcap), X(intcont), X(luck), X(purity),
		  X(hp), X(maxhp), X(games), X(level),
		  X(oldscore), X(newscore), X(oldmoves), X(newmoves),
		  X(wizstuff), X(ignorestuff), X(maimstuff), X(displaymode),
		  X(clientmode), X(where), X(snoop), X(record_no),
		  X(invloc), X(spare1), X(spare2), X(spare3));
  say (temp);
}

#undef X

/*--------------------------------------------------------------------------*/

void statitem ( playerno, itemno )
    int playerno;
    int itemno;
{
    struct object *item = &(obj[itemno]);
    int i;

    if ( ! ARCHMODEP ( playerno ) ) {
	say ( "Naaaaah - only yer Arch-wizards can stat items!\n" );
	return;
    }
    say ( form ( "Object: %d\nDescription: \"%s\"\n", itemno, item->desc ) );
    for ( i = 0; i < item->name.wordc; i++ )
	say ( form ( "Name %d: \"%s\"\n", i+1, item->name.wordv[i] ) );
    say ( form ( "\
Location : %d\tValue : %d\tWeight: %d\tSize : %d\n\
Damage   : %d\tSeen  : %d\tKey   : %d\tSpeed: %d\n",
	     item->location,
	     item->value,
	     item->weight,
	     item->size,
	     item->damage,
	     item->seen,
	     objectkey(itemno),
	     item->speed ) );
    say ( form ( "Container: (0x%x)\n\t   %s %s %s %s\n\t   %s %s\n",
		item->container,
		transparentp(itemno) ? "(transparent)" : "(opaque)",
		openp(itemno)        ? "(open)" : "(closed)",
		openablep(itemno)    ? "(openable)" : "(unopenable)",
		containerp(itemno)   ? "(container)" :
    		    ( surfacep(itemno)   ? "(surface)" : "(solid)" ),
		lockablep(itemno)    ? "(lockable)" : "(notlockable)",
		lockedp(itemno)      ? "(locked)" : "(unlocked)" ) );
    say ( form ( "Flags    : (0x%x)\n\t   %s %s %s %s\n\t   %s %s %s %s\n",
		item->flag,
		moveablep(itemno)    ? "(moveable)" : "(immovable)",
		wearablep(itemno)    ? "(wearable)" : "(unwearable)",
		drinkablep(itemno)   ?
                    (poisonp(itemno)     ? "(poison)" : "(drinkable)") :
		    (ediblep(itemno)     ? "(edible)" : "(nonfood)"),
		singleshotp(itemno)  ? "(pistol)" :
		    (multishotp(itemno)  ? "(smg)" : "(nongun)" ),
		nuclearp(itemno)     ? "(nuclear)" : "(nonnuclear)",
		nuclearsavep(itemno) ? "(saver)"   : "(nonsaver)",
		weaponp(itemno)      ? "(weapon)"  :
  		    (cursedweaponp(itemno) ? "(cursed)" : "(nonweapon)"),
		explodep(itemno)     ? "(explode)" : "(stable)" ) );
    say ( form ( "\
Actions  : (0x%x)\n\t  %s %s %s %s\n\t   %s %s %s %s\n\t   %s %s %s\n",
		item->actions,
		act_movep(itemno)   ? "(act_moves)"   : "(act_static)",
		act_openp(itemno)   ? "(act_opens)"   : "(act_notopens)",
		act_sayp(itemno)    ? "(act_says)"    : "(act_notsays)",
		act_takep(itemno)   ? "(act_takes)"   : "(act_nottakes)",
		act_dropp(itemno)   ? "(act_drops)"   : "(act_notdrops)",
		act_lockp(itemno)   ? "(act_locks)"   : "(act_notlocks)",
		act_unlockp(itemno) ? "(act_unlocks)" : "(act_notunlocks)",
		act_shoutp(itemno)  ? "(act_shouts)"  : "(act_notshouts)",
		act_givep(itemno)   ? "(act_gives)"   : "(act_notgives)",
		act_stealp(itemno)  ? "(act_steals)"  : "(act_notsteals)",
		act_attackp(itemno) ? "(act_attacks)" : "(act_notattacks)" ) );
}
/*--------------------------------------------------------------------------*/

void statroom ( playerno, roomno )
    int playerno;
    int roomno;
{
    struct location *lp = &(loc[roomno]);

    if ( !ARCHMODEP ( playerno ) ) {
	say ( "Naaaah - only Arch-wizards can stat rooms!\n" );
	return;
    }

    say ( form ( "Room: %d\tName: %s\tCode: %s\n\
Exits:\n\
P : %s%s\tN : %s%s\tS : %s%s\tE : %s%s\tW: %s%s\n\
U : %s%s\tD : %s%s\tI : %s%s\tO : %s%s\n\
Ne: %s%s\tSe: %s%s\tNw: %s%s\tSw: %s%s\n\
Flags: (0x%x)\n\t(%s)  (%s)  (%s)  (%s)\n\t(%s)  (%s)  (%s)  (%s)  (%s)\n",
		roomno, lp->name, lp->code,
		exit_name(lp->exits[0]), exit_name(lp->exits[1]), exit_name(lp->exits[2]),
		exit_name(lp->exits[3]), exit_name(lp->exits[4]), exit_name(lp->exits[5]),
		exit_name(lp->exits[6]), exit_name(lp->exits[7]), exit_name(lp->exits[8]),
		exit_name(lp->exits[9]), exit_name(lp->exits[10]), exit_name(lp->exits[11]),
		exit_name(lp->exits[12]),
		lp->flags,
		loc_pitp(roomno) ? "Pit" : "Not_Pit",
		loc_noteletop(roomno) ? "No_Teleport_to" : "Teleport_to",
		loc_notelefrp(roomno) ? "No_Teleport_from" : "Teleport_from",
		loc_echop(roomno) ? "Has_echo" : "No_echo",
		loc_safeforcep(roomno) ? "No_forcing" : "Forcing",
		loc_safestealp(roomno) ? "No_stealing" : "Stealing",
		loc_safekillp(roomno) ? "No_fighting" : "Fighting",
		loc_safesnoopp(roomno) ? "No_snooping" : "Snooping",
		loc_safesummonp(roomno) ? "No_summoning" : "Summoning"
		) ) ;
}


/*-------------------------------------------------------------------------*/

void statstring (playerno, string)
  int playerno;
  char *string;
{
    char word[LINELEN];
    int ind, ingame;
    
    if (*string == '\0') {
	statperson (playerno, player[playerno]);
	return;
    }

    if (!WIZMODEP (playerno)) {
	say ("Only Wizards can use \"stat\" with arguments.\n");
	return;
    }
	    
    while (*string != '\0') {
	getnextobject (word, string);

	if ( (ingame = strmatchplayer (word)) != -1 ) {
	    statperson (playerno, player[ingame] );
	    continue;
	}
	if ((ind = strmatchpersona (word)) != -1) {
	    statperson (playerno, persona_list[ind]);
	    continue;
	}
	if ((ind = strtonoun ( word, player[playerno].where )) != -1 ) {
	    statitem (playerno, ind );
	    continue;
	}
	if ((ind = strtoroomcode ( word )) != -1 ) {
	    statroom ( playerno, ind );
	    continue;
	}
	if ( strlen ( word ) < 2 ) {
	    say (form ("\"%s\" is not a valid object/persona.\n", word));
	    continue;
	}
	ind = atoi ( &(word[1]) );
	switch ( word[0] ) {
	case 'l':
	case 'L':
	case 'r':
	case 'R':
	    if ( ind >= 0 && ind < no_locations )
		statroom ( playerno, ind );
	    else
		say ( form ( "No such room number %d!\n", ind ) );
	    break;
	case 'o':
	case 'O':
	case 'i':
	case 'I':
	    if ( ind >= 0 && ind < no_objects )
		statitem ( playerno, ind );
	    else 
		say ( form ( "No such item number %d!\n", ind ) );
	    break;
	default:
	    if ( ind >= 0 && ind < no_personas )
		statperson ( playerno, persona_list[ind] );
	    else
		say ( form ( "No such persona number %d!\n", ind ) );
	    break;
	}
    }
}

/*--------------------------------------------------------------------------*/
int snooploop ( playerno, victim )
    int playerno;
    int victim;
{
    int next = victim;

    while ( next != -1 ) {
	if ( (next = player[next].snoop) == playerno )
	    return ( 1 );
    }
    return ( 0 );
}

/*--------------------------------------------------------------------------*/

void snoopon (playerno, string)
  int playerno;
  char *string;
{
  int victim = -1;
  char item[LINELEN];

  if ( ! WIZMODEP ( playerno ) ) {
      say ( "Ah yes, very witty. A mortal snooping.  Ha!\n" );
      return;
  }

#ifdef ARCH_SNOOP
  if ( ! ARCHMODEP ( playerno ) ) {
      say ( "Frog off - Only Arch-wizards can do that!\n" );
      return;
  }
#endif

  getnextword (item, string);

  if (*item == '\0') {
      if (player[playerno].snoop != -1) 
	  say ( form ( "You are currently snooping on %s.\n",
		      player[player[playerno].snoop].name ) );
      else
	  say ( "You are not currently snooping on anyone.\n" );
      return;
  }

  if ((victim = strtoplayer (item)) == -1) {
      say ( form ( "There is no-one called  \"%s\" in the game!\n",
		  item ) );
      return;
  }
  if (victim == playerno || snooploop(playerno,victim)) {
      say ( "Recursion is RIGHT OUT!\n" );
      return;
  }
  if (player[playerno].snoop == victim) {
      say ( form ( "You are already snooping on %s.\n",
		  player[victim].name ) );
      return;
  }
  if (!wizdominatep ( playerno, victim )) {
      say ( form ( "Your not powerful enough to snoop on %s!\n",
		  player[victim].name ) );
      safewrite ( victim, form ( "*** %s tried to snoop on you.\n",
				player[playerno].name ) );
      return;
  }
  if (player[playerno].snoop != -1)
      say ( form ( "You stop snooping on %s.\n",
		  player[player[playerno].snoop].name) );
  player[playerno].snoop = victim;
  say ( form ( "You start snooping on %s.\n",
	      player[player[playerno].snoop].name ) );
}

/*--------------------------------------------------------------------------*/

void snoopoff (playerno)
  int playerno;
{
  if ( ! WIZMODEP ( playerno ) ) {
      say ( "Mortals can't snoop => can't unsnoop.  Simple!\n");
      return;
  }

#ifdef ARCH_SNOOP
  if ( ! ARCHMODEP ( playerno ) ) {
      say ( "Frog off - Only Arch-wizards can do that!\n" );
      return;
  }
#endif

  if (player[playerno].snoop == -1) {
      say ( "You are not snooping *now*, you blurfl!\n" );
      return;
  }
  say ( form ( "You stop snooping on %s.\n",
	      player[player[playerno].snoop].name ) );
  player[playerno].snoop = -1;
}

/*--------------------------------------------------------------------------*/

void force (playerno, string)
    int playerno;
    char *string;
{
    int victim = -1, cost;
    char item[LINELEN];
    
    if (!WIZMODEP (playerno)) {
	say ("You just concentrate on your OWN game, mortal!\n");
	return;
    }

    getnextword (item, string);

    if (*item == '\0') {
	say ("Force WHO, gonzo?\n");
	return;
    }
    if ((victim = strtoplayer (item)) == -1) {
	say ( form ( "There is no-one called  \"%s\" in the game!\n",
			item ) );
	return;
    }

    if ( loc_safeforcep(player[victim].where) && !ARCHMODEP(playerno) ) {
	say ( form ( "%s is in a safe location!\n", item ) );
	return;
    }

    if ( ! wizdominatep ( playerno, victim ) ) {
	say ( form ( "You not powerful enough to force %s!\n",
		    player[victim].name ) );
	safewrite ( victim, form ( "*** %s tried to force you to \"%s\"!\n",
		   player[playerno].name, string ) );
	return;
    }
    if ( ! ARCHMODEP ( playerno ) && player[victim].where == 0 ) {
	say ( "Only arch-wizards may force people in limbo!\n" );
	return;
    }

    if ( ! ARCHMODEP ( playerno ) && player[victim].where == UPITLOC ) {
	say ( "Only arch-wizards may force people who are under the pit!\n" );
	return;
    }
    if ( !ARCHMODEP (playerno ) ) {
	cost = player[victim].intelligence * 2 + 400;
	if ( ! castablep ( playerno, cost ) ) {
	    say ( "Frog off you aint got enough points!\n" );
	    return;
	}
    }

    if (*string == '\0') {
	say ( form ( "Use: \"force %s <action>\"\n",
			  player[victim].name ) );
	return;
    }
    say ( form ( "You force %s to \"%s\".\n",
		player[victim].name, string ) );
    safewrite ( victim, form ( "%s forces you to \"%s\".\n",
			      player[playerno].name, string ) );
    handle (victim, string);
}

/*--------------------------------------------------------------------------*/

void delete_string (playerno, string)
  int playerno;
  char *string;
{
  int number = 0;
  int victim_ind;
  char item[LINELEN];

  if (!ARCHMODEP (playerno)) {
    say ("Chicken off!  Only Arch-wizzes can delete personas!\n");
    return;
  }

  while (*string != '\0') {
    number++;
    getnextobject (item, string);
    if ((victim_ind = strmatchpersona (item)) != -1)
      delete_persona (playerno, victim_ind);
    else
      say (form ("There's no-one on file called \"%s\".\n", item));
  }
  
  if (number == 0) {
    if (player[playerno].record_no == -1)
      say ("Your persona hasn't yet been saved.\n");
    else
      delete_persona (playerno, player[playerno].record_no);
  }
}

/*-------------------------------------------------------------------------*/

void crash (playerno)
  int playerno;
{
    int pitted = totalvalue (UPITLOC) + totalvalue (STORELOC);
    int i;

    if (pitted < totalvalue (-1) && !ARCHMODEP(playerno)) {
	say ( "Non arch-wizards can't reset until all treasure has been pitted.\n");
	return;
    }

    for ( i = 0; i < MAXUSERS; i++ ) {
	if ( (player[i].fd != -1) && ISACTIVE(i) ) {
	    save_persona ( i, "" );
	    safewrite ( i, "*** Persona saved! ***\n" );
	}
    }
    
	
    shout ("\nSomething very magical is happening ...\n", NO_ONE, MESG_WARN);
    shout ("Everything is being reset to its start position.\n",
	   NO_ONE, MESG_WARN);
    shout ("Login to MUNDI again in a few seconds.\n", NO_ONE, MESG_WARN);
    quit (16);
}

/*--------------------------------------------------------------------------*/

void exorcise (playerno, string)
    int playerno;
    char *string;
{
    char victimstr[LINELEN], *mesg;
    int i, victim, cost;

#ifndef EXBYSCORE
    if ( ! WIZMODEP ( playerno ) ) {
	say ( "You a mere motal exorcise! Pah!\n" );
	return;
    }
#endif    
    
    getnextword (victimstr, string);
    if (*victimstr == '\0') {
	say ("Exorcise whom?\n");
	return;
    }
    if ((victim = strtoplayer (victimstr)) == -1) {
	say ( form ( "There is no-one called \"%s\" logged in.\n",
		    victimstr ) );
	return;
    }
    if (victim == playerno) {
	say ("No.  It's much too silly.\n");
	return;
    }
    if ( ( player[playerno].where != player[victim].where ) &&
	( ! WIZMODEP ( playerno ) ) ) {
	say ( form ( "%s isn't here!\n",
		    player[victim].name ) );
	return;
    }
    if ( ! wizdominatep ( playerno, victim ) ) {
	say ( form ( "You not powerful enough to exorcise %s!\n",
		    player[victim].name ) );
	safewrite ( victim, form ( "*** %s tried to exorcise you!\n",
		   player[playerno].name ) );
	return;
    }
    if ( ( player[victim].where == 0 ) && ( ! WIZMODEP ( playerno ) ) ) {
	say ("Limbo is a safe location - no EXing here!\n");
	return;
    }
    if ( !ARCHMODEP (playerno ) ) {
	cost = totalvalue(player[victim].invloc) +
	       player[victim].intelligence + 200;
	if ( ! castablep ( playerno, cost ) ) {
	    say ( "Frog off you aint got enough points!\n" );
	    return;
	}
	change_purity ( playerno, -cost / 10 );
    }

    safewrite ( victim, form ( "%s%s\n", EXORCISED,
	       player[playerno].name ) );
    say ( form ( "*** You exorcise %s!\n",
		player[victim].name ) );
    shout ( form ( "*** %s has exorcised %s!\n",
		  player[playerno].name, player[victim].name ),
	   victim, MESG_GLOBAL );
    killoff (victim);
}

/*--------------------------------------------------------------------------*/

void superex (playerno, string)
    int playerno;
    char *string;
{
    char victimstr[LINELEN];
    int i, victim;
    
    if (!WIZMODEP (playerno)) {
	say ("NO chance matey!  You are MORTAL!\n");
	return;
    }
    if (!ARCHMODEP (playerno)) {
	say ("Pah!  You might be a wizard, but you ain't no Arch, boy!\n");
	return;
    }
    getnextword (victimstr, string);
    if (*victimstr == '\0') {
	say ("SuperExorcise whom?\n");
	return;
    }
    if ((victim = strtoplayer (victimstr)) == -1) {
	say ( form ( "There is no-one called \"%s\" logged in.\n",
			  victimstr ) );
	return;
    }
    if (victim == playerno) {
	say ("No.  It's much too silly.\n");
	return;
    }
    if (ARCHMODEP (victim) && !METAMODEP(playerno) ) {
	say ("You can't SuperExorcise another Arch-wiz!!\n");
	safewrite ( victim, form ( "### %s tried to SuperExorcise you!\n",
				  player[playerno].name) );
	return;
    }
    safewrite ( victim, form ( "%s%s\n", SUPEREXED,
			      player[playerno].name ) );
    say ( form ( "### You SuperExorcise %s!\n",
		player[victim].name ) );
    shout ( form ( "### %s has SuperExorcised %s!\n",
		  player[playerno].name, player[victim].name ),
	   victim, MESG_GLOBAL );
    killoff (victim);
}

/*--------------------------------------------------------------------------*/

void plist (playerno, command)
    int playerno;
    char *command;
{
    int i, passwd = 0;
    char match[LINELEN];

    if (!ARCHMODEP (playerno)) {
	say ("Only arch-wizards can do that kinda heavy stuff.\n");
	return;
    }

    getnextword ( match, command );
    if ( samenstring ( match, "passwd" ) ) {
	if ( METAMODEP (playerno) ) {
	    passwd = 1;
	    getnextword ( match, command );
	} else {
	    say ( "FROG OFF - you can't see passwd's.\n" );
	    return;
	}
    }

    for (i = 0; i < no_personas; i++) {
	if ( samestring ( persona_list[i].name, match ) ||
	    *match == '\0' ) {
	    say (form ("Record# %d\tName: \"%s\"\tPassword: \"%s\"\tWiz: %d\n",
		       persona_list[i].record_no, persona_list[i].name,
		       (passwd ? persona_list[i].passwd : "<CENSORED>"),
		       persona_list[i].wizstuff));
	}
    }
}

/***************************************************************************\
|*									   *|
|*  The function frog() is available to arch-wizards and above only since  *|
|*  it involves low-level (ie. program- rather than game-level) twiddling  *|
|*  of persona structure data.  It is equivalent to MUD's "frig".	   *|
|*									   *|
|*  Syntax is: "frog persona attribute value"				   *|
|*									   *|
\***************************************************************************/

#define FRIG_STR(xxx, len) (samestring (#xxx, attr)) { \
	(void) strncpy (pptr->xxx, value, len); \
	say (form ("Attribute \"%s\" changed to \"%s\".\n", #xxx, value)); }

#define FRIG_INT(xxx) (samestring (#xxx, attr)) { \
	pptr->xxx = intval; \
	say (form ("Attribute \"%s\" changed to \"%ld\".\n", #xxx, intval)); }

void frog (playerno, args)
    int playerno;
    char *args;
{
    static char attr[LINELEN];
    static char value[LINELEN];
    static char person[LINELEN];	/* persona being frigged (or null) */
    long int intval;		/* The value of the string "value" :-) */
    int tmp, level;		/* Index into player or persona array */
    struct persona *pptr;	/* Pointer to persona being twiddled */
    
    if ( !ARCHMODEP ( playerno ) ) {
	say ("Go stick your head in a ferret -- You ain't no kinda archwiz!\n");
	return;
    }

    getnextword (person, args);    
    getnextword (attr, args);
    getnextobject (value, args );
    intval = atol (value);
    
    if ((tmp = strtoplayer (person)) != -1)
	pptr = &player[tmp];
    else
	if ((tmp = strtopersona (person)) != -1)
	    pptr = &persona_list[tmp];
	else {
	    if ( strlen(person) > 1 && person[0] == '#' ) {
		tmp = atoi(&(person[1]));
		pptr = &persona_list[tmp];
	    } else {
		say ( form ( "There is no persona \"%s\"!\n", person ) );
		return;
	    }
	}
    
    if (*value == '\0') {
	say ("Syntax is: \"frog person attribute value\".\n");
	return;

    }

    if ( ( pptr->wizstuff & CANARCH ) && !METAMODEP(playerno) )
	if ( player[playerno].record_no != pptr->record_no ) {
	    say ( "Sorry you can't frog around with other arch's!\n" );
	    return;
	}
    
    if FRIG_STR (name, NAMELEN)	/* These stats wont frog up the daemon */
    else if FRIG_STR (passwd, PWLEN)
    else if FRIG_STR (free, CODELEN)
    else if FRIG_STR (physical, PHYSLEN)
    else if FRIG_STR (haircolour, HAIRLEN)
    else if FRIG_STR (eyecolour, EYELEN)
    else if FRIG_STR (sex, SEXLEN)
    else if FRIG_INT (strength)
    else if FRIG_INT (intelligence)
    else if FRIG_INT (agility)
    else if FRIG_INT (hoopiness)
    else if FRIG_INT (intcap)
    else if FRIG_INT (intcont)
    else if FRIG_INT (luck)
    else if FRIG_INT (purity)
    else if FRIG_INT (hp)
    else if FRIG_INT (maxhp)
    else if FRIG_INT (maimstuff)
    else if FRIG_INT (level)
    else if FRIG_INT (newscore)
    else if FRIG_INT (oldscore)
    else if FRIG_INT (ignorestuff) /* These are dangerous...DONT fiddle too */
    else if FRIG_INT (record_no)   /* much unless you know what your doing  */
    else if FRIG_INT (fd)		
    else if FRIG_INT (displaymode)
    else if FRIG_INT (clientmode)
    else if FRIG_INT (where)
    else if FRIG_INT (spare1)
    else if FRIG_INT (spare2)
    else if FRIG_INT (spare3)
    else if (METAMODEP(playerno)) { 
	if FRIG_INT (wizstuff)
    }
    else {
	say (form ("There is no attribute \"%s\".\n", attr));
	return;
    }
    
    /* Check level vomit */
    level = compute_level ( sc_levels, 
			   pptr->oldscore + pptr->newscore,
			   SC_LEVELS );
    if ( level != pptr->level ) {
	pptr->level = level;
	if ( level < MIN_WIZLEVEL )
	    pptr->wizstuff &= ~CANWIZ;
	else
	    pptr->wizstuff |= CANWIZ;
    } 
}

/*-------------------------------------------------------------------------*/

/***************************************************************************\
|*									   *|
|*  The function tea() is available to arch-wizards and above only since   *|
|*  it involves low-level (ie. program- rather than game-level) twiddling  *|
|*  of object structure data. 						   *|
|*									   *|
|*  Syntax is: "tea attribute value item"				   *|
|*									   *|
\***************************************************************************/

/* 2001: changed all these macros to use ANSI-style stringification */
#define TEA_STR(xxx, len) (samestring (#xxx, attr)) { \
	(void) strncpy (optr->xxx, value, len); \
	say (form ("Attribute \"%s\" changed to \"%s\".\n", #xxx, value)); }

#define TEA_INT(xxx) (samestring (#xxx, attr)) { \
	optr->xxx = intval; \
	say (form ("Attribute \"%s\" changed to \"%ld\".\n", #xxx, intval)); }

#define TEA_INTLOC(xxx) (samestring (#xxx, attr)) { \
        if ((ind = strtoroomcode ( value ))==-1) \
	    optr->xxx = intval; \
	else \
            intval = optr->xxx = ind; \
	say (form ("Attribute \"%s\" changed to \"%ld\".\n", #xxx, intval)); }

#define TEA_INTOBJ(xxx) (samestring (#xxx, attr)) { \
        if ((ind = strtonoun ( value, player[playerno].where ))==-1) \
	    optr->xxx = intval; \
	else \
            intval = optr->xxx = ind; \
	say (form ("Attribute \"%s\" changed to \"%ld\".\n", #xxx, intval)); }

#define TEA_FLAGON(yy,xxx,bit) (samestring (#xxx, attr)) { \
        optr->yy |= 1<<bit; \
        say ( form ("Attribute \"%s\" flag \"%s\" set.\n", #yy, #xxx )); }

#define TEA_FLAGOFF(yy,xxx,bit) (samestring (#xxx, attr)) { \
        optr->yy &= ~(1<<bit); \
        say ( form ("Attribute \"%s\" flag \"%s\" set.\n", #yy, #xxx)); }

#define TEA_NAME(xxx,len) (samestring (attr, #xxx)) { \
        ind = atoi ( &(attr[strlen ( attr )-1]) ); \
        if ( ind > 0 && ind <= optr->name.wordc ) \
	    say ( form ( "Attribute \"%s\" changed to \"%s\".\n", \
			attr, value ) ); \
        else { \
	    ind = ++(optr->name.wordc); \
            say ( form ( "Creating Name %d attribute as \"%s\"\n", \
			ind, value ) ); \
	} \
        strncpy ( optr->name.wordv[ind-1], value, len ); }

							      
void tea (playerno, value)
  int playerno;
  char *value;
{
  static char attr[LINELEN];
  static char item[LINELEN];	/* item being tea'd */
  long int intval;		/* The value of the string "value" :-) */
  int tmp, ind;			/* Index into item */
  struct object *optr;		/* Pointer to object being twiddled */

  if ( !ARCHMODEP ( playerno ) ) {
      say ("Go stick your head in a ferret -- You ain't no kinda archwiz!\n");
      return;
  }

  getnextword (item, value);
  getnextword (attr, value);
  intval = atol (value);

  if (*item == '\0' || *attr == '\0' ) {
      say ("Syntax is: \"tea item attribute [value]\".\n");
      return;
  }

  if ((tmp = strtonoun ( item, player[playerno].where )) == -1 ) {
      if ( strlen(item) > 1 && item[0] == '#' ) {
	  tmp = atoi(&(item[1]));
      } else {
	  say ( form ( "I can't seem to find an object \"%s\"!\n",
		      item ) );
	  return;
      }
  }
  optr = &(obj[tmp]);
  
  if TEA_STR (desc, LINELEN)	/* These stats wont frog up the daemon */
  else if TEA_NAME(name, WORDLEN)
  else if TEA_INTLOC (location)
  else if TEA_INT (value)
  else if TEA_INT (weight)
  else if TEA_INT (size)

  else if TEA_FLAGON (flag,moveable,0)
  else if TEA_FLAGOFF (flag,immoveable,0)
  else if TEA_FLAGON (flag,wearable,1)
  else if TEA_FLAGOFF (flag,unwearable,1)
  else if TEA_FLAGON (flag,drinkable,2)
  else if TEA_FLAGOFF (flag,edible,2)
  else if TEA_FLAGON (flag,fireable,3)
  else if TEA_FLAGOFF (flag,unfireable,3)
  else if TEA_FLAGON (flag,multishot,4)
  else if TEA_FLAGOFF (flag,singleshot,4)
  else if TEA_FLAGON (flag,nuclear,5)
  else if TEA_FLAGOFF (flag,nonnuclear,5)
  else if TEA_FLAGON (flag,saver,6)
  else if TEA_FLAGOFF (flag,nonsaver,6)
  else if TEA_FLAGON (flag,nonweapon,7)
  else if TEA_FLAGOFF (flag,weapon,7)
  else if TEA_FLAGON (flag,explode,8)
  else if TEA_FLAGOFF (flag,stable,8)

  else if TEA_FLAGOFF (container,opaque,0)
  else if TEA_FLAGON (container,transparent,0)
  else if TEA_FLAGOFF (container,closed,1)
  else if TEA_FLAGON (container,open,1)
  else if TEA_FLAGOFF (container,unopenable,2)
  else if TEA_FLAGON (container,openable,2)
  else if TEA_FLAGOFF (container,container,3)
  else if TEA_FLAGON (container,surface,3)
  else if TEA_FLAGOFF (container,notlockable,4)
  else if TEA_FLAGON (container,lockable,4)
  else if TEA_FLAGOFF (container,unlocked,5)
  else if TEA_FLAGON (container,locked,5)

  else if TEA_INT (seen)
  else if TEA_INT (damage)
  else if TEA_INTOBJ (key)
  else if TEA_INT (speed)

  else if TEA_FLAGON (actions,act_moves,0 )
  else if TEA_FLAGOFF (actions,act_notmoves,0)
  else if TEA_FLAGON (actions,act_opens,1 )
  else if TEA_FLAGOFF (actions,act_notopens,1)
  else if TEA_FLAGON (actions,act_says,2 )
  else if TEA_FLAGOFF (actions,act_notsays,2)
  else if TEA_FLAGON (actions,act_takes,3 )
  else if TEA_FLAGOFF (actions,act_nottakes,3)
  else if TEA_FLAGON (actions,act_drops,4 )
  else if TEA_FLAGOFF (actions,act_notdrops,4)
  else if TEA_FLAGON (actions,act_locks,5 )
  else if TEA_FLAGOFF (actions,act_notlocks,5)
  else if TEA_FLAGON (actions,act_unlocks,6 )
  else if TEA_FLAGOFF (actions,act_notunlocks,6)
  else if TEA_FLAGON (actions,act_shouts,7 )
  else if TEA_FLAGOFF (actions,act_notshouts,7)
  else if TEA_FLAGON (actions,act_gives,8 )
  else if TEA_FLAGOFF (actions,act_notgives,8)
  else if TEA_FLAGON (actions,act_steals,9 )
  else if TEA_FLAGOFF (actions,act_notsteals,9)
  else if TEA_FLAGON (actions,act_attacks,10 )
  else if TEA_FLAGOFF (actions,act_notattacks,10)

  else {
    say (form ("There is no attribute \"%s\".\n", attr));
    return;
  }
}

/*--------------------------------------------------------------------------*/
				/* Substitutes \n and \t for correct values */
char *sub_escapes ( text )
    char *text;
{
    int esc = 0;
    char ch, *np = text, *ret = text;

    while ( ch = *text++ ) {
	if ( esc ) {
	    switch ( ch ) {
	    case 'n':
		*np++ = '\n';
		break;
	    case 't':
		*np++ = '\t';
		break;
	    default:
		*np++ = '\\';
		*np++ = ch;
		break;
	    }
	    esc = 0;
	} else {
	    if ( ch == '\\' )
		esc = 1;
	    else
		*np++ = ch;
	}
    }
    *np = '\0';
    return ( ret );
}

/****************************************************************************\
  Decorate will change attributes of rooms.

  Ie. decorate room
\****************************************************************************/

#define DECOR_FLAGON(yy,xxx,bit) (samestring (#xxx, attr)) { \
        rptr->yy |= 1<<bit; \
        say ( form ("Attribute \"%s\" flag \"%s\" set.\n", #yy, #xxx )); }

#define DECOR_FLAGOFF(yy,xxx,bit) (samestring (#xxx, attr)) { \
        rptr->yy &= ~(1<<bit); \
        say ( form ("Attribute \"%s\" flag \"%s\" set.\n", #yy, #xxx )); }

#define DECOR_STR(xxx, len) (samestring (#xxx, attr)) { \
	(void) strncpy (rptr->xxx, value, len); \
	say (form ("Attribute \"%s\" changed to \"%s\".\n", #xxx, value)); }

#define DECOR_INT(xxx) (samestring (#xxx, attr)) { \
	rptr->xxx = intval; \
	say (form ("Attribute \"%s\" changed to \"%ld\".\n", #xxx, intval)); }

				/* Re mallocs text in attr */
#define DECOR_MALLOC(xxx) (samestring (#xxx, attr)) { \
	free ( rptr->xxx ); rptr->xxx = strdup ( sub_escapes ( value ) ); \
	say ( form ("Attribute \"%s\" realloced.\n" , #xxx)); }

				/* Adds more text to attr */
#define DECOR_MALLOCADD(xxx,yyy) (samestring (#xxx, attr)) { \
        ind = strlen ( rptr->yyy ); \
        (void) sub_escapes ( value ); \
	rptr->yyy = realloc ( rptr->yyy, ind + strlen ( value ) + 1 ); \
        strcat ( rptr->yyy, value ); \
	say ( form ("Attribute \"%s\" added to.\n", #xxx )); }

				/* Adds exit to location */
#define DECOR_SUBINT(field,xxx,sub) (samestring (#xxx, attr)) { \
        if ((ind = strtoroomcode ( value )) == -1 ) \
	    rptr->field[sub] = intval; \
        else \
	    intval = rptr->field[sub] = ind; \
	say (form ("Attribute \"%s\" changed to \"%ld\".\n", #xxx, intval)); }

void decorate ( playerno, value )
    int playerno;
    char *value;
{
  static char attr[LINELEN];
  static char room[LINELEN];	/* room being decorated */
  long int intval;		/* The value of the string "value" :-) */
  int tmp, ind;			/* Index into room */
  struct location *rptr;	/* Pointer to room being twiddled */

  if ( !ARCHMODEP ( playerno ) ) {
      say ("Go stick your head in a ferret -- You ain't no kinda archwiz!\n");
      return;
  }

  getnextword (room, value);
  getnextword (attr, value);
  if ((intval = strtoroomcode ( value )) == -1 )
      intval = atol (value);

  if (*room == '\0' || *attr == '\0' ) {
      say ("Syntax is: \"decorate item attribute [value]\".\n");
      return;
  }

  if ((tmp = strtoroomcode ( room )) == -1 ) {
      if ( strlen(room) > 1 && room[0] == '#' ) {
	  tmp = atoi(&(room[1]));
      } else {
	  say ( form ( "I can't seem to find a room \"%s\"!\n",
		      room ) );
	  return;
      }
  }
  rptr = &(loc[tmp]);

  if DECOR_STR(name,LINELEN)
  else if DECOR_STR(code,WORDLEN)
  else if DECOR_MALLOC(desc)
  else if DECOR_MALLOCADD(adddesc,desc)
  else if DECOR_SUBINT(exits,pit,0)
  else if DECOR_SUBINT(exits,north,1)
  else if DECOR_SUBINT(exits,south,2)
  else if DECOR_SUBINT(exits,east,3)
  else if DECOR_SUBINT(exits,west,4)
  else if DECOR_SUBINT(exits,up,5)
  else if DECOR_SUBINT(exits,down,6)
  else if DECOR_SUBINT(exits,in,7)
  else if DECOR_SUBINT(exits,out,8)
  else if DECOR_SUBINT(exits,ne,9)
  else if DECOR_SUBINT(exits,northeast,9)
  else if DECOR_SUBINT(exits,se,10)
  else if DECOR_SUBINT(exits,southeast,10)
  else if DECOR_SUBINT(exits,nw,11)
  else if DECOR_SUBINT(exits,northwest,11)
  else if DECOR_SUBINT(exits,sw,12)
  else if DECOR_SUBINT(exits,southwest,12)
  else if DECOR_FLAGON(flags,_pit,0)
  else if DECOR_FLAGOFF(flags,no_pit,0)
  else if DECOR_FLAGON(flags,no_tele_to,1)
  else if DECOR_FLAGOFF(flags,tele_to,1)
  else if DECOR_FLAGON(flags,no_tele_from,2)
  else if DECOR_FLAGOFF(flags,tele_from,2)
  else if DECOR_FLAGON(flags,echo,3)
  else if DECOR_FLAGOFF(flags,no_echo,3)
  else if DECOR_FLAGON(flags,force,4)
  else if DECOR_FLAGOFF(flags,no_force,4)
  else if DECOR_FLAGON(flags,steal,5)
  else if DECOR_FLAGOFF(flags,no_steal,5)
  else if DECOR_FLAGON(flags,kill,6)
  else if DECOR_FLAGOFF(flags,no_kill,6)
  else if DECOR_FLAGON(flags,snoop,7)
  else if DECOR_FLAGOFF(flags,no_snoop,7)
  else if DECOR_FLAGON(flags,summon,8)
  else if DECOR_FLAGOFF(flags,no_summon,8)
  else {
      say ( form ("There is no attribute \"%s\".\n", attr ));
      return;
  }

}

/*--------------------------------------------------------------------------*/

void mkobject ( playerno )
    int playerno;
{
    int ind;

    if ( !ARCHMODEP (playerno) ) {
	say ( "Sorry only archwizards can create objects!\n" );
	return;
    }

    if ( no_objects >= MAXOBJECTS-1 ) {
	say ( "Object table is full - sorry!\n" );
	return;
    }

    ind = no_objects++;

    obj[ind].location = player[playerno].invloc;
    strcpy ( obj[ind].desc, form ( "Obj#%d", ind ) );
    strcpy ( obj[ind].name.wordv[0], form ( "Obj#%d", ind ) );
    obj[ind].name.wordc = 1;

    say ( form ( "\"Obj#%d\" has been created in your inventory!\n",
		ind ) );
}

/*--------------------------------------------------------------------------*/

void mkroom ( playerno )
    int playerno;
{
    int i, ind;
    if ( !ARCHMODEP (playerno) ) {
	say ( "Sorry only archwizards can create rooms!\n" );
	return;
    }

    if ( no_locations >= MAXLOCATIONS-1 ) {
	say ( "Location table is full - sorry!\n" );
	return;
    }

    ind = no_locations++;

    loc[ind].desc = strdup ( form ( "Loc#%d\n", ind ) );
    strcpy ( loc[ind].name, form ( "Loc#%d", ind ) );
    strcpy ( loc[ind].code, form ( "Loc#%d", ind ) );
    for ( i = 0; i < no_directions; i++ )
	loc[ind].exits[i] = 0;
    say ( form ( "\"Loc#%d\" has been created in hyperspace!\n",
		ind ) );
}

/*--------------------------------------------------------------------------*/

void saveobj ( playerno, args )
    int playerno;
    char *args;
{
    FILE *fp;
    int item;

    if ( ! ARCHMODEP ( playerno ) ) {
	say ( "Phlegm off you aint no kind of archwiz matey!\n" );
	return;
    }

    if ((item = strtonoun ( args, player[playerno].where ) ) == -1 ) {
	say ( form ( "There isn't an object \"%s\" in the game.\n", args ) );
	return;
    }

    if ( ( fp = fopen ( form ( "%s/%s/%s", MUNDI_DIR,
			      SAVE_DIR, OBJECTS_SAVE ), "a" ) ) == NULL ) {
	say ( form ( "Could not open object save file!\n" ) );
	return;
    }

    fprintf ( fp, "\n# -------- Dumped: %16.16s --------\n\n", thedate () );

    write_object ( fp, item );

    fclose ( fp );
    say ( form ( "Saved object \"%s\" to disk.\n", obj[item].desc ) );
}

/*--------------------------------------------------------------------------*/

void saveroom ( playerno, args )
    int playerno;
    char *args;
{
    FILE *fp;
    int room;

    if ( ! ARCHMODEP ( playerno ) ) {
	say ( "Phlegm off you aint no kind of archwiz matey!\n" );
	return;
    }

    if ((room = strtoroomcode ( args ) ) == -1 ) {
	say ( form ( "There isn't a room \"%s\" in the game.\n", args ) );
	return;
    }

    if ( ( fp = fopen ( form ( "%s/%s/%s", MUNDI_DIR,
			      SAVE_DIR, LOCATIONS_SAVE ), "a" ) ) == NULL ) {
	say ( form ( "Could not open room save file!\n" ) );
	return;
    }
    fprintf ( fp, "\n# -------- Dumped: %16.16s --------\n\n", thedate () );

    write_room ( fp, room );

    fclose ( fp );
    say ( form ( "Saved room \"%s\" to disk.\n", loc[room].name ) );
}

/****************************************************************************\
  List persona's record number by matching with a particular field and
  value! Usefull for recovering damaged personas (which seems to happen
  too much for my liking!)
\****************************************************************************/

#define MATCH_STR(field) \
    if ( samestring("field",attr) ) { \
	if (   (strcmp(persona_list[i].field,lowvalue) >= 0) \
	    && (strcmp(persona_list[i].field,hivalue) <= 0) ) \
	    say ( form ( "Record#%d:\tName:\"%s\"\t%s: \"%s\"\n", \
			i, persona_list[i].name, \
			"field", persona_list[i].field ) ); \
	continue; \
    }

#define MATCH_INT(field) \
    if ( samestring("field",attr) ) { \
	if (   (persona_list[i].field >= lowval) \
	    && (persona_list[i].field <= hival) ) \
	    say ( form ( "Record#%d:\tName:\"%s\"\t%s: %d\n", \
			i, persona_list[i].name, \
			"field", persona_list[i].field ) ); \
	continue; \
    }


#define MATCH_LAST(field)  \
    if ( samestring("field",attr) ) { \
	time_t t_t = (persona_list[i].field-time(0))/3600;\
	if (   (t_t >= lowval) \
	    && (t_t <= hival) ) { \
	    t_t = persona_list[i].field-(time(0)/3600);\
	    say ( form ( "Record#%d:\tName:\"%s\"\t%s: %s\n", \
			i, persona_list[i].name, \
			"field", ctime(&t_t) ) ); \
	} \
	continue; \
    }

#define MATCH_TIME(field) \
    if ( samestring("field",attr) ) { \
	if (   (persona_list[i].field >= lowval) \
	    && (persona_list[i].field <= hival) ) \
	    say ( form ( "\
Record#%d:\tName:\"%s\"\t%s: %d Days, %2.2d:%2.2d:%2.2d\n", \
			i, persona_list[i].name, \
			"field", \
			persona_list[i].field / (3600*24), \
			(persona_list[i].field % (3600*24)) / 3600, \
			(persona_list[i].field % (3600)) / 60, \
			(persona_list[i].field % 60) ) ); \
	continue; \
    }

void match ( playerno, args )
    int playerno;
    char *args;
{
    static char attr[LINELEN];
    static char lowvalue[LINELEN];
    static char hivalue[LINELEN];
    int lowval, hival, i;

    if ( !ARCHMODEP ( playerno ) ) {
	say ("Go stick your head in a ferret -- You ain't no kinda archwiz!\n");
	return;
    }
    getnextword (attr, args);
    getnextword (lowvalue, args);
    getnextword (hivalue, args);

    lowval = atol (lowvalue);
    hival = atol (hivalue);
    
    if (*attr == '\0' || *lowvalue == '\0' || *hivalue == '\0') {
	say ("Syntax is: \"match attribute lower upper.\n");
	return;
    }

    for ( i = 0; i < no_personas; i++ ) {
	MATCH_STR(name);
	MATCH_STR(passwd);
	MATCH_STR(code);
	MATCH_STR(free);
	MATCH_STR(physical);
	MATCH_STR(haircolour);
	MATCH_STR(eyecolour);
	MATCH_STR(sex);
	MATCH_LAST(logtime);
	MATCH_TIME(totaltime);
	MATCH_INT(strength);
	MATCH_INT(intelligence);
	MATCH_INT(agility);
	MATCH_INT(hoopiness);
	MATCH_INT(intcap);
	MATCH_INT(intcont);
	MATCH_INT(luck);
	MATCH_INT(purity);
	MATCH_INT(hp);
	MATCH_INT(maxhp);
	MATCH_INT(games);
	MATCH_INT(oldscore);
	MATCH_INT(newscore);
	MATCH_INT(oldmoves);
	MATCH_INT(newmoves);
	MATCH_INT(level);
	MATCH_INT(wizstuff);
	say ( form ( "Cannot match values on attribute \"%s\"!\n", attr ) );
	break;
    }
}

/*--------------------------------------------------------------------------*/

void sgetstring (playerno, string)
  int playerno;
  char *string;
{
  int number = 0, where;
  int ind = -1, cost;
  int gotone = 0;
  char item[LINELEN];
  char temp[LINELEN];

  if (!WIZMODEP (playerno)) {
    say ("No wiz-mode => no SuperGet.  Simple.  Frog off.\n");
    return;
  }

  while (*string != '\0') {
    number++;
    getnextobject (item, string);
    if ((ind = strtonoun (item, player[playerno].where)) == -1)
      say (form ("I don't know what a \"%s\" is.\n", item));
    else {
	where = real_loc ( ind );
      if ((where == UPITLOC) && (!ARCHMODEP (playerno)))
	say ("Sorry, it has been pitted.\n");
      else {
	if (where > MAXLOCATIONS) {
	    if ( ( where = whoholdsitem ( ind ) ) != -1 )
		say (form ("Sorry, the %s is being carried by %s.\n",
			   obj[ind].desc,
			   player[where].name));
	    else
		say (form ("Sorry, the %s is not in the game!\n",
			   obj[ind].desc ) );
	} else {
	    if (!ARCHMODEP(playerno)) {
		if ( (totallocweight(player[playerno].invloc)
		      + obj[ind].weight ) > weight_limit(playerno) ) {
		    say ( "You can't carry as much as that!\n" );
		    return;
		}
	    }
	    if ( !ARCHMODEP (playerno ) ) {
		cost = abs ( obj[ind].value )
		    +  abs ( obj[ind].damage ) * 10
		    +  obj[ind].weight * 10
		    +  obj[ind].size * 10
		    +  obj[ind].flag * 10;
		if ( ! castablep ( playerno, cost ) ) {
		    say ( "Frog off you aint got enough points!\n" );
		    continue;
		}
	    }
	    speak ( form ( "\
A magic hand picks up the %s, and both vanish in a puff of magic.",
			  obj[ind].desc ),
		   NO_ONE, where );
	    obj[ind].location = playerno+MAXLOCATIONS;
	    say ( form ("\
With a \"Ni!\" sound, the %s appears in your hands.\n",
			obj[ind].desc));
	}
      }
    }
  }
  
  if (number == 0)
    say ("Where are the nouns, you putrid pile of camembert?\n");
}

/*--------------------------------------------------------------------------*/

void delay_cmd ( playerno, command )
    int playerno;
    char *command;
{
    int	delay;			/* Time to delay in seconds */
    char value[LINELEN];

    if ( ! ARCHMODEP ( playerno ) ) {
	say ( "Sorry only Arch-wizards can delay commands\n" );
	return;
    }

    getnextword ( value, command );

    if ( *value == '\0' || *command == '\0' ) {
	say ( "usage: delay time command\n" );
	return;
    }

    delay = atoi ( value );

    add_event ( delay, 0, ev_handle_command, "dss",
	       playerno, player[playerno].name, command );
}

