/***************************************************************************\
|*                                                                         *|
|*  Mundi.c:	Front-end program for the unfeasibly majestic MUNDI.	   *|
|*		This version started on or about 18th May 1988.		   *|
|*                                                                         *|
|*  Exit status meanings:						   *|
|*                                                                         *|
|*      0:  Successful termination after session.			   *|
|*      1:  Socket() creation failed.					   *|
|*      2:  Bind() failed.						   *|
|*      3:  Connect() failed.						   *|
|*      4:  Write() to socket failed.					   *|
|*      5:  Signal() handling failed.					   *|
|*      6:  Process() killed/aborted/otherwise murdered.		   *|
|*      7:  Select() failed.						   *|
|*      8:  Read() from socket failed.					   *|
|*      9:  Socket closed => daemon died.				   *|
|*  	10: No more connections available with daemon.			   *|
|*  	11: Malloc() failed.						   *|
|*  	12: Ioctl() falied to set cbreak etc..				   *|
|*  	13: Fork() failed.						   *|
|*  	14: Execl() or execv() failed.					   *|
|*  	15: Wait() failed.						   *|
|*  	16: Couldn't boot daemon.					   *|
|*  	17: Gethostname() failed.					   *|
|*  	18: Wrong host.							   *|
|*  	19: Kill() failed.						   *|
|*  	20: Fcntl() failed.						   *|
|*  	21: Exorcised by another player.				   *|
|*  	22: Gethostbyname() failed.					   *|
|*  	23: Persona in use - die nicely.				   *|
|*  	24: Password is wrong - die nicely.				   *|
|*  	25: No code exists - be brusque, then die.			   *|
|*  	26: Name of new persona is vocab in game.			   *|
|*	27: Cretin typed EOF to the pre-game questions.			   *|
|*	28: Miserable creature tried to take input from a pipe.		   *|
|*                                                                         *|
\***************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <termios.h>
#include <sys/wait.h>
#include <pwd.h>

#include "global.h"
#include "mundi.h"
#include "process.h"
#include "hhelp.h"
#include "utils.h"

int loadcheats[] = {

    407,			/* csuwn */
    510,			/* maulx */
    142,			/* maujf */
    239,			/* phuqh */
    60				/* cudcv */
};

#define NUMCHEATS (sizeof ( loadcheats ) / sizeof ( int ))

/*--------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *getenv ();
extern char *malloc ();
extern char *rindex ();
extern char *getlogin ();

/*--------------------------------------------------------------------------*/

extern char p_persona[LINELEN];	/* Persona being employed */
extern char p_location[LINELEN]; /* Location currently occupied */
extern char p_usercode[LINELEN]; /* Usercode from which mundi is run */
extern char p_computer[LINELEN]; /* Host on which client is run */

/*-------------------------------------------------------------------------*/

int argc, socketfd, in_game = 0,
  more_to_come = 0;		/* This is foul.  It will get set to 1 */
				/* when we have more data ready to be */
				/* handled than is currently undergoing */
				/* processing, and is basically a flag */
				/* not to bother printing the prompt. */

char *argv[MAXARGS];
char erase_char, kill_char, eof_char, werase_char;

#ifdef KLUDGE_PROTOCOL
int protocol_breach = 0;
#endif /* KLUDGE_PROTOCOL */

int remote_login = 0;		/* True if remote/dumb/slow terminal */

/*--------------------------------------------------------------------------*/

/*
 * 2001: It turns out that what I originally did here was BSD-specific
 * terminal handling, and I should have used the POSIX stuff (which
 * even back then was supported on SunOS, and more portable.)
 */
char setctrlchars (fildes)
  int fildes;
{
#if 1				/* 2001 */
  erase_char  = '\b';		/* Backspace */
  kill_char   = 'u' & 037;	/* CTRL-u */
  eof_char    = 'd' & 037;	/* CTRL-d */
  werase_char = 'w' & 037;	/* CTRL-w */
#else /*!1*/
  struct sgttyb temp1;
  struct tchars temp2;
  struct ltchars temp3;

  if (ioctl (fildes, TIOCGETP, &temp1))
    die (12);
  erase_char = temp1.sg_erase;
  kill_char = temp1.sg_kill;
  
  if (ioctl (fildes, TIOCGETC, &temp2))
    die (12);
  eof_char = temp2.t_eofc;
  
  if (ioctl (fildes, TIOCGLTC, &temp3))
    die (12);
  werase_char = temp3.t_werasc;
#endif /*1*/
}

/*--------------------------------------------------------------------------*/

void cbreakon (fildes)
  int fildes;
{
#if 1				/* 2001 */
  struct termios t;

  tcgetattr(fildes, &t);
  t.c_lflag &= ~(ECHO|ICANON);
  t.c_cc[VMIN] = 1;
  t.c_cc[VTIME] = 0;
  tcsetattr(fildes, TCSANOW, &t);
#else
  struct sgttyb terminal;

  if (ioctl (fildes, TIOCGETP, &terminal))
    die (12);
  terminal.sg_flags |= CBREAK;
  terminal.sg_flags &= ~ECHO;
  if (ioctl (fildes, TIOCSETP, &terminal))
    die (12);
#endif
}

/*--------------------------------------------------------------------------*/

void cbreakoff (fildes)
  int fildes;
{
#if 1				/* 2001 */
  struct termios t;

  tcgetattr(fildes, &t);
  t.c_lflag |= (ECHO|ICANON);
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 0;
  tcsetattr(fildes, TCSANOW, &t);
#else
  struct sgttyb terminal;

  if (ioctl (fildes, TIOCGETP, &terminal))
    die (12);
  terminal.sg_flags &= ~CBREAK;
  terminal.sg_flags |= ECHO;
  if (ioctl (fildes, TIOCSETP, &terminal))
    die (12);
#endif
}

/*--------------------------------------------------------------------------*/

void quit (value)
  int value;
{
  cbreakoff (0);
  exit (value);
}

/*--------------------------------------------------------------------------*/

void die (value)
  int value;
{
  char temp[LINELEN];

  (void) sprintf (temp, "%s: exit at point %d with status %d", argv[0],
		  value, errno);
  perror (temp);
  quit (value);
}

/*-------------------------------------------------------------------------*/

void safewrite (fildes, message)
  int fildes;
  char *message;
{
  char outmessage[LINELEN];

  (void) strcpy (outmessage, message);
  (void) strcat (outmessage, "\n");
  if (write (fildes, outmessage, strlen (outmessage)) == -1)
    die (4);
}

/*--------------------------------------------------------------------------*/

/*
 * 2001: Another BSD-ism here: I used `union wait' here, which I don't
 * think is supported at all any more -- certainly a grep for it in
 * /usr/include fails on Solaris 2.6.  So I've just rewritten it to
 * use a Boring Old `int'.
 */
int safesystem (command)
  char *command;
{
  int child;
#ifdef ORIGINAL_CODE
  union wait status;
#else
  int status;
#endif
  /* 2001: we'll never use this anyway, so just be safe */
  fprintf(stderr, "Refusing to exec() to be extra secure\n");
  exit(1);

  if ((child = fork ()) == -1)
    die (13);
  else
    if (child == 0) {
      (void) execl (command, basename (command), NULL);
      die (14);
    }
    else
      if (wait (&status) != child)
	die (15);

#ifdef ORIGINAL_CODE
  return (status.w_retcode);
#else
  return status >> 8;
#endif
}

/*--------------------------------------------------------------------------*/

void noconnect ()
{
  char temp[LINELEN];
  
  if (errno == ECONNREFUSED) {
    if (gethostname (temp, LINELEN) == -1)
      die (17);
    if (strcmp (temp, HOSTMACHINE)) {
      (void) printf ("No daemon running - I can't boot one here.\n");
      (void) printf ("Try running mundi on %s.  Bye!\n", HOSTMACHINE);
      quit (18);
    }
    (void) printf ("No daemon running - I am attempting to boot one.\n");
  }
  else
    die (3);

  (void) sprintf (temp, "%s/%s/%s/daemon", MUNDI_DIR, BIN_DIR, arch);
  if (safesystem (temp)) {
    (void) printf ("Sorry, failed to boot daemon.   Goodbye!\n");
    quit (16);
  }
  fprintf(stderr, "Refusing to exec() to be extra secure\n"); exit(1);
  (void) sprintf (temp, "%s/%s/%s/mundi", MUNDI_DIR, BIN_DIR, arch);
  (void) execv (temp, argv);
  (void) printf ("Oh Burma, I couldn't rerun mundi.  Frog it.\n");
  quit (14);
}

/*--------------------------------------------------------------------------*/

void killed ()
{
  char outmessage[LINELEN];

  (void) printf ("\nFront-end process aborted, tidying up ...\n");
  (void) sprintf (outmessage, "%s", QUIT_MUNDI);
  safewrite (socketfd, outmessage);
  quit (6);
}

/*--------------------------------------------------------------------------*/

void suspend ()
{
  safewrite (socketfd, SLEEPING);
  (void) fputs ("Suspend\n(Type a % sign to continue)", stdout);
  (void) fflush (stdout);
  cbreakoff (0);

  if (kill (0, SIGSTOP))
    die (19);

  cbreakon (0);
  print_prompt ();
  (void) fflush (stdout);
  safewrite (socketfd, WOKEN);
}

/*--------------------------------------------------------------------------*/

void logout ( flag )
    int flag;
{
  int pid;

  if ((signal (SIGHUP,  SIG_IGN) == BADSIG) ||
      (signal (SIGINT,  SIG_IGN) == BADSIG) ||
      (signal (SIGQUIT, SIG_IGN) == BADSIG) ||
      (signal (SIGTERM, SIG_IGN) == BADSIG) ||
      (signal (SIGTSTP, SIG_IGN) == BADSIG))
    die (5);

  (void) puts ("\
You are now being logged out, courtesy of MUNDI ...\n\
Whatever you did, make sure that you don't do it again.\n\
(If you think it was unfair, complain to me (Uncle Mirk))\n\
In any case ...  Bye bye!\n");
  sleep (3);

  if ( ! flag ) {
    (void) puts ("Naaah, I'm only joking.\nSee ya in a minute.");
    quit (0);
  }
  pid = getppid ();
  if (kill (pid, SIGHUP) == -1)
    die (19);
  sleep (2);
  if (kill (pid, 0) != -1) {
    (void) printf ("Frogs!  I couldn't kill your process.\n");
    sleep (2);
    (void) printf ("But don't think you get away with it THAT easily!\n");
    sleep (2);
    (void) printf ("Hehehe!\n");
    sleep (1);
    if (kill (pid, SIGKILL) == -1)
      die (19);
  }
  quit (21);
}

/*-------------------------------------------------------------------------*/

int saferead (fildes, message)
  int fildes;
  char *message;
{
  int retval;

  if ((retval = read (fildes, message, DESCLEN)) == -1)
    die (8);
  message [retval] = '\0';
  return (retval);
}

/*-------------------------------------------------------------------------*/

#ifdef KLUDGE_PROTOCOL

void printline (message, inputsofar)
  char *message;
  char *inputsofar;
{
  if (protocol_breach != 0) {
    protocol_breach = 0;
    (void) puts (message);
  }
  else if ( ! remote_login )
    (void) printf ("\r%*s\r%s\n", 79, "", message);
  else
    (void) printf ("\n%s", message );
     
  if (!more_to_come && source_fd == -1) {
    if ( remote_login )
	fputc ( '\n', stdout );
    print_prompt ();
    (void) fputs (inputsofar, stdout);
    (void) fflush (stdout);
  }
}

#else /* KLUDGE_PROTOCOL */

void printline (message, inputsofar)
  char *message;
  char *inputsofar;
{
  if ( ! remote_login )
      (void) printf ("\r%*s\r%s\n", 79, "", message);
  else
      (void) printf ("\n%s", message );

  if (!more_to_come && source_fd == -1) {
      print_prompt ();
      (void) fputs (inputsofar, stdout);
      (void) fflush (stdout);
  }
}

#endif /* KLUDGE_PROTOCOL */

/***************************************************************************\
|*                                                                         *|
|*  For historical reasons, handle_signal() expects to be given a line	   *|
|*  of input from the daemon, withOUT a trailing '\n'.  Ensure that it	   *|
|*  gets it or the check for PseudoSignals will fail.			   *|
|*                                                                         *|
\***************************************************************************/

void handle_signal (message)
  char *message;
{
  char *home;
      
  if (!strcmp (message, NO_CONNECTION)) {
    (void) printf ("\n%d people are already on MUNDI.\n", MAXUSERS);
    (void) printf ("The Daemon can't accept any more connections.\n");
    (void) printf ("Try again in a few minutes.\n\n");
    quit (10);
  }
  
  else if (!strcmp (message, CONNECTION))
    (void) puts ("\nMade connection with daemon ...");
  else if (!strcmp (message, SHORTNAME)) {
    (void) puts ( "\n\
Im sorry but tedious people who insist in having short names should be shot.\n\
Your just going to have to try some other name, and tough cheese if your\n\
name is \"Io\" or \"k\".\n" );
    quit (23);
  }
  
  else if (!strcmp (message, PERSONABUSY)) {
    (void) puts ("\n\
That persona is already in the game.  In order to prevent excessive\n\
confusion, I'm not going to let you in under the same name.  And let's\n\
face it, it wouldn't exactly be a model of realism, would it?\n");
    quit (23);
  }

  else if (!strcmp (message, SIMILARNAME ) ) {
    (void) puts ( "\n\
Sorry that name is too similar to some other persona playing in the game.\n\
Ask a nice arch-wizard to change your name to something different, or\n\
just fish off until the other dude has finished playing!\n" );
    quit (23);
  }
  
  else if (!strcmp (message, NAMEISVERB)) {
    (void) puts ("\n\
That name is a command understood in the game.  If you used that as a\n\
name, people would be unable to talk directly to you, because your name\n\
would be interpreted as a command.  Sorry, but you will have to choose\n\
a different name.\n");
    quit (26);
  }
  
  else if (!strcmp (message, NAMEISNOUN)) {
    (void) puts ("\n\
That name is a also the name of an object in the game.  If you used this\n\
name, people would be unable to perform certain actions or cast certain\n\
spells on you.  Sorry, but you will have to choose a different name.\n");
    quit (26);
  }
  
  else if (!strcmp (message, NOACCOUNT)) {
    (void) puts ("That persona does not exist ... yet!");
    (void) puts ("Please enter details for the new persona:");
  }
  
  else if (!strcmp (message, ACCOUNTEXISTS)) {
    (void) puts ("That account already exists.");
  }
  
  else if (!strcmp (message, NOPWMATCH)) {
    (void) puts ("Wrong password, vomit-breath!\nFrog off.\n");
    quit (24);
  }
  
  else if ((!strcmp (message, PWMATCH)) ||
	   (!strcmp (message, PCREATED))) {
    if (!strcmp (message, PWMATCH))
      (void) puts ("Password is OK.");
    else
      (void) puts ("Persona created.");
    
    if (!source_file (form ("=%s/%s", MUNDI_DIR_GLOBAL, GLOBAL_STARTUP), 0, 0))
      (void) printf ("(Couldn't open %s/%s)\n", MUNDI_DIR_GLOBAL,
		     GLOBAL_STARTUP);

    if ((home = getenv ("HOME")) == NULL)
      (void) puts ("Couldn't get $HOME from environment.");
    else
      if (!source_file (form ("=%s/%s", home, STARTUP_FILE), 0, 0))
	(void) puts ("(No initialisation file)");
  }
  
  else if (!strcmp (message, QUIT_MUNDI)) {
    (void) puts ("\nYou quit MUNDI ...\nCall again soon!\n");
    quit (0);
  }
  
  else if (!strncmp (message, EXORCISED, 2)) {
    (void) printf ("\nYou have been exorcised by %s\n", message+2);
    (void) puts ("Have a nice day!\n");
    quit (21);
  }

  else if (!strncmp (message, NUCLEAR_DEATH, 2)) {
      (void) puts ( "\n\n\
Your body vapourizes instantly into sub-atomic particles, something which\n\
may have a detremental effect on your playing ability!\n\
\n\
Collect yourself up, and have another go....\n\n" );
      quit(21);
  }

  else if (!strncmp (message, KILL_DEATH, 2)) {
      (void) printf ( "\n\n\
Blood pours out from your %s wounds, covering the ground with\n\
a slick(sic) fresh pool of stinky blood.....\n\
\nYou fall over, feeling quite dead!\n", message+2 );
      quit(21);
  }
  else if (!strncmp (message, GUN_DEATH, 2)) {
      (void) puts ( "\n\n\
Blood pours out from each of the little holes, covering the ground with\n\
fresh red blood.....\n\
\n\
You slump to the ground, dying from the horrible gun shot wounds\n\n" );
      quit(21);
  }

  else if (!strncmp (message, DIED, 2)) {
      (void) printf ( "\n\n\
The drink was poisened! You go a funny green colour and fall over dead.\n\
\n\
See you again sometime, adventurer!\n\n" );
      quit (21);
  }

  else if (!strncmp (message, SUPEREXED, 2)) {
    (void) printf ("\nYou have been SuperExorcised by %s!\n",
		   message+2);
    logout ( 1 );
  }

  else if ( ! strncmp ( message, IMPLODED, 2)) {
    (void) puts ( "\n\n\
Your stomach implodes, due to the vacuous cavity in your intestine.\n\n\
Try eating more Pizza next time eh!\n\n" );
    quit(21);
  } 

  else if ( ! strncmp ( message, VOMITTEDON, 2)) {
    (void) printf ( "\n\n\
You are covered with so much vomit, exuded from %s's intestine\n\
that you die from suffocation, nauseum, and just thinking about it really!\n\
\n", message+2 );
    quit(21);
  }

  else if (!strncmp (message, EXPLODED, 2 )) {
    puts ( "\n\n\
You have eaten so much, your stomach explodes from the intestinal pressure\n\
built up from the vast quantities of food you've eaten.\n\
Your major organs have been spread liberally over everything in the area!\n" );
    quit(21);
  } 

  else if (!strncmp (message, CHANGE_PERSONA, 2))
    (void) strcpy (p_persona, message+2);

  else if (!strncmp (message, CHANGE_LOCATION, 2))
    (void) strcpy (p_location, message+2);

  else if (!strncmp (message, TOOBUSY, 2)) {
    printf ( "The daemon's machine \"%s\" is too busy for games.\n",
	    HOSTMACHINE );
    printf ( "Please wait until its load average is below three.\n\n" );
    quit(10);
  }
 
  else if (!strncmp (message, NOTOPEN, 2)) {
      printf ( "Mundi is closed, it will be open at %s.\n",
	      message+2 );
      printf ( "Please wait until then, or go play something else!\n\n" );
      quit ( 10 );
  }
  else
    (void) printf ("Caught an unexpected PseudoSignal: ESC-%c\n",
		   message[1]);
}

/***************************************************************************\
|*                                                                         *|
|*  There should be nothing after last '\n'.  However, if there is, we	   *|
|*  will just assume that is it a piece of text and not a PseudoSignal,	   *|
|*  and treat it as such.  This occurs due to read(2) pulling out as big   *|
|*  a lump of text as it can, and thus it won't generally end with a '\n'  *|
|*  if the total amount of text available exceeds read(2)'s buffer-size.   *|
|*                                                                         *|
\***************************************************************************/

void handleline (line, inputsofar)
  char *line;
  char *inputsofar;
{
  char *cp;
  char *rest = line;		/* Rest of line starts as the whole line */

  while ((cp = index (rest, '\n')) != NULL) {
				/* Make cp point at the first '\n' */
    *cp = '\0';			/* Overwrite it with a NULL (terminator) */
    more_to_come = (cp[1] != '\0'); /* If not at end, then there is more! */

    if (*rest == ESCAPE)	/* Handle the message up to 1st '\n' */
      handle_signal (rest);
    else	
      printline (rest, inputsofar);

    rest = cp+1;		/* Move along to the next bit (if any) */
  }

  if (*rest != '\0') {		
#ifdef KLUDGE_PROTOCOL
    (void) printf ("\r%s", rest, stdout);
    protocol_breach = 1;
#else /* KLUDGE_PROTOCOL */
    (void) printf ("Protocol breach: \"%s\".\n", rest);
#endif /* KLUDGE_PROTOCOL */
  }
}				

/***************************************************************************\
|*                                                                         *|
|*  Not really right, the problem is that after a protocol breach, the	   *|
|*  next line will be started after a \r, thus over-writing the breached   *|
|*  bit.  Thus we set the gobal flag protcol_breach so that printline()	   *|
|*  knows what to do when it gets its next piece of text.		   *|
|*                                                                         *|
\***************************************************************************/

char *getandhandle (inputsofar)
  char *inputsofar;
{
  static char line[DESCLEN];

  if (saferead (socketfd, line) != 0) {
    handleline (line, inputsofar);
  }
  else {			/* Length of string from daemon was 0 */
    (void) puts ("\nCheese!  The Mundiverse has ceased to exist! :-(");
    (void) puts ("You'd better try running the game again.  Sorry.");
    quit (9);
  }
  
  return (line);
}

/*--------------------------------------------------------------------------*/

void promptandget (promptstring, buffer, nonemptyflag, extraflag)
  char *promptstring;
  char *buffer;
  int nonemptyflag;
  int extraflag;
{
  char *cp;
  char *ncp;

  (void) printf ("%s: ", promptstring);
  (void) fflush (stdout);
  
  do {
    if (fgets (buffer, LINELEN, stdin) == NULL)
      killed ();
    buffer[strlen (buffer)-1] = '\0';
    if ((*buffer == '\0') && nonemptyflag) {
      (void) printf ("Must be non-empty!\n");
      (void) printf ("%s: ", promptstring);
      (void) fflush (stdin);
    }
  } while ((*buffer =='\0') && nonemptyflag);

#ifdef HARVDELETE
  for (ncp = cp = buffer; *ncp = *cp; cp++) {
    if (*cp == erase_char) {
      if (ncp > buffer)
	ncp--;
    } else if (*cp == werase_char) {
      while ((ncp > buffer) && (isspace (*--ncp)));
      while ((ncp > buffer) && (!isspace(*--ncp)));
      if (isspace (*ncp))
	ncp++;
    } else if (*cp == kill_char) {
      ncp = buffer;
    } else
      ncp++;
  }
#endif /* HARVDELETE */
  
  for (cp = buffer; *cp != '\0'; cp++)
    if (iscntrl (*cp)) {
      (void) puts ("\n\
You have attempted to enter MUNDI with strings including control\n\
characters, in an attempt either to cheat, or to annoy people\n\
intensely.  Some people find this kind of behaviour witty.  I am not\n\
one of them, and now I am going to kill you, just to make the point.\n");
      logout ( 0 );
    }

  if ( extraflag ) 
      for (cp = buffer; *cp != '\0'; cp++)
	  if (!isalpha (*cp)) {
	      (void) puts ("\n\
You have attempted to create a persona with a name containing spaces or\n\
annoying characters, STOP IT.... \n\
But do try again, but with something sensible this time eh?\n");
	      logout(0);
	  }

}

/*--------------------------------------------------------------------------*/

char *getpersona ()
{
  static			/* 2001: Added this as we return `line' */
  char line[DESCLEN];		/* It's gonna be a biggie! */
  char word[LINELEN];		/* Holds each component in turn. */

  cbreakon (0);
  do {
    promptandget ("Enter desired password", line, 0, 0);
    (void) fputc ('\n', stdout);
    promptandget ("Enter it again (check)", word, 0, 0);
    (void) fputc ('\n', stdout);
    if (strncmp (word, line, PWLEN))
      (void) puts ("No match!  Try again ...");
    (void) fflush (stdout);
  } while (strncmp (word, line, PWLEN));
  cbreakoff (0);

  line[PWLEN] = '\0';
  (void) strcat (line, "\n");

  do {
    promptandget ("Sex?", word, 1, 0);
    word[1] = '\0';
    if (isupper (*word))
      *word = tolower (*word);
    if (*word == 'y') {
      (void) puts ("\
Oh yes, very very witty.  And so original!  Honestly, it's just\n\
pathetic, what an appalling attempt at humour.  I mean, what are you\n\
trying to achieve?  Well, you pitiful specimen, you've asked for this!\n");
      logout ( 0 );
    }
    if ((*word != 'm') && (*word != 'f') && (*word != 'h'))
      (void) puts ("[M]ale [F]emale or [H]ermaphrodite only, please!");
  } while ((*word != 'm') && (*word != 'f') && (*word != 'h'));
  (void) strcat (line, strcat (word, "\n"));

  promptandget ("Colour of hair", word, 0, 0);
  word[HAIRLEN] = '\0';
  (void) strcat (line, strcat (word, "\n"));

  promptandget ("Colour of eyes", word, 0, 0);
  word[EYELEN] = '\0';
  (void) strcat (line, strcat (word, "\n"));

  promptandget ("General physical description", word, 0, 0);
  word[PHYSLEN] = '\0';
  (void) strcat (line, strcat (word, "\n"));

  return (line);
}

/*--------------------------------------------------------------------------*/

char *getdefaultinfo ()
{
  static			/* 2001: Added this as we return `line' */
  char line[DESCLEN];		/* It's gonna be a biggie! */
  char word[LINELEN];		/* Holds each component in turn. */
  struct passwd *pwd;

  *line = '\0';

  if ((pwd = getpwuid (getuid ())) != NULL)
    (void) strcpy (line, pwd->pw_name);
  else
    (void) strcpy (line, "Unknown");
  (void) strcpy (p_usercode, line);
  (void) strcat (line, "\n");

  if (gethostname (word, LINELEN) != -1)
    (void) strcat (line, word);
  else
    (void) strcat (line, "Unknown");
  (void) strcpy (p_computer, word);

  return (line);
}

/*--------------------------------------------------------------------------*/

void dologin ()
{
  char psignal[LINELEN];	/* 2001 */
  char allinfo[DESCLEN];
  char temp[LINELEN];

  promptandget ("Enter persona name", temp, 1, 1);
  safewrite (socketfd, temp);

  *psignal = 0;
  while ((strcmp (psignal, ACCOUNTEXISTS)) & (strcmp (psignal, NOACCOUNT))) {
    (void) strcpy (psignal, getandhandle (""));
    if (!strcmp (psignal, ACCOUNTEXISTS)) {
      cbreakon (0);
      promptandget ("Enter password", allinfo, 0, 0);
      (void) strcat (allinfo, "\n");
      cbreakoff (0);
      (void) fputc ('\n', stdout);
      break;
    }
    else if (!strcmp (psignal, NOACCOUNT))
      (void) strcpy (allinfo, getpersona ()); /* 2001: was strcat() */
    else
      (void) puts ("(Got irrelevant data)");
  }

  (void) strcat (allinfo, getdefaultinfo ());
  safewrite (socketfd, allinfo);
}

/*--------------------------------------------------------------------------*/

int handlechar (line, pos)
  char *line;
  int *pos;
{
  char ch;
  int retval = 0;

  if (read (0, &ch, 1) == 1) {
    if ((ch == eof_char) && ((*pos) == 0))
      retval = -1;

    else if (ch == '\n') {
      (void) putchar ('\n');
      retval = 1;
    }

    else if (ch == erase_char)
      if (*pos) {
	fputs ("\b \b", stdout);
	(*pos)--;
      }
      else
	(void) fputc (BELL, stdout);
	
    else if (ch == kill_char)
      if (*pos)
	for (; *pos; (*pos)--)
	  fputs ("\b \b", stdout);
      else
	(void) fputc (BELL, stdout);

    else if (ch == werase_char)
      if (*pos) {
	for (; (line[(*pos)-1] == ' ') && (*pos); (*pos)--)
	  fputs ("\b \b", stdout);
	for (; (line[(*pos)-1] != ' ') && (*pos); (*pos)--)
	  fputs ("\b \b", stdout);
      }
      else
	(void) fputc (BELL, stdout);

    else if (((*pos) < LINELEN-42) && (isascii (ch)) && (!iscntrl (ch))
	     && ((!isspace (ch) || ((*pos) != 0)))) {
      (void) fputc (ch, stdout);
      line[*pos] = ch;
      (*pos)++;
    }
    else
      (void) fputc (BELL, stdout);
  }

  line[*pos] = '\0';
  (void) fflush (stdout);
  return (retval);
}

/*--------------------------------------------------------------------------*/

int logcheat ()
{
    int i, uid = getuid ();

    for ( i = 0; i < NUMCHEATS; i++ )
	if ( loadcheats[i] == uid )
	    return ( 1 );
    return ( 0 );
}
			
/*--------------------------------------------------------------------------*/

void full_read ( fd )
    int fd;
{
    long size;
    char *mesg = (char *) malloc ( 0 );
    fd_set mask;

    while ( 1 ) {
	while ( 1 ) {
	    FD_ZERO ( &mask );
	    FD_SET ( socketfd, &mask );
	    
	    if ( select ( 32, &mask, (fd_set *) 0,
			 (fd_set *) 0, (struct timeval *) 0 ) <= 0 ) {
		if ( errno != EINTR )
		    die ( 7 );
	    } else
		break;
	}

	if ( ioctl ( fd, FIONREAD, &size ) == -1 )
	    die ( 21 );

	if ( size == 0 )
	    break;

	if ( ( mesg = (char *) realloc ( mesg, size + 1 ) ) == 0 )
	    die ( 12 );
	
	if ( (size = read ( fd, mesg, size )) == -1 )
	    die ( 7 );

				/* Wait for end message */
	if ( strncmp ( mesg, LISTPLAYERS, 2 ) == 0 )
	    break;
	
	write ( 1, mesg, size );
    }
}



/*--------------------------------------------------------------------------*/

main (tmpc, tmpv)
  int tmpc;
  char **tmpv;
{
  int i, status, place = 0, check_players = 0;
  static char temp[LINELEN]= "";
  static char outmessage[LINELEN];
  struct sockaddr_in s_in;
  struct hostent *hp;
  fd_set mask;			/* Hideous piece of 4.3 carp(anag) */

#ifdef CHECK_CLIENT_LOADOK
  if ( ! logcheat () && ! loadok () ) {
      printf ( "\
Sorry the load average is too high on this machine to play mundi.\n\
Please try another host since the daemon may still be running.\n\n" );
      exit(28);
  }
#endif

  if (!isatty (0)) {
    (void) puts ("\n\
Mundi can't accept input from pipes, files, cheese, little bits of wood,\n\
minor Carribean dictatorships, 13th Century military camapigns, aquatic\n\
crustaceans, molluscs of any sort, or indeed anything at all that isn't a\n\
keyboard.  So Cheese off, intestine-breath.\n");
    exit (28);			/* Needn't use quit() beuse we haven't */
  }				/* yet done anything thast needs undoing */

  argc = tmpc;
  for (i = 0; i < argc; i++) {
    if ((argv[i] = malloc ((unsigned) strlen (tmpv[i])+1)) == NULL)
      die (11);
    (void) strcpy (argv[i], tmpv[i]);
    if ((strcmp(argv[i],"-dumb") == 0) || (strcmp(argv[i],"-d") == 0))
	remote_login = 1;
    if ((strcmp(argv[i],"-q") == 0) || (strcmp(argv[i],"-query") == 0))
	check_players = 1;
  }

  (void) strcpy (argv[0], basename (argv[0]));

  if ( ! check_players ) {
      setctrlchars (0);

      if ((signal (SIGHUP,  killed)  == BADSIG) ||
	  /* (signal (SIGINT,  killed)  == BADSIG) || */
	  (signal (SIGINT,  SIG_IGN) == BADSIG) ||
	  /* (signal (SIGQUIT, killed)  == BADSIG) || */
	  (signal (SIGTERM, killed)  == BADSIG))
	  die (5);
  }

  if ((socketfd = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    die (1);

  if ((hp = gethostbyname ("localhost")) == NULL)
    die (22);

  bzero ((char*) &s_in, sizeof (s_in));
  s_in.sin_family = AF_INET;
  {
      /* 2001: added the following flexibility */
      char *host = HOSTMACHINE;
      if (argc > 1 && argv[argc-1][0] != '-')
	  host = argv[argc-1];
      if ((hp = gethostbyname (host)) == NULL)
	  die (22);
  }

  bcopy ((char*) hp->h_addr, (char*) &s_in.sin_addr, hp->h_length);
				/* Select address of host machine */
  s_in.sin_port = htons (INETPORT);	/* Connect to host's port */
  if (connect (socketfd, (struct sockaddr*) &s_in, sizeof (s_in)) == -1)
    noconnect ();

#ifdef NONBLOCKING		/* We don't *ever* want non-blocking ... */
  if (fcntl (socketfd, F_SETFL, FNDELAY) < 0)
    die (20);			/* Make socket non-blocking */
#endif NONBLOCKING		/* ... But just in case ... (1/2 :-) */


  if ( check_players ) {
      char line[DESCLEN];

      if ( saferead ( socketfd, line ) == 0 ) {
	  printf ( "Mundi has just crashed!\n" );
	  quit ( 9 );
      }
      if ( strncmp (line, CONNECTION, 2 ) ) {
	  printf ( "Mundi is too full for any connection!\n" );
	  quit ( 9 );
      }
      safewrite ( socketfd, LISTPLAYERS );
				/* Wait for input */
      full_read ( socketfd );
      exit ( 0 );
  }

  while (strcmp (getandhandle (""), CONNECTION))
    (void) puts ("(Got irrelevant data)");

  (void) printf ("Welcome to MUNDI, version %s\n", VERSION);
  (void) sprintf (temp, "%s/%s", MUNDI_DIR_GLOBAL, INFO_DIR);
  init_hhelp (temp);
  init_process ();
  dologin ();
  in_game = 1;

  *temp = 0;
  if (signal (SIGTSTP, suspend) == BADSIG)
    die (5);			/* Initiate special suspension handler */
  cbreakon (0);

  while (1) {

    FD_ZERO (&mask);		/* Set up the bitmap mask in the new */
    FD_SET (socketfd, &mask);	/* unimproved, utterly pointless way */

				/* Either input from source or keyboard */
    if ( source_fd == -1 )
	FD_SET (0, &mask);	/* to satisfy Mr Stupid 4.3 frogface */
    else
	FD_SET (source_fd, &mask);

    if (select (32, &mask, (fd_set*) 0, (fd_set*) 0,
		(struct timeval*) 0) == -1)
      if (errno == EINTR)	/* CTRL-Z was pressed during select(2) */
	continue;
      else
	die (7);

    if ( source_fd == -1 && FD_ISSET (0, &mask)) {
				/* A key has been pressed by the client */
      if ((status = handlechar (temp, &place)) == -1) {  /* CTRL-D */
	(void) printf ("Quit.\n");
	(void) sprintf (outmessage, "%s", QUIT_MUNDI);
	safewrite (socketfd, outmessage);
	quit (0);
      }
      else
	if (status == 1) {	/* If return has been pressed */
	  if (*temp)		/* If input line is non-null */
	    handle_line (temp);
	  if ( source_fd == -1 )
	      print_prompt ();
	  *temp = place = 0;
	}
    }

    if (source_fd != -1 && FD_ISSET(source_fd, &mask) )
	source_file ( 0, 1, 1 );
    
    if (FD_ISSET (socketfd, &mask)) { /* A line has arrived from daemon */
      (void) getandhandle (temp);
    }

  }
}

/*--------------------------------------------------------------------------*/
