
/***************************************************************************\
|*                                                                         *|
|*  Persona.c:	Part of the server program for MUNDI - handle personas.	   *|
|*		This version started on or about 26th May 1988.		   *|
|*                                                                         *|
\***************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>

#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "handle.h"
#include "look.h"
#include "utils.h"

/*--------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *malloc ();
extern char *realloc ();
extern char *ctime ();
extern off_t lseek ();
extern time_t time ();

/*-------------------------------------------------------------------------*/

int passwd_fd;
struct persona player[MAXUSERS];
struct persona *persona_list;

/*-------------------------------------------------------------------------*/

struct level sc_levels[SC_LEVELS] = {
  { 0,		"scum"		},
  { 100,	"low-life"	},
  { 200,	"pleb"		},
  { 500,	"peasant"	},
  { 1000,	"pretender"	},
  { 2000,	"contender"	},
  { 5000,	"survivor"	},
  { 10000,	"adventurer"	},
  { 20000,	"master"	},
  { 30000,	"over-master"	},
  { 50000,      "bossman"       },
  { 80000,      "lord"          },
  { 120000,     "illusionist"   },
  { 180000,     "mangician"     },
  { 300000,     "Neophyte Wizard"         },
  { 600000,     "Apprentice Wizard"       },
  { 1000000,    "Hero Wizard"             },
  { 2000000,    "Champion Wizard"         },
  { 5000000,    "Warrior Wizard"          },
  { 10000000,   "Exceller Wizard"         },
  { 20000000,   "Deadly Wizard"           },
  { 40000000,   "Frogging Deadly Wizard"  },
  { 100000000,  "Psuedo Arch-Wizard"      },
  { 500000000,  "ARCH Wizard"             },
  { 1000000000, "META Wizard"             }
};

/*-------------------------------------------------------------------------*/

struct level pu_levels[PU_LEVELS] = {
  { -50000,	"evil"		},
  { -20000,	"demonic"	},
  { -10000,	"malevolent"	},
  { -5000,	"depraved"	},
  { -2000,	"sadistic"	},
  { -1000,	"vindictive"	},
  { -500,	"vicious"	},
  { -200,	"nasty"		},
  { -100,	"unpleasant"	},
  { -50,	"selfish"	},
  { 0,		"neutral"	},
  { 50,		"helpful"	},
  { 100,	"friendly"	},
  { 200,	"pleasant"	},
  { 500,	"nice"		},
  { 1000,	"philanthropic"	},
  { 2000,	"benevolent"	},
  { 5000,	"good"		},
  { 10000,	"luverly"	},
  { 20000,	"holy"		},
  { 50000,	"saintly"	}
};

/*-------------------------------------------------------------------------*/

struct level fa_levels[FA_LEVELS] = {
  { 0,		"wafer-thin"	},
  { 1,		"emaciated"	},
  { 5,		"anorexic"	},
  { 10,		"malnourished"	},
  { 20,		"skinny"	},
  { 50,		"slim"		},
  { 100,	"well-built"	},
  { 200,	"overweight"	},
  { 300,	"tubby"		},
  { 500,	"large"		},
  { 700,	"fat"		},
  { 1000,	"obese"		},
  { 2000,	"gross"		},
  { 5000,	"monstrous"	},
  { 10000,	"gargantuan"	}
};

/*--------------------------------------------------------------------------*/

int compute_level ( table, score, maxlevel )
struct level *table;
int           score;
int	      maxlevel;
{
    int ret = 0;

    while ( table[ret].minscore <= score ) {
	if ( ret++ == (maxlevel-1) )
	    break;
    }
    return ( ret ? (ret - 1) : 0 );
}

/****************************************************************************\
  Calculates the persons stats for the new level
\****************************************************************************/

void calculate_stats( playerno, level )
    int playerno, level;
{
    int range, basetotal, addtotal;

    player[playerno].level = level;

    range = ( ( SC_LEVELS + 5 - level ) * 3 );
				/* (3 * Sum{SC_LEVELS+5-i | 0 < i <= level} */
				/* + range/2) / 6 + 16 */
    basetotal = ( ( ( 3 * (level+1) ) *
		    ( 5 + SC_LEVELS - ( level - 1)/2 ) )
		 + range/2 ) / 6 + 16;


    if ( ARCHMODEP(playerno) ) {
	player[playerno].strength     = 1000;
	player[playerno].intelligence = 1000;
	player[playerno].agility      = 1000;
	player[playerno].hoopiness    = 1000;
	player[playerno].intcap       = 1000;
	player[playerno].luck         = 1000;
    } else { 

	do {
	    addtotal  = range / 4;
	    addtotal -= player[playerno].strength     = ( random() % range/12 );
	    addtotal -= player[playerno].intelligence = ( random() % range/12 );
	    addtotal -= player[playerno].agility      = ( random() % range/12 );
	    addtotal -= player[playerno].hoopiness    = ( random() % range/12 );
	    addtotal -= player[playerno].intcap       = ( random() % range/12 );
	} while ( ( addtotal < 0 ) || ( addtotal > range/12 ) );
	
	player[playerno].strength     += basetotal;
	player[playerno].intelligence += basetotal;
	player[playerno].agility      += basetotal;
	player[playerno].hoopiness    += basetotal;
	player[playerno].intcap       += basetotal;
	player[playerno].luck          = basetotal + addtotal;
	
    }
}

/***************************************************************************\
|*									   *|
|*  I originally wrote this as: "IFS='\n'; getnextword(); IFS=' ';"	   *|
|*  This turns out to be a bad idea, since getnextword throws away all	   *|
|*  whitespace, including multiple consecutives, so any null component	   *|
|*  of the sigon string will be lost.  I'm about to write getcomponent	   *|
|*  by basically ripping the getnextword() code out of utils.c, and	   *|
|*  poking it about a bit.  Redoing it like this also means that we	   *|
|*  get the elegance of not having to set and reset a global variable	   *|
|*  just for the benfit of this one function.				   *|
|*									   *|
\***************************************************************************/

void getcomponent (word, line)
  char *word;
  char *line;
{
  char *cp;
  int difference = line-word;
  
  for (cp = line; (*cp != '\n') && (*cp != '\0'); cp++)
    *(cp-difference) = *cp;	/* Copy first component to word */
  
  *(cp-difference) = '\0';	/* Terminate word with a '\0' */
  if (*cp == '\0') {
    *line = '\0';
    return;
  }
  
  cp++;				/* Skip past terminator */
  difference = cp-line;		/* Shunt up the rest of the line */
  while (*(cp-1) != '\0') {	/* to fill the gap left by word. */
    *(cp-difference) = *cp;
    cp++;
  }
}



/***************************************************************************\
|*                                                                         *|
|*  This routine will read the player's name from the socket, and check	   *|
|*  that is is not the name of someone who is already playing, that it	   *|
|*  is not a verb in the game, and that it is not a noun.  It will then	   *|
|*  decide whether the login name already exists in the pasword file, and  *|
|*  acts correctly based on this knowledge.  Dependant on this, it may	   *|
|*  send the follwing PseudoSignals back to the front-end program:	   *|
|*                                                                         *|
|*	case PERSONABUSY:                                                  *|
|*	case NAMEISVERB:                                                   *|
|*	case NAMEISNOUN:                                                   *|
|*	case ACCOUNTEXISTS:                                                *|
|*	case NOACCOUNT:                                                    *|
|*                                                                         *|
\***************************************************************************/

int verify (playerno, line)
    int playerno;
    char *line;
{
    int i = 176;			/* Why the cheese not, A? */
    int found_persona = 0;	/* Set to 1 if persona already exists */
    int temp_fd;

    line[strlen (line)-1] = '\0';
    if (strlen(line) > NAMELEN)	/* 2001: added test, prevent write-past-end */
      line[NAMELEN] = '\0';
    if (!strcmp (line, QUIT_MUNDI))
	return (-1);

    if (!strcmp (line, LISTPLAYERS)) {
				/* List players */
	int count = 0;

	for ( i = 0; i < MAXUSERS; i++ ) {
	    if ( player[i].fd != -1 && ISACTIVE(i) ) {
		if ( count++ == 0 )
		    send_reply ( playerno, "Currently playing;\n" );
		send_reply ( playerno, person_sd ( player[i] ) );
	    }
	}
	if ( ! count )
	    send_reply ( playerno, "Nobody is playing mundi!\n" );
/*	send_reply ( playerno, form ( "%s\n", LISTPLAYERS ) ); */
	return (-1);
    }

#ifdef CAPITALISE
  if (islower (*line))
      *line = toupper (*line);
    for (i = strlen (line); i > 0; i--)
	if (isupper (line[i]))
	    line[i] = tolower (line[i]);
#endif CAPITALISE


    for (i = 0; i < MAXUSERS; i++) {
	if ((i != playerno) && ISNAMED(i) &&
	    (samenstring (line, player[i].name))) {
	    (void) sprintf (line, "%s\n", PERSONABUSY);
	    send_reply ( playerno, line );
	    return (-1);
	}
	if ((i != playerno) && ISNAMED(i) &&
	    (similarto (line, player[i].name))) {
	    (void) sprintf (line, "%s\n", SIMILARNAME);
	    send_reply ( playerno, line );
	    return (-1);
	}
    }
				/* Check for SHORTNAME */
    if (   (strlen ( line ) > 2 && (line[0] == ' ' || line[1] == ' ') )
	|| (strlen ( line ) < 2 ) ) {
	(void) sprintf ( line, "%s\n", SHORTNAME );
	send_reply ( playerno, line );
	return ( -1 );
    }

				/* Check for NAMEISVERB */
    for ( i = 0; comms[i].command != NULL; i++ ) {
	if ( samenstring ( line, comms[i].command ) ) {
	    (void) sprintf ( line, "%s\n", NAMEISVERB );
	    send_reply ( playerno, line );
	    return ( -1 ) ;
	}
    }
				/* Check for NAMEISNOUN */
    for ( i = 0; i < no_objects; i++ ) {
      /* 2001: was comparing `line' with `obj[i].name'! */
      int j;

      for (j = 0; j < obj[i].name.wordc; j++) {
	if ( samenstring ( line, obj[i].name.wordv[j] ) ) {
	    (void) sprintf ( line, "%s\n", NAMEISNOUN );
	    send_reply ( playerno, line );
	    return ( -1 );
	}
      }
    }
    for ( i = 0; i < no_locations; i++ ) {
	if ( samenstring ( line, loc[i].name ) ||
	     samenstring ( line, loc[i].code ) ) {
	    (void) sprintf ( line, "%s\n", NAMEISNOUN );
	    send_reply ( playerno, line );
	    return ( -1 );
	}
    }
	 
    (void) strcpy (player[playerno].name, line);
				/* Ignore first persona - its junk! */
    for (i = 1; i < no_personas; i++)
	if ((samenstring (line, persona_list[i].name)) &&
	    (persona_list[i].record_no != -1)) {
	     (void) sprintf (line, "%s\n", ACCOUNTEXISTS);
	     temp_fd = player[playerno].fd;
	     bcopy ((char*) &persona_list[i], (char*) &player[playerno],
		    sizeof (struct persona));
	     player[playerno].fd = temp_fd;
	     found_persona = 1;
	     player[playerno].clientmode = AWAITINGSIGN;
	     break;
	 }
	 
    if (found_persona == 0) {
	(void) sprintf (line, "%s\n", NOACCOUNT);
	player[playerno].clientmode = AWAITINGINFO;
				/* Definitately nowiz mode */
	player[playerno].wizstuff = 0;
    }

    if ( !CANARCHP(playerno) && !mundi_open () ) {
	int d, h = next_open_time ( &d );
	(void) sprintf ( line, "%s%s %d:00 hours\n", NOTOPEN,
			days[d], h );
	send_reply ( playerno, line );
	return(-1);
    }
    
#ifdef CHECK_DAEMON_LOADOK    
    if ( !CANARCHP(playerno) && !loadok () ) {
	(void) sprintf ( line, "%s\n", TOOBUSY );
	send_reply ( playerno, line );
	return(-1);
    }
#endif	    

    send_reply ( playerno, line );
    return (1);
}

/***************************************************************************\
|*                                                                         *|
|*  The function generate() does the parts of the client initialisation	   *|
|*  that are specific to personas that are being created.		   *|
|*                                                                         *|
\***************************************************************************/

void generate (playerno)
  int playerno;
{
  player[playerno].oldscore = 0;
  player[playerno].oldmoves = 0;
  player[playerno].displaymode = PRVERBOSE | PREXITS | PROBJECTS;
  player[playerno].games = 0;
  player[playerno].level = 0;
  player[playerno].wizstuff = 0;
  player[playerno].ignorestuff = 0;
  player[playerno].maimstuff = 0;
  player[playerno].record_no = -1;
  player[playerno].maxhp = (random() % 10) + 20;
  player[playerno].hp = player[playerno].maxhp;
  calculate_stats( playerno, 0 );
  player[playerno].intcont      = player[playerno].intcap;
  player[playerno].purity       = 0;

				/* This is a special value meaning that */
				/* when we go to lseek() in the save */
				/* routine, we need a new record_no */
#ifdef MIRKCHEAT
  if ((!strcmp (player[playerno].code, "mirk")) ||
      (!strcmp (player[playerno].code, "maujf")))
    player[playerno].wizstuff = CANWIZ;
  else
    player[playerno].wizstuff = 0;
#endif MIRKCHEAT
}


/****************************************************************************\
  This function will re-compute a players stats... usefull for
  old personas!
  This can't actually happen anymore, but I'll leave it in
  for posterity.
\****************************************************************************/

void revise_old_persona ( playerno )
    int playerno;
{    
    int i, oldlevel;

    send_reply ( playerno, "\
*** WARNING: This character was created with zero statistics ***\n\
*** I will generate some typical stats for you ...           ***\n" );

    calculate_stats( playerno, player[playerno].level );
}

/*--------------------------------------------------------------------------*/

char *heal_player ( playerno, delta, lowlimit, addyour )
    int    playerno;
    int    delta;
    int    lowlimit;
    int    addyour;
{
    int dhp;
    char *feel;

    dhp = player[playerno].hp;
    
    player[playerno].hp += delta;
    
    if ( player[playerno].hp > player[playerno].maxhp ) 
	player[playerno].hp = player[playerno].maxhp;
    
    if ( lowlimit )
	if ( (player[playerno].hp * 4) < player[playerno].maxhp )
	    player[playerno].hp = player[playerno].maxhp / 4;
    
    dhp = player[playerno].hp - dhp;

    if ( dhp <= -500 )
	feel = "absolutely drained in spirit";
    else if ( dhp > -500 && dhp <= -250 )
	feel = "drained of life and spirit";
    else if ( dhp > -250 && dhp <= -100 )
	feel = "physically drained of vitality";
    else if ( dhp > -100 && dhp <= -50 )
	feel = "physically very tired";
    else if ( dhp > -50 && dhp <= -25 )
	feel = "exhausted";
    else if ( dhp > -25 && dhp <= -10 )
	feel = "tired";
    else if ( dhp > -10 && dhp <= -5 )
	feel = "sleepy";
    else if ( dhp > -5 && dhp < 0 )
	feel = "slightly off colour";
    else if ( dhp == 0 )
	return ( 0 );
    else if ( dhp > 0 && dhp <= 5 ) 
	feel = "mentally refreshed";
    else if ( dhp > 5 && dhp <= 10 )
	feel = "slightly better";
    else if ( dhp > 10 && dhp <= 25 )
	feel = "physically refreshed";
    else if ( dhp > 25 && dhp <= 50 )
	feel = "more socially adjusted (;-})";
    else if ( dhp > 50 && dhp <= 100 )
	feel = "greatly refreshed";
    else if ( dhp > 100 && dhp <= 250 )
	feel = "invogriated";
    else if ( dhp > 250 && dhp <= 500 )
	feel = "revitalised";
    else
	feel = "like a new person";

    if ( addyour )
	return ( form ( "You feel %s.\n", feel ) );
    else
	return ( feel );
}

/*--------------------------------------------------------------------------*/

char *rest_player ( playerno, delta, lowlimit, addyour )
    int playerno;
    int delta;
    int lowlimit;
    int addyour;
{
    int dint;
    char *feel;
   
    dint = player[playerno].intcont;
    
    player[playerno].intcont += delta;
    
    if ( player[playerno].intcont > player[playerno].intcap ) 
	player[playerno].intcont = player[playerno].intcap;
    
    if ( lowlimit )
	if ( (player[playerno].intcont * 4) < player[playerno].intcap )
	    player[playerno].intcont = player[playerno].intcap / 4;
    
    dint = (player[playerno].intcont - dint) * 100 / player[playerno].intcap;
    
    if ( dint <= - 60 )
	feel = "less like a beached whale";
    else if ( dint > -60 && dint <= -35 )
	feel = "less like a camels intestines";
    else if ( dint > -35 && dint <= -20 )
	feel = "less like a huge churn of vomit";
    else if ( dint > -20 && dint <= -15 )
	feel = "less bloated";
    else if ( dint > -15 && dint <= -10 )
	feel = "less full";
    else if ( dint > -10 && dint <= 10 )
	return ( 0 );
    else if ( dint > 10 && dint <= 15 ) 
	feel = "slightly less empty";
    else if ( dint > 15 && dint <= 20 )
	feel = "a bit fuller";
    else if ( dint > 20 && dint <= 35 )
	feel = "a lot filled";
    else if ( dint > 35 && dint <= 60 )
	feel = "greatly filled";
    else if ( dint > 60 )
	feel = "completely full";

    if ( addyour )
	return ( form ( "Your stomach feels %s.\n", feel ) );
    else
	return ( feel );
}

/***************************************************************************\
|*                                                                         *|
|*  The function restore() does the parts of the client initialisation	   *|
|*  that are specific to personas that have previously been saved.	   *|
|*                                                                         *|
\***************************************************************************/

void restore (playerno)
  int playerno;
{
  time_t timeofday;
  time_t deltatime;
  char *repl;

  player[playerno].oldscore += player[playerno].newscore;
  player[playerno].level     = compute_level ( sc_levels,
					      player[playerno].oldscore,
					      SC_LEVELS );
  if ( (player[playerno].level < MIN_WIZLEVEL) &&
       WIZMODEP ( playerno) ) {
      send_reply ( playerno,
		  "\n*** Your wizard status has been revoked ***\n" );
      player[playerno].wizstuff &= ~CANWIZ;
  }
      
  player[playerno].oldmoves += player[playerno].newmoves;
  player[playerno].games++;
				/* Check if old persona! */
  if ( player[playerno].intcap == 0 ||
       player[playerno].luck   == 0 )
      revise_old_persona ( playerno );

  player[playerno].snoop = -1;	/* No snooping going on */

  send_reply (playerno, form ("\nThis is game number %d for this persona.\n",
			      player[playerno].games+1));

  (void) time (&timeofday);
  send_reply (playerno, form ("Last game was on %s",
			      ctime (&(player[playerno].logtime)) ) );
  deltatime = timeofday - player[playerno].logtime;
  send_reply (playerno, form ("That was %s ago.\n",
		    hmstime ( deltatime ) ) );

				/* Possile healing - elapsed realtime */
  if ( ( repl = heal_player ( playerno, deltatime / 900, 1, 1 ) ) != 0 )
      send_reply (playerno, repl );

				/* Food will be reset */
  if ( ( repl = rest_player ( playerno, deltatime / 900, 1, 1 ) ) != 0 )
      send_reply (playerno, repl );
}

/***************************************************************************\
|*                                                                         *|
|*  The function setup() does the parts of the client initialisation	   *|
|*  that are relevant both to new and to old personas.			   *|
|*                                                                         *|
\***************************************************************************/

void setup (playerno)
  int playerno;
{
  player[playerno].newscore = 0;
  player[playerno].newmoves = 0;
  player[playerno].snoop = -1;
/*  player[playerno].where = STARTLOC; */
  player[playerno].where = 0;	/* Start at limbo now! ;-) */
  player[playerno].clientmode = PLAYINGGAME;
  player[playerno].invloc = MAXLOCATIONS+playerno;
  player[playerno].wizstuff &= (~WIZMODE) & (~ARCHMODE) & (~METAMODE);
  (void) time (&(player[playerno].logtime));
}

/***************************************************************************\
|*									   *|
|*  The function signon() takes a signon string sent from a client, and    *|
|*  parses it into password/code/host strings, along with stuff like sex,  *|
|*  eye colour and so on if this is a first-time login.  It returns -1 if  *|
|*  for any reason, the login fails (eg., client sent QUIT_MUNDI signal,   *|
|*  or passwords didn't match), otherwise initialises the persona struct   *|
|*  by calling setup() and either generate() or restore() depending on     *|
|*  whether this is a new persona or an old one.			   *|
|*									   *|
\***************************************************************************/

int signon (playerno, line)
  int playerno;
  char *line;
{
  char word[LINELEN];

  if (!strncmp (line, QUIT_MUNDI, 2)) /* If the player wants to quit */
    return (-1);		/* Then login should immediately fail */

  getcomponent (word, line);
  word[PWLEN] = '\0';

  if (player[playerno].clientmode == AWAITINGINFO) {
    (void) strcpy (player[playerno].passwd, word);

    getcomponent (word, line);
    word[SEXLEN] = '\0';
    (void) strcpy (player[playerno].sex, word);
    
    getcomponent (word, line);
    word[HAIRLEN] = '\0';
    (void) strcpy (player[playerno].haircolour, word);
    
    getcomponent (word, line);
    word[EYELEN] = '\0';
    (void) strcpy (player[playerno].eyecolour, word);

    getcomponent (word, line);
    word[PHYSLEN] = '\0';
    (void) strcpy (player[playerno].physical, word);

    generate (playerno);	/* Get value for strength, int., etc. */
    send_reply (playerno, form ("%s\n", PCREATED));
  }
  else {
    if (!strcmp (word, player[playerno].passwd))
      send_reply (playerno, form ("%s\n", PWMATCH));
    else {
      send_reply (playerno, form ("%s\n", NOPWMATCH));
      return (-1);
    }

    restore (playerno);		/* Update scores etc. */
  }

  getcomponent (word, line);
  word[CODELEN] = '\0';
  (void) strcpy (player[playerno].code, word);

  getcomponent (word, line);
  word[HOSTLEN] = '\0';
  (void) strcpy (player[playerno].host, word);

  setup (playerno);
  send_reply (playerno, form ("%s%s\n%s%s\n\n", CHANGE_PERSONA,
			      player[playerno].name,
			      CHANGE_LOCATION,
			      loc[player[playerno].where].code));

  shout (form ("*** %s has just signed on from the code %s.\n",
	       player[playerno].name, player[playerno].code),
	 playerno, MESG_GLOBAL);
  logerror ("Signon: client #%d, name \"%s\".\n", playerno,
	    player[playerno].name);
  return (1);
}

/***************************************************************************\
|*  									   *|
|*  This function does all initialisation connected with the persona file  *|
|*  and the variable player[], which is an array of struct personas.  At   *|
|*  the moment, the password file stuff is unwritten, and the sole	   *|
|*  function of the function, (Whoops, nasty bit of linguistic tautology   *|
|*  there :-) is to set up any bits of the player[] variable that need to  *|
|*  be done before play commences.					   *|
|*  									   *|
|*  Ideally this sort of thing should be in a different function from this *|
|*  one, and this function should be in initialise.c, but what the frog, a *|
|*  little bit of distinction-blurring never hurt anyone, and now that	   *|
|*  I have a bit of a mess around the place, I don't have to be so	   *|
|*  particular about not creating any more, so coding will be more fun.	   *|
|*  Anyway, mark my words, there will *never* be a real AI program that	   *|
|*  works without some kind of blurring, so I am setting a research	   *|
|*  precedent here.  (15:01:1989)					   *|
|*  									   *|
|*  In fact, I have just stuck the code below into daemon.c as well, so	   *|
|*  this function is completely redundant, but I'm not going to delete	   *|
|*  it, 'cos I'm just that kind of guy.  You must remind me to move this   *|
|*  function into initialise.c some time.  (A bit later on 15:01:1989)	   *|
|*  									   *|
|*  Haaargh!  Completely false, I am just about to utterly rewrite the	   *|
|*  routine so that it gets an armful of personas from the system/passwd   *|
|*  file (Oh alright then, all of them), and leaves the file open for	   *|
|*  further transactions (saves, mostly, or maybe deletions)  29:04:1989   *|
|*  									   *|
\***************************************************************************/

int getpersonas ()
{
  int status = 0;
  char pathname[LINELEN];
  struct stat buffer;

  (void) sprintf (pathname, "%s/%s", MUNDI_DIR, PASSWDFILE);

  if ((passwd_fd = open (pathname, O_RDWR | O_CREAT, 0644)) == -1) {
    logerror ("Couldn't open password file: \"%s\"\n", pathname);
    die (23);
  }

  if (fstat (passwd_fd, &buffer) == -1) {
    logerror ("Couldn't stat password file\n");
    die (24);
  }

  /* 2001: added +1 here to avoid malloc(0) on empty password file */
  if ((persona_list = (struct persona*) malloc ((unsigned) buffer.st_size+1))
      == NULL) {
    logerror ("Couldn't allocate memory for personas.\n");
    die (27);
  }
  
  if ((status = read (passwd_fd, (char*) persona_list,
		      (int) buffer.st_size)) == -1) {
    logerror ("Couldn't read password file.\n");
    die (25);
  }
  
  if (status != buffer.st_size) {
    logerror ("Password file lied about its size.\n");
    die (26);
  }

  return (buffer.st_size/sizeof (struct persona));
}

/*--------------------------------------------------------------------------*/

void blind_save (p_index)
  int p_index;
{
  if (lseek (passwd_fd, (off_t) p_index*sizeof (struct persona),
	     L_SET) == -1) {
    logerror ("Couldn't lseek in password file.\n");
    die (29);
  }
  
  if (write (passwd_fd, (char*) (persona_list+p_index),
	     sizeof (struct persona)) == -1) {
    logerror ("Couldn't write persona to file.\n");
    die (30);
  }
}

/*-------------------------------------------------------------------------*/

void save_persona (playerno, victimstr)
  int playerno;
  char *victimstr;
{
  int victimno, i = 0;
				/* If we have no record-number yet, then */
				/* find the first deleted persona.  If */
				/* there isn't one, make a new one. */
  if (*victimstr == '\0')
    victimno = playerno;
  else {
    if (!ARCHMODEP (playerno)) {
      say ("Only arch-wizzes can save other personas.\n");
      return;
    }
    if ((victimno = strmatchplayer (victimstr)) == -1) {
      if ((i = strmatchpersona (victimstr)) != -1) {
	persona_list[i].record_no = i;
	goto GET_ON_WITH_IT;
      }
      else {
	say (form ("There is no-one called \"%s\", froghead!\n", victimstr));
	return;
      }
    }
  }
  
  if ((i = player[victimno].record_no) == -1) {
    for (i = 1; i < no_personas; i++)
      if (persona_list[i].record_no == -1) {
	player[victimno].record_no = i;
	break;
      }
    
    /* 2001: There's confusion throughout the MUNDI source as to
     * whether the persona array starts at [0] or [1], and the next
     * block of code consequently fails with a memory-overrun (which
     * may be overlooked on some systems) when adding the very first
     * persona to a new password file.  Hence my one-line fixup hack.
     */
    if (no_personas == 0) no_personas++;

    if (i == no_personas) {	/* If this needs a new slot in the file */
      no_personas++;		/* Increment total number stored */
      player[victimno].record_no = i;
      if ((persona_list = (struct persona*) realloc ((char*) persona_list,
	     (unsigned)	no_personas*sizeof (struct persona))) == NULL) {
	logerror ("Couldn't reallocate memory.\n");
	die (28);
      }
    }
  }

  bcopy ((char*) &player[victimno], (char*) &persona_list[i],
	 sizeof (struct persona));

GET_ON_WITH_IT:
  blind_save (i);
  say (form ("Saved %s.\n", persona_list[i].name));
}

/*-------------------------------------------------------------------------*/

void delete_persona (playerno, victim)
  int playerno;			/* Index into player[] - report to here */
  int victim;			/* Index into persona_list[] - delete */
{
  if (persona_list[victim].record_no == -1)
    say (form ("Persona \"%s\" has already been deleted.\n",
	       persona_list[victim].name));
  else {
    persona_list[victim].record_no = -1;
    blind_save (victim);
    say (form ("Deleted %s.\n", persona_list[victim].name));
  }
}

/*-------------------------------------------------------------------------*/

void write_pwfile (playerno)
  int playerno;
{
  if (!ARCHMODEP (playerno))
    say ("Chicken off, you weedy non-arch-wiz-type person!\n");
  else {
    if (lseek (passwd_fd, (off_t) 0, L_SET) == -1) {
      logerror ("Couldn't lseek in password file.\n");
      die (29);
    }
    
    if (write (passwd_fd, (char*) persona_list,
	       no_personas*sizeof (struct persona)) == -1) {
      logerror ("Couldn't write personas to file.\n");
      die (30);
    }
    
    say ("Persona file written out to disk.\n");
  }
}


    



