/***************************************************************************\
|*                                                                         *|
|*  Persona.h:	Part of the server program for MUNDI - handle personas.	   *|
|*		This version started on or about 26th May 1988.		   *|
|*                                                                         *|
\***************************************************************************/

#undef	CAPITALISE		/* Make personas have capitalised names */
#undef	MIRKCHEAT		/* Make all maujf/mirk logins wizzes */
#define PASSWDFILE "lib/system/passwd" /* Where to find persona records */

/*-------------------------------------------------------------------------*/

#define	CANWIZ		1	/* Each of these is a binary mask to */   
#define	CANARCH		2	/* undergo bitwise and'ing with the */
#define	CANMETA		4	/* wizstuff flag, to determine status. */ 
#define	WIZMODE		8	/* The CAN* masks determine the ability */
#define	ARCHMODE	16	/* to enter a mode, and the *MODE masks */  
#define	METAMODE	32	/* show whether a player actually is. */  
#define IMMORTAL	64	/* If killed, can you come back as wiz? */

#define	IGNORESHOUT	1	/* These are bit values to be bitwise ... */
				/* (These three not used) */
#define	IGNORESPEAK	2	/* ... and'ed with the value of the ... */
#define	IGNOREARRIVE	4	/* ... ignorestuff flag, to decide ... */
#define	IGNORETAKE	8	/* ... whether to send certain messages. */
#define IGNOREGLOBAL   16	/* Ignore global information... */
#define IGNOREALL      31	/* Ignore all messages */

#define	ISBLIND		1	/* Again, these are bitmasks, this time ... */
#define	ISDEAF		2	/* ... giving status on the various ... */
#define	ISCRIPPLED	4	/* ... things that people can do to you. */
#define	ISDUMB		8	/* Some of these disablement spells ... */
#define	ISDISMEMBERED	16	/* ... will only be available in wiz- ... */
#define	ISGHOST		32	/* ... or maybe even arch-mode. */
#define ISDISEMBOWELED	64	/* Go stick youir head in a cheese */

#define	PRVERBOSE	1	/* Yet more bitmasks.  This time they ... */
#define	PREXITS		2	/* ... are masks for the displaymode ... */
#define	PROBJECTS	4	/* ... flag, and determine which ... */
#define	PRALLEXITS	8	/* ... elements of the full location ... */
#define	PRNUMBERS	16	/* ... description are printed on screen. */

#define AVAILABLE	0	/* Client-number completely unused */
#define AWAITINGNAME	1	/* Connection made, awaiting name */
#define AWAITINGSIGN	2	/* We have the name, and need passwd etc. */
#define AWAITINGINFO	3	/* We have the name, and need passwd etc. */
#define PLAYINGGAME	4	/* Normal state while playing game */
#define WRITINGMAIL	5	/* Mode in which input is sent to a file */
#define ASLEEP		6	/* Client is logged in, but suspended */

/*-------------------------------------------------------------------------*/

#define CANWIZP(playerno)	(player[playerno].wizstuff & CANWIZ)
#define CANARCHP(playerno)	(player[playerno].wizstuff & CANARCH)
#define CANMETAP(playerno)	(player[playerno].wizstuff & CANMETA)
#define WIZMODEP(playerno)	(player[playerno].wizstuff & \
				 (WIZMODE | ARCHMODE | METAMODE))
#define ARCHMODEP(playerno)	(player[playerno].wizstuff & \
				 (ARCHMODE | METAMODE))
#define METAMODEP(playerno)	(player[playerno].wizstuff & METAMODE)
#define IMMORTALP(playerno)	(player[playerno].wizstuff & IMMORTAL)

#define ISACTIVE(playerno)	(player[playerno].clientmode >= PLAYINGGAME)
#define ISNAMED(playerno)	(player[playerno].clientmode >= AWAITINGSIGN)

#define IGNORESHOUTP(playerno)  (player[playerno].ignorestuff & IGNORESHOUT)
#define IGNOREGLOBALP(playerno) (player[playerno].ignorestuff & IGNOREGLOBAL)

/*--------------------------------------------------------------------------*/

struct persona {
  char name[NAMELEN+1];		/* Name used by player when in game. */
  char passwd[PWLEN+1];		/* Password used by player to access game. */
  char code[CODELEN+1];		/* Code from which player last signed on. */
  char host[HOSTLEN+1];		/* Host from which player last signed on. */
  char free[CODELEN+1];		/* Code from which no password is required. */
  time_t logtime;		/* Time at which last login occurred */
  time_t totaltime;		/* Total accumulated game-time */
  int fd;			/* File descriptor connected to client */
  int record_no;		/* Sequential number of this persona */

  char physical[PHYSLEN+1];	/* Arbitrary player-chosen physique etc. */
  char haircolour[HAIRLEN+1];	/* Arbitrary colour of character's hair. */
  char eyecolour[EYELEN+1];	/* Arbitrary colour of character's eyes. */
  char sex[SEXLEN+1];		/* M=Male; F=Female; H=Hermaphrodite. */

  long int strength;		/* Affects damage done when hitting. */
  long int intelligence;	/* Affects ability to cast spells. */
  long int agility;		/* Affects chance to hit people in combat. */
  long int hoopiness;		/* General measure of togetherness */
  long int intcap;		/* Capacity of intestines; ability to eat. */
  long int intcont;		/* Contents of intestines; fatness. */
  long int luck;		/* Affects whatever I choose while coding. */
  long int purity;		/* Measure of moral standards in game. */

  long int hp;			/* Damage that can be done before death. */
  long int maxhp;		/* Highest possible value for stamina. */
  int games;			/* Number of games played so far. */
  long int oldscore;		/* Score when current session started. */
  long int newscore;		/* Polong ints scored during this session. */
  long int oldmoves;		/* No. turns when current session started. */
  long int newmoves;		/* Number of turns during this session. */
  int level;			/* Experience/Power level attained. */

  long int wizstuff;		/* Player's status wrt. sundry wiz-modes */
  long int ignorestuff;		/* What messages is the player ignoring? */
  long int maimstuff;		/* In what ways is the player disabled? */
  long int displaymode;		/* Are exits, objects, players etc. shown? */
  int clientmode;		/* Client's mode, eg. game, mail, login */

  int invloc;			/* Location signifying player's inventory */
  int where;			/* Player's current location */
  int snoop;			/* On whom is the player snooping? */

  long int spare1;		/* In case I think of anything else ... */
  long int spare2;		/* ... and I don't want to have to ... */
  long int spare3;		/* ... rebuild the password file. */
};

/*--------------------------------------------------------------------------*/

extern int passwd_fd;
extern struct persona player[MAXUSERS];
extern struct persona *persona_list;

/*-------------------------------------------------------------------------*/

#define SC_LEVELS 25		/* Number of score-levels (ability) */
#define PU_LEVELS 21		/* Number of purity-levels (alignment) */
#define FA_LEVELS 15		/* Number of fatness-levels (intcont) */

#define MAX_FATNESS 100000
#define MAX_PURITY  100000

/*-------------------------------------------------------------------------*/

struct level {
  long int minscore;		/* Score needed to attain level */
  char name[LINELEN];		/* Textual description of level */
};

/*-------------------------------------------------------------------------*/

extern struct level sc_levels[SC_LEVELS];
extern struct level pu_levels[PU_LEVELS];
extern struct level fa_levels[FA_LEVELS];

/*--------------------------------------------------------------------------*/

extern int verify ();
extern int signon ();
extern int getpersonas ();
extern void save_persona ();
extern void delete_persona ();
extern void write_pwfile ();
extern int  compute_level ();
extern char *heal_player ();
extern char *rest_player ();

/*--------------------------------------------------------------------------*/
