/***************************************************************************\
|*                                                                         *|
|*  Process.c:	Module to preprocess input lines before sending to daemon. *|
|*		This version started about 12th June 1988.		   *|
|*                                                                         *|
\***************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>

#include "global.h"
#include "process.h"
#include "mundi.h"
#include "utils.h"
#include "hhelp.h"

/*-------------------------------------------------------------------------*/

extern char *malloc ();
extern char *getenv ();
/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *strtok ();

/***************************************************************************\
|*									   *|
|*  The first batch of global strings here are those which will be used	   *|
|*  in the print_prompt() function to display location names etc.  These   *|
|*  will be updated by mundi.c and thus must be declared extern there.	   *|
|*									   *|
\***************************************************************************/

char p_persona[LINELEN];	/* Persona being employed */
char p_location[LINELEN];	/* Location currently occupied */
char p_usercode[LINELEN];	/* Usercode from which mundi is run */
char p_computer[LINELEN];	/* Host on which client is run */

extern int remote_login;	/* True if remote login (slow/dumb terminal) */

int source_fd = -1;		/* -1 if not sourcing, fd otherwise */
int current = 0;		/* Current history event-number. */
int client_mode = PLAYING;	/* Kludge until daemon comes back. */
char prompt_string[LINELEN] = PROMPT_STRING;
char history[MAXHISTORY][LINELEN];
struct Alias *alias_ptr;
char *IFS = " \t";
char META_ARRAY[107] = "!@><%-^&*=#|";
				/* Why 107?  Well, y'know? */

char *META_CHAR =	META_ARRAY+0;  /* Introduces history substitution */
char *PROMPT_CHAR =	META_ARRAY+1;  /* Changes client's prompt */
char *ALIAS_CHAR =	META_ARRAY+2;  /* Introduces a new alias */
char *NOALIAS_CHAR =	META_ARRAY+3;  /* Deletes an old alias */
char *HISTORY_CHAR =	META_ARRAY+4;  /* Prints history list */
char *BEFORE_CHAR =	META_ARRAY+5;  /* With META_CHAR, finds old event */
char *REPLACE_CHAR =	META_ARRAY+6;  /* Substitute string in last event */
char *IFS_CHAR =	META_ARRAY+7;  /* Change whitespace characters */
char *METACHAR_CHAR =	META_ARRAY+8;  /* Change meaning of metachars */
char *SOURCE_CHAR =	META_ARRAY+9;  /* Read events from named file */
char *COMMENT_CHAR =	META_ARRAY+10; /* Ignore the rest of the line */
char *ECHO_CHAR =	META_ARRAY+11; /* Print the rest of the line */

/***************************************************************************\
|*                                                                         *|
|*  The function print_prompt(), besides being very tricky to type in a    *|
|*  hurry, will print the string prompt_string, expanding a number of      *|
|*  %-escape sequences that I will program later.			   *|
|*                                                                         *|
|*  Oh look, I have done some of them, (well alright, two), but I know     *|
|*  what they are all going to be now, (subject to possible changes).      *|
|*  They are:								   *|
|*                                                                         *|
|*		%t:	Insert current time (hh:mm:ss format).		   *|
|*		%h:	Insert current history-number.			   *|
|*		%m:	Insert current client-mode.			   *|
|*		%M:	Insert current client-mode (upper case).	   *|
|*		%p:	Insert persona-name.				   *|
|*		%l:	Insert current location.			   *|
|*		%u:	Insert usercode being used.			   *|
|*		%c:	Insert computer being used.			   *|
|*		%%:	Insert a percent-sign.				   *|
|*                                                                         *|
|*  Of these, %m is currently kludged, but will ultimately be sent by      *|
|*  the daemon in a message of the form "ab<string>\n", where (a) is       *|
|*  a globally-defined CHANGEPROMPT, (b) is the character representing     *|
|*  the prompt-filed that needs chnaging, and <string> is its new value.   *|
|*  %l will also be done this way, hence the need for character (b).       *|
|*  The strings for %p, %u and %c will be filled in once at the start      *|
|*  of the run, when they are delved out to be sent to the daemon.         *|
|*                                                                         *|
\***************************************************************************/

void print_prompt ()
{
  char *ch;

  for (ch = prompt_string; *ch != 0; ch++)
    if (*ch != '%')
      (void) putchar (*ch);
    else
      switch (*++ch) {

      case 't':
	print_time ();
	break;

      case 'h':
	(void) printf ("%d", current);
	break;

      case 'm':
	if (client_mode == INFO)
	  (void) fputs ("info", stdout);
	else
	  (void) fputs ("play", stdout);
	break;

      case 'M':
	if (client_mode == INFO)
	  (void) fputs ("INFO", stdout);
	else
	  (void) fputs ("PLAY", stdout);
	break;

      case 'p':
	(void) fputs (p_persona, stdout);
	break;

      case 'l':
	(void) fputs (p_location, stdout);
	break;

      case 'u':
	(void) fputs (p_usercode, stdout);
	break;

      case 'c':
	(void) fputs (p_computer, stdout);
	break;

      case '%':
      default:
	(void) putchar ('%');
      }
  
  (void) fflush (stdout);
}

/***************************************************************************\
|*                                                                         *|
|*  The function init_process() initialises the tables and variables       *|
|*  that are used in the processing of input-lines, eg. setting all the    *|
|*  all the history-table entries to null-strings, and initialising the    *|
|*  alias table (pah!)  Oddly enough, I've decided to use a linked-list    *|
|*  for the aliases, after briefly flirting with the idea of a fixed       *|
|*  array, a dynamically-allocated array, an ongoingly-REalloced array     *|
|*  and even a hash-table.  The last idea is kinda pointless, since        *|
|*  when it comes to actually doing the substitutions, we need some kind   *|
|*  of a "foreach i (aliases)" loop, which you can't do in a hash-table    *|
|*  in any remotely sensible way, the other ideas are equally valid        *|
|*  in many ways, but I fancy having no hard limit to the number of        *|
|*  aliases, and I thought, what the frog, I haven't written a linked      *|
|*  list in years ...  After all, every good Computer Scientist should     *|
|*  keep his hand in by writing the occasional linked list ...  Hehehe     *|
|*                                                                         *|
\***************************************************************************/

void init_process ()
{
  int i;
  
  for (i = 0; i < MAXHISTORY; i++)
    *history[i] = '\0';

  alias_ptr = NULL;

  *p_persona = *p_location = *p_computer = *p_usercode = 0;
}

/***************************************************************************\
|*                                                                         *|
|*  The function update() keeps the global history[][] array up to date.   *|
|*  The integer current is the history event-number, and when reduced      *|
|*  positive-modulo MAXHISTORY (with the macro MAP()) is an index into     *|
|*  this array.  Sunny says I should liven this comment up by typing it    *|
|*  in reverse, so here goes:						   *|
|*  									   *|
|*  .etad ot pu yarra ][][yrotsih labolg eht speek )(etadpu noitcnuf ehT   *|
|*     decuder nehw dna ,rebmun-tneve yrotsih eht si tnerruc regetni ehT   *|
|*    otni xedni na si ))(PAM orcam eht htiw( YROTSIHXAM oludom-evitisop   *|
|*   ti gnipyt yb pu tnemmoc siht nevil dluohs I syas ynnuS  .yarra siht   *|
|*  					       :seog ereh os ,esrever ni   *|
|*                                                                         *|
\***************************************************************************/

void update (line)
  char *line;
{
  (void) strcpy (history[MAP (current)], line);
  current++;
}

/***************************************************************************\
|*                                                                         *|
|*  The function list_meta() lists on the screen the current bindings of   *|
|*  meta-commands to meta-keys, since it can get pretty confusing now	   *|
|*  that you can change them						   *|
|*                                                                         *|
\***************************************************************************/

void list_meta ()
{
  (void) printf ("\nMETA_CHAR is     '%c'\n", *META_CHAR);
  (void) printf ("PROMPT_CHAR is   '%c'\n", *PROMPT_CHAR);
  (void) printf ("ALIAS_CHAR is    '%c'\n", *ALIAS_CHAR);
  (void) printf ("NOALIAS_CHAR is  '%c'\n", *NOALIAS_CHAR);
  (void) printf ("HISTORY_CHAR is  '%c'\n", *HISTORY_CHAR);
  (void) printf ("BEFORE_CHAR is   '%c'\n", *BEFORE_CHAR);
  (void) printf ("REPLACE_CHAR is  '%c'\n", *REPLACE_CHAR);
  (void) printf ("IFS_CHAR is      '%c'\n", *IFS_CHAR);
  (void) printf ("METACHAR_CHAR is '%c'\n", *METACHAR_CHAR);
  (void) printf ("SOURCE_CHAR is   '%c'\n", *SOURCE_CHAR);
  (void) printf ("COMMENT_CHAR is  '%c'\n", *COMMENT_CHAR);
  (void) printf ("ECHO_CHAR is     '%c'\n", *ECHO_CHAR);
  (void) puts ("See \"info\" for more details.\n");
}

/***************************************************************************\
|*                                                                         *|
|*  The function source_file() reads commands from a named file, and       *|
|*  processes them as if they had been typed at the keyboard, (either      *|
|*  as metacommands, or as commands to be passed to the daemon.  Unlike    *|
|*  change_meta &c., it is a void function.  This is because we want to    *|
|*  store even unsuccessful source'ing attempts in the history.  Actually, *|
|*  I think I'll just make it always return 1, just in case I change my    *|
|*  mind.								   *|
|*                                                                         *|
\***************************************************************************/

/****************************************************************************\
  There was a bug previously with this function, which meant that
  all lines in the source file was handled in one lump. If the person
  sourcing had died before the sourcing had finished, the rest of the input
  was still sent - this could cause serious frog ups later on...
\****************************************************************************/

int source_file (line, printflag, interactive)
    char *line;
    int printflag;
    int interactive;
{
    static FILE *fp;
    static char file[LINELEN];
    static char temp[LINELEN];
    char *cp;
				/* Handle non-interactive source */
    if ( ! interactive ) {
	while ( 1 ) {
	    if ( ! source_file ( line, 0, 1 ) )
		return ( 0 );
	    if ( source_fd == -1 )
		break;
	}
	return ( 1 );
    }
	
    if ( source_fd == -1 ) {
				/* Open source file */
	for (cp = line+1; HOOPY_ISSPACE (*cp) && *cp != '\0'; cp++);
	strcpy ( file, cp );
	if ((fp = fopen (file, "r")) == NULL) {
	    if (printflag)
		(void) printf ("Couldn't read file \"%s\"\n", file);
	    source_fd = -1;
	    return (0);
	}
	
	if (printflag)
	    (void) printf ("* Handling lines from file \"%s\"\n", file);
				/* Set source_fd */
	source_fd = fileno(fp);
	return(1);
    } else {
				/* Execute one line */

	if ( fgets ( temp, LINELEN, fp ) != NULL ) {
	    temp[strlen (temp)-1] = '\0';
	    handle_line (temp);
	} else {
				/* Close and finish sourcing */
	    if (printflag) {
		(void) printf ("* Finished file \"%s\"\n", file);
		print_prompt (); 
	    }
	    (void) fclose (fp);
	    source_fd = -1;

	}
	return (1);
    }
}

/***************************************************************************\
|*                                                                         *|
|*  The function change_prompt() doesn't really need to be a function at   *|
|*  all, it sits neatly in-line in process(), but for compatibility with   *|
|*  change_meta() and so on, I thought it would be best.  You can guess    *|
|*  what it does for yourself.						   *|
|*                                                                         *|
\***************************************************************************/

int change_prompt (line)
  char *line;
{
  (void) strcpy (prompt_string, line+1);
  return (1);
}

/***************************************************************************\
|*                                                                         *|
|*  The function change_IFS needs no comments at all, s'completely	   *|
|*  transparent, and it makes me wonder why I'm writing this comment.	   *|
|*  S'the trouble with going all structured, I tend to go over the top	   *|
|*  and stick the comments all over the place.  Errm, well now that	   *|
|*  I've started on this one, I'd better get some substance to it.  So	   *|
|*  here's what Sunny says:						   *|
|*									   *|
|*  Nyaaa.  This function .. o, er.  HE, I mean *HE* calls it the	   *|
|*  "INTERNAL field seperator".  Yes, INTERNAL.  Pah!  But WE all know	   *|
|*  what it really means, don't we, A?A?A?  Yup, we surely do.  Nunno	   *|
|*  this INTERNAL carp, naaaa.  It REALLY stands for INTESTINAL.  And	   *|
|*  what about this 'ere  cheesing "FIELD", A?A?A?  Naaa, it's FISH!	   *|
|*  O O Mater, and what about the frogging "SEPERATOR", A?A?A?A?A?A?A?	   *|
|*  Carp!!  I said, CAARRP!  It's actually CHEESECLOTH.			   *|
|*                                                                         *|
\***************************************************************************/

int change_IFS (line)
  char *line;
{
  char *cp;

  if (line[1] == '\0') {
    (void) puts ("Current field-seperator characters:");
    for (cp = IFS; *cp != '\0'; cp++)
      (void) printf ("   '%c' (%d)", *cp, *cp);
    (void) putchar ('\n');
  }
  else
    (void) strcpy (IFS, line+1);
  
  return (1);
}

/***************************************************************************\
|*                                                                         *|
|*  The function change_meta takes as its argument a string which should   *|
|*  be of exactly two characters in length; the first is the character to  *|
|*  change from, the second is that one to change it to.  This could get   *|
|*  messy.  Well, let's face it, it WILL get messy.  Still, never mind, A? *|
|*                                                                         *|
\***************************************************************************/

int change_meta (line)
  char *line;
{
  char *oldp;			/* Pointer into META_ARRAY */

  if (line[1] == '\0') {
    list_meta ();
    return (1);
  }

  if ((oldp = index (META_ARRAY, line[1])) == NULL) {
    (void) printf ("\"%c\" is not in use as a metachar.\n", line[1]);
    return (0);
  }

  if (line[2] == '\0') {
    (void) puts ("You must specify what to change the metachar TO");
    return (0);
  }

  if (line[2] == line[1]) {
    (void) puts ("ABEND: 0xh00p1; Error #000465c at BRKp:34; Abort");
    return (0);			/* Snigger */
  }

  if (index (META_ARRAY, line[2]) != NULL) {
    (void) printf ("\"%c\" is already in use as a metachar.\n", line[2]);
    return (0);
  }

  *oldp = line[2];
  return (1);
}  

/*-------------------------------------------------------------------------*/

int replace_match (line)
  char *line;
{
  int i;

  for (i = MAP (current-1); i != (MAP (current)); i = PREVIOUS (i))
    if (samestring (history[i], line+1))
      break;
  
  if (i != MAP (current)) {
    (void) strcpy (line, history[i]);
    return (1);
  }

  (void) printf ("Couldn't match \"%s\" in history.\n", line+1);
  return (0);
}

/*--------------------------------------------------------------------------*/

int replace_number (line)
  char *line;
{
  int i = atoi (line+1);

  if ((i > current) || (i <= current-MAXHISTORY)) {
    (void) printf ("Couldn't find event #%d in history.\n", i);
    return (0);
  }

  (void) strcpy (line, history[MAP (i)]);
  return (1);
}

/*--------------------------------------------------------------------------*/

int replace_relative (line, position)
  char *line;
  int position;
{
  int i = atoi (line+position);

  if ((i >= MAXHISTORY) || (i >= current)) {
    (void) puts ("Recorded history doesn't regress that far.");
    return (0);
  }

  if (i <= 0) {
    (void) puts ("Don't be so frogging stupid.");
    return (0);
  }

  (void) strcpy (line, history[MAP (current-i)]);
  return (1);
}

/*-------------------------------------------------------------------------*/

int replace_string (line)
  char *line;
{
  char *cp = NULL;
  static char old[LINELEN];
  static char new[LINELEN];
  static char temp[LINELEN];

  if ((cp = index (line+1, *REPLACE_CHAR)) == NULL)
    *new = '\0';
  else {
    *cp = '\0';
    (void) strcpy (new, cp+1);
  }
  (void) strcpy (old, line+1);

  (void) strcpy (temp, history[MAP (current-1)]);
  if ((cp = strindex (temp, old)) != NULL) {
    *cp = '\0';
    (void) strcpy (line, temp);
    (void) strcat (line, new);
    (void) strcat (line, cp+strlen (old));
    return (1);
  }
  else {
    (void) puts ("(No substitution made)");
    (void) strcpy (line, temp);
    return (0);
  }
}

/*--------------------------------------------------------------------------*/

void print_history (line)
  char *line;
{
  int i, length;
  
  length = atoi (line+1);
  if (length < 0)
    (void) puts ("Phlegm off, duodenum-features.");
  else {
    if ((length == 0) || (length > MAXHISTORY-1))
      length = MAXHISTORY-1;
    
    for (i = ((current >= length) ? (current-length+1) : 0);
	 i != current; i++)
      (void) printf ("%4d: %s\n", i, history[MAP (i)]);
    (void) printf ("%4d: %s\n", current, line);
  }
}

/***************************************************************************\
|*                                                                         *|
|*  The function print_aliases() traverses the whole of the linked list    *|
|*  pointed to by alias_ptr, printing the values of each component of      *|
|*  each component.  Hey, I'm really getting into these minor semantic     *|
|*  ambiguities!  Cool, aren't they?  Let's try some more ... um ...       *|
|*  Cheese, I can't think of any now.					   *|
|*                                                                         *|
\***************************************************************************/

void print_aliases ()
{
  struct Alias *ptr;

  for (ptr = alias_ptr; ptr != NULL; ptr = ptr->next_ptr)
    (void) printf ("%-*s%s\n", WORDLEN+1, ptr->name, ptr->expand);

  if (alias_ptr == NULL)
    (void) puts ("No aliases defined.");
}

/***************************************************************************\
|*                                                                         *|
|*  The function print_alias() traverses the linked list pointed to by     *|
|*  alias_ptr, until it finds the component it has been handed as an       *|
|*  argument, which it then prints.					   *|
|*                                                                         *|
\***************************************************************************/

void print_alias (word)
  char *word;
{
  struct Alias *ptr;

  for (ptr = alias_ptr; ptr != NULL; ptr = ptr->next_ptr)
    if (samenstring (word, ptr->name)) {
      (void) printf ("%-*s%s\n", WORDLEN+1, ptr->name, ptr->expand);
      break;
    }

  if (ptr == NULL)
    (void) printf ("%s is not aliased.\n", word);
}

/***************************************************************************\
|*                                                                         *|
|*  The function add_alias() inserts a new element into the linked list    *|
|*  of aliases, probably by sticking it in at the front, but I haven't     *|
|*  decided yet, and might zip along and tack it onto the end, just for    *|
|*  the intestines of it.						   *|
|*                                                                         *|
|*  Haaargh!  I've decided now -- in fact I *have* to go down the list     *|
|*  to see if the same thing has already been defined, in which case       *|
|*  I just replace the old one.						   *|
|*                                                                         *|
\***************************************************************************/

void add_alias (word, line)
  char *word;
  char *line;
{
  struct Alias *ptr;

  word[WORDLEN] = '\0';		/* Just in case we have a cretin here */
  for (ptr = alias_ptr; ptr != NULL; ptr = ptr->next_ptr)
    if (samenstring (ptr->name, word)) {
      (void) strcpy (ptr->expand, line);
      return;
    }

  if ((ptr = (struct Alias*) malloc (sizeof (struct Alias))) == NULL)
    die (11);
  
  (void) strcpy (ptr->name, word);
  (void) strcpy (ptr->expand, line);
  ptr->next_ptr = alias_ptr;
  alias_ptr = ptr;
}

/***************************************************************************\
|*                                                                         *|
|*  The function make_alias() adds a new alias onto the end of the list,   *|
|*  prints the value of an existing alias, or prints the whole list.       *|
|*  Actually, I think it's got a pretty misleading name, but then I've     *|
|*  been kinda hoopily-structured by breaking this down into the three     *|
|*  subfunctions instead of just writing all the bits in-line, which is    *|
|*  my natural tendancy, so I'll forgive myself that minor indiscretion.   *|
|*                                                                         *|
\***************************************************************************/

void make_alias (line)
  char *line;
{
  char newline[LINELEN];
  char word[LINELEN];
  char *cp;
  
  for (cp = line+1; (HOOPY_ISSPACE (*cp) && (*cp != '\0')); cp++);
  if (*cp == '\0')
    print_aliases ();
  else {
    (void) strcpy (newline, cp);
    getnextword (word, newline);
    if (*newline == '\0')
      print_alias (word);
    else
      add_alias (word, newline);
  }
}

/***************************************************************************\
|*                                                                         *|
|*  The function destroy_alias() sucks an alias out of the middle of the   *|
|*  linked list.  Why "sucks"?  Well, it seems appropriate for the way     *|
|*  the list kinda contracts around the deleted alias.  		   *|
|*                                                                         *|
\***************************************************************************/

void destroy_alias (line)
  char *line;
{
  struct Alias *ptr, *last_ptr = NULL;
  char *cp;

  for (cp = line+1; (HOOPY_ISSPACE (*cp) && (*cp != '\0')); cp++);
  if (*cp == '\0')
    (void) puts ("Nothing to unalias!");
  else {
    for (ptr = alias_ptr; ptr != NULL; ptr = ptr->next_ptr) {
      if (samenstring (cp, ptr->name)) {
	if (last_ptr == NULL)
	  alias_ptr = ptr->next_ptr;
	else
	  last_ptr->next_ptr = ptr->next_ptr;
	free ((char*) ptr);
	break;
      }
      last_ptr = ptr;
    }
    
    if (ptr == NULL)
      (void) printf ("%s is not aliased.\n", cp);
  }
}

/***************************************************************************\
|*                                                                         *|
|*  The function echo_line() prints a line.  You might not think it	   *|
|*  needs it's own line, but trust me, it does.  It has to suck up any	   *|
|*  whitespace, for a start.  And it's generally more elegant this way.	   *|
|*                                                                         *|
\***************************************************************************/

void echo_line (line)
  char *line;
{
  char *cp;

  for (cp = line+1; (HOOPY_ISSPACE (*cp) && (*cp != '\0')); cp++);
  (void) puts (cp);
}

/***************************************************************************\
|*                                                                         *|
|*  The function process_aliases () acts on a character-array, returning   *|
|*  no value, but transforming the array by substitution of any aliases    *|
|*  that are embedded in it.						   *|
|*                                                                         *|
|*  Max sez: Ive had a bad day and dont really want to work out why things *|
|*  are going wrong here for the first command - Ill leave the orginal     *|
|*  ifdef'd out.							   *|
\***************************************************************************/

void process_aliases (line)
    char *line;
{
    static char newline[DESCLEN];
    char *word = line;
    struct Alias *ptr;
    int len;

    newline[0] = 0;

    while ( *word != '\0' ) {	/* While we have a word */
				/* Look at next word */
	len = strcspn ( word, IFS );
				/* If alias then expand */
	for (ptr = alias_ptr; ptr != NULL; ptr = ptr->next_ptr) {
	    if (   (  len == strlen ( ptr->name ) )
		&& ( strncmp ( word, ptr->name, len ) == 0 ) ) {
		(void) strcat ( newline, ptr->expand );
		break;
	    }
	}
				/* Else just add word */
	if ( ptr == NULL )
	    (void) strncat ( newline, word, len );
				/* Add space to end */
	word += len;
	if ( ( len = strspn ( word, IFS ) ) == 0 )
	    break;
	strncat ( newline, word, len );
	word += len;
    }
				/* Copy newline back over line */
    (void) strcpy ( line, newline);
}

#ifdef OLD_PROCESS_ALIASES
void process_aliases (line)
  char *line;
{
  static char newline[DESCLEN];
  char *cp = line, *np = newline;
  struct Alias *ptr;

  while (*cp != '\0') {		/* Check each word in the line ... */
				/* ... against each alias */
    for (ptr = alias_ptr; ptr != NULL; ptr = ptr->next_ptr) {
      if (
#ifdef ALIAS_WORDS
	  ((HOOPY_ISSPACE (*(cp+strlen (ptr->name)))) ||
	   (*(cp+strlen (ptr->name)) == '\0')) &&
#endif ALIAS_WORDS
	  (samenstring (cp, ptr->name))) {
	(void) strcpy (np, ptr->expand);
	np += strlen (ptr->expand);
	cp += strlen (ptr->name);
	break;
      }
    }
    
    if (ptr == NULL) {		/* If we didn't subsitute the word, */
      while (!HOOPY_ISSPACE (*cp) && (*cp != '\0')) {
	*np = *cp;		/* Advance through it ... */
	np++;			/* ... copying as we go */
	cp++;
      }
    }
    
    while (HOOPY_ISSPACE (*cp) && (*cp != '\0')) {
      *np = *cp;		/* Advance to next word ... */
      np++;			/* ... copying as we go */
      cp++;
    }
  }

  *np = '\0';
  (void) strcpy (line, newline);
}
#endif

/***************************************************************************\
|*                                                                         *|
|*  For an explanation of the processing that this function does on the    *|
|*  argument line, see the MUNDI info file "metacommands".  This function  *|
|*  is called prior to sending anything to the daemon.			   *|
|*                                                                         *|
\***************************************************************************/

int process (line)
  char *line;
{
  int send_to_daemon = 1;	/* Send unless told otherwise */
  int add_to_history = 1;	/* Add unless told otherwise */

  skip_spaces (line);		/* Throw away leading cruddy bits */
  process_aliases (line);	/* Do this first (deliberately) */

  if ((*line == *META_CHAR) &&	/* Repeat previous command */
      ((line[1] == *META_CHAR) ||
       (line[1] == *META_CHAR)))
    (void) strcpy (line, history[MAP (current-1)]);
  
  if ((*line == *META_CHAR) &&	/* If command character is a digit */
      (isdigit (line[1])))	/* Replace by the numbered event */
    add_to_history = send_to_daemon = replace_number (line);

  if ((*line == *META_CHAR) &&
      (line [1] == *BEFORE_CHAR)) /* Pick out n'th previous event */
    add_to_history = send_to_daemon = replace_relative (line, 2);
  
  if (*line == *BEFORE_CHAR) /* Pick out n'th previous event */
    add_to_history = send_to_daemon = replace_relative (line, 1);
  
  if ((*line == *META_CHAR) &&
      (NOT_SPECIAL (line[1])))	/* Replace by last event that matches */
    add_to_history = send_to_daemon = replace_match (line);
  
  if (*line == *REPLACE_CHAR)	/* Return the line with alterations. */
    add_to_history = send_to_daemon = replace_string (line);
  
  if (((*line == '`') || (*line == '"') || (*line == '\'')) &&
      (line[1] == '\0'))
    (void) strcat (line, history[MAP (current-1)]);
  
  if (*line == *ALIAS_CHAR) {
    make_alias (line);
    send_to_daemon = 0;
  }
  
  if (*line == *NOALIAS_CHAR) {
    destroy_alias (line);
    send_to_daemon = 0;
  }

  if (*line == *HISTORY_CHAR) {
    print_history (line);
    send_to_daemon = 0;
  }
  
  if (*line == *PROMPT_CHAR) {
    add_to_history = change_prompt (line);
    send_to_daemon = 0;
  }

  if (*line == *IFS_CHAR) {
    add_to_history = change_IFS (line);
    send_to_daemon = 0;
  }

  if (*line == *METACHAR_CHAR) {
    add_to_history = change_meta (line);
    send_to_daemon = 0;
  }

  if (*line == *SOURCE_CHAR) {
    if ( source_fd == -1 ) {
      add_to_history = source_file (line, 1, 1);
      send_to_daemon = 0;
    } else
      printf ( "*** Sorry - cannot source recursivly ***\n" );
  }

  if (*line == *ECHO_CHAR) {
    echo_line (line);
    add_to_history = send_to_daemon = 0;
  }

  if (add_to_history)
    update (line);		/* Keep the history list up to date */
  
  return (send_to_daemon);	/* Tell main() whether to send the string */
}

/***************************************************************************\
|*                                                                         *|
|*  This is a simple function that handles whatever is passed to it,       *|
|*  either as a meta-command, or as a command for the daemon.  This is     *|
|*  the function that should be called both by mundi.c, when a line of     *|
|*  input has been received, and by the source_file() function.  It is     *|
|*  the only function (other than print_prompt() and the initialiser)      *|
|*  that needs to be visible to the outside world.			   *|
|*                                                                         *|
\***************************************************************************/

int handle_line (line)
  char *line;
{
  char *info_ptr;
  extern int socketfd;		/* The file-descriptor through which */
  extern void safewrite();	/* transformed lines should be sent to */
				/* the daemon, and the function to do it. */

  if ((*line != '\0') && (*line != *COMMENT_CHAR))
    if (process (line))
      switch (client_mode) {

      case INFO:
	if ((info_ptr = handle_hhelp (line)) != NULL)
	  pagestr (info_ptr);
	else {
	  client_mode = PLAYING;
	  (void) puts ("(Leaving info mode)");
	}
	break;

      default:
	if (samestring (line, "info")) {
	  client_mode = INFO;
	  (void) puts ("(Entering info mode)");
	  (void) fputs (handle_hhelp ("list"), stdout);
	}
	else if (samestring (line, "help")) {
	  (void) puts ("\n\
Welcome to MUNDI -- this is, as has been pointed out on a number of\n\
occasions, not only a very powerful and impressive program, but also a\n\
very confusing one, to the extent that even the author is frequently\n\
surprised by the result of some metacommands.  You have two options:\n\
You can either ignore all the flashy stuff, and just play it like you\n\
would MIST or any other primitive pile of carp(anag), or you can take\n\
the metaphorical bull by the horns, and frogging well learn how to use\n\
MUNDI properly, metacommands, aliases, history and all.\n\n\
If you wanna go for the second option, the on-line documentation is\n\
probably a good place to start, so you'll be wanting to type \"info\",\n\
then \"help\" to get the help for the info mode.  Then follow your\n\
instincts!\n\n\
Well, whatever ...  Have fun! -- Unkle Mirk.\n");
	}
	else
	  safewrite (socketfd, line);
      }
}

/*--------------------------------------------------------------------------*/
