/***************************************************************************\
|*                                                                         *|
|*  utils.c:	This module is going to contain all the neat little        *|
|*  		functions that I need in both the client and the server    *|
|*		programs for MUNDI, along with anything needed in only     *|
|*		one of the programs, but which doesn't really fit.         *|
|*  Started:	Fri May 19 15:47:58 BST 1989				   *|
|*                                                                         *|
\***************************************************************************/

#include <sys/wait.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>		/* 2001: need this too, for strdup() */
#include <sys/types.h>
#include <time.h>
#include <varargs.h>
#include <ctype.h>

#include "global.h"
#include "utils.h"
#include "look.h"

/*-------------------------------------------------------------------------*/

extern time_t time ();

/***************************************************************************\
|*									   *|
|*  The function pagestr() takes as its argument a null-terminated	   *|
|*  string of arbirary length (usually very long), consisting of a	   *|
|*  number of '\n'-seperated lines.  It prints all this to stdout,	   *|
|*  SCREEN_DEPTH lines at a time, and requires a key to be pressed	   *|
|*  between each screenful.  SInce this will (for now) only be used	   *|
|*  as a part of MUNDI, I am going to assume that the terminal is in	   *|
|*  cbreak mode.							   *|
|*									   *|
|*  Commands at page-break are currently:				   *|
|*	q/Q	Quit reading this pagefull.				   *|
|*	Other	Read next pagefull.					   *|
|*									   *|
\***************************************************************************/

void _pagestr (text)
  char *text;
{
  int lines = 0;		/* Lines printed so far on this page */
  char *here = text;		/* Points at bit about to be printed */
  char *cp;			/* Points at the next '\n', if any */
  char ch;			/* Character typed between pages */

  while ((cp = index (here, '\n')) != NULL) {
    *cp = '\0';
    (void) puts (here);
    *cp = '\n';
    here = cp+1;
    if (++lines == SCREEN_DEPTH-1) {
      (void) fputs ("--- MORE ---", stdout);
      ch = getchar ();
      (void) fputs ("\r            \r", stdout);
      lines = 0;
      if ((ch == 'q') || (ch == 'Q'))
	return;
    }
  }

  if (*here != '\0') {
    (void) printf ("SPARE BIT: %s\n", here);
  }
}

/*
 * 2001: renamed pagestr() to _pagestr() and added this wrapper
 * function, because _pagestr() writes NULs into its argument string.
 * That's usually OK, but not when it's a string constant as in (for
 * example) what gets passed in if the user types "help".
 */
void pagestr (text)
    char *text;
{
    char *tmp = strdup(text);
    _pagestr(tmp);
    free(tmp);
}

/***************************************************************************\
|*                                                                         *|
|*  The function getnextword() takes as arguments two strings.  The value  *|
|*  of the first is ignored, the second is dissected into two components,  *|
|*  the first of which is its value up to (but not including) the first    *|
|*  whitespace, and the second of which is the rest of itself after that   *|
|*  whitespace.  The first of these components is placed into the first    *|
|*  of the function's arguments, the second is placed into the second,     *|
|*  kinda ambiguous, huh?  Well, you know what I mean.  Eat cheese.        *|
|*                                                                         *|
|* Ps. Max rewrote it cos it didn't work/couldn't understand it            *|
\***************************************************************************/

void getnextword (word, line)
  char *word;
  char *line;
{
  char *old = line;		/* Remeber line */
				/* Skip leading blanks */
  while ( HOOPY_ISSPACE (*line) && (*line != '\0'))
      line++;
				/* Copy word */
  while (!HOOPY_ISSPACE (*line) && (*line != '\0'))
    *word++ = *line++;
  *word++ = '\0';		/* Null at end of word */
				/* Move over spaces in line*/
  while (HOOPY_ISSPACE (*line) && (*line != '\0'))
    line++;
				/* recopy line to front */
  while ( *line )
      *old++ = *line++;
  *old++ = '\0';
}

/****************************************************************************\
  This function gets the next object, instead of word
  from a string which has the form;


  <object> [" and " <object>]*

  An object can have the form <word> [ " " <word>]*

\****************************************************************************/

void getnextobject (word, line)
  char *word;
  char *line;
{
    char temp[LINELEN];
    int f = 0;

    *word = '\0';

    while ( *line != '\0' ) {
				/* Get next word */
	getnextword ( temp, line );
				/* Is word "and" then end */
	if ( samenstring ( temp, "and" ) )
	    break;
	if ( samenstring ( temp, "," ) )
	    break;
				/* Add word with space after */
	f = 1;
	strcat ( word, temp );
	strcat ( word, " " );
    }


    if ( f == 1 )
	word[strlen(word)-1] = '\0';
}


/***************************************************************************\
|*                                                                         *|
|*  The function skip_spaces() takes as its argument a null-terminated     *|
|*  string, and pseudo-returns it all shunted up a bit such that, if it    *|
|*  started with characters that were in IFS, then it no longer does!      *|
|*                                                                         *|
\***************************************************************************/

void skip_spaces (line)
  char *line;
{
  char *cp = line;
  char *frog;

  if (HOOPY_ISSPACE (*line)) {
    while (HOOPY_ISSPACE (*cp) && *cp != '\0')
      cp++;

    if (*cp == '\0')
      *line = '\0';
    else {
      for (frog = cp; *frog != '\0'; frog++)
	*(frog-cp+line) = *frog;
      *(frog-cp+line) = *frog;
    }
  }
}

/***************************************************************************\
|*                                                                         *|
|*  The function samestring() returns 1 if the strings passed to it	   *|
|*  are the same, excepting differences of case between letters, up to	   *|
|*  the length of the second.  Otherwise it returns 0			   *|
|*                                                                         *|
\***************************************************************************/

int samestring (template, try)
  char *template;
  char *try;
{
  char *cp;
  
  if (try == NULL || *try == '\0')
    return (0);

  for (cp = try; *cp != 0; cp++)
    if (!((*cp == *(cp-try+template)) ||
	  ((islower (*cp)) && (toupper (*cp)) == *(cp-try+template)) ||
	  ((isupper (*cp)) && (tolower (*cp)) == *(cp-try+template))))
      return (0);		/* Naughty, but nice [1] */
  
  return (1);
}

/*
 * [1] Explanation of this comment.  It's naughty to just subtract the
 * corresponding chars modulo ('a' - 'A') ... or, er Mirk want's this back ...
 * hang on a second.. [2]
 */

/***************************************************************************\
|*                                                                         *|
|*  [2] Excuse for this comment.  Sunny started typing it in, and had got  *|
|*  that far before I asked him to mail it back to me, so that I could     *|
|*  use it for all the history-matching and so on in this module.  This    *|
|*  explains (A) the hideous comment-format, and (B) the gratuitous        *|
|*  apostrophe in the word "want's" (sic).  The reason that it's naughty   *|
|*  to subtract the characters in this way is that it allows other things  *|
|*  that have ASCII codes differing by 'a'-'A' to compare equally as a     *|
|*  part of the case-insensitivity.  The reason it's nice is that I 	   *|
|*  couldn't give a frog that it's naughty[3].				   *|
|*                                                                         *|
\***************************************************************************/

/***************************************************************************\
|*									   *|
|*  It turns out that I no longer don't care that it's naughty, since	   *|
|*  self-same piece of naughtiness in hh_samestring was responsible	   *|
|*  for the legendary "files beginning with 'n' bug".  Thus I have	   *|
|*  re-written this bit to use ctype macros and be generally careful.	   *|
|*  Therefore, Sunny's original comment is entirely lacking meaning,	   *|
|*  but I'm chickened if I'm gonna delete it.				   *|
|*									   *|
\***************************************************************************/

/***************************************************************************\
|*                                                                         *|
|*  The function samenstring() returns 1 if the strings passed to it	   *|
|*  are of the same length, and are the same excepting differences of	   *|
|*  case between letters.  Otherwise it returns 0			   *|
|*                                                                         *|
\***************************************************************************/

int samenstring (template, try)
  char *template;
  char *try;
{
  if (strlen (template) != strlen (try))
    return (0);

  return (samestring (template, try));
}

/***************************************************************************\
|*                                                                         *|
|*  The function print_time will output an ASCII representation of the     *|
|*  current time, for use when printing a prompt that has %t in it.        *|
|*                                                                         *|
\***************************************************************************/

void print_time ()
{
  time_t timeofday;
  static char line[LINELEN];
  
  (void) time (&timeofday);
  (void) strcpy (line, ctime (&timeofday));
  line[19] = '\0';
  (void) printf ("%s", &line[11]);
}

/***************************************************************************\
|*                                                                         *|
|*  The function strindex() takes two strings as arguments, and returns a  *|
|*  pointer to the first occurrence in string1 of string2, or NULL if      *|
|*  there is no such ocurrence.  It is analogous to index() in strings.h   *|
|*                                                                         *|
\***************************************************************************/

char *strindex (str1, str2)
  char *str1;
  char *str2;
{
  int i, vomit = strlen (str2);

  for (i = 0; i <= strlen (str1)-vomit; i++)
    if (samestring (&str1[i], str2))
      return (&str1[i]);
  
  return (NULL);
}

/***************************************************************************\
|*									   *|
|*  Here's a liddle chap that I need in every program I write.  It takes   *|
|*  as its argument a string, and returns a pointer to its first path-	   *|
|*  component, (ie. the first character after the last slash)		   *|
|*									   *|
\***************************************************************************/


char *basename (name)
  char *name;
{
  char *slash;

  if (slash = rindex (name, '/'))
    return (slash+1);
  else
    return (name);
}

/***************************************************************************\
|*                                                                         *|
|*  The function form() is functionally identical to the function of the   *|
|*  same name in C++(eurgh!)'s <stream.h> library.  I mean, let's be       *|
|*  completely frank around here, you can stick all your "cout << cheese"  *|
|*  blurfl in your intestines, but they hit the nail on the head a bit     *|
|*  with your friend and mine, form(), di'n't they, A?  So this takes      *|
|*  a printf(3)-like format string, together with a number of items to     *|
|*  be printed, assembles it in a lump (nice technical term, that, lump)   *|
|*  and returns it in a static string, which you can then send down        *|
|*  sockets or whatever else takes your proverbial.  S'cool, bro!	   *|
|*                                                                         *|
\***************************************************************************/

char *form (va_alist)
  va_dcl			/* Resist the tempation to add a ';'! */
{
  va_list pvar;			/* Parameter-traversing variable */
  static char temp[DESCLEN];	/* Variable to hold printf(3)-like bit */
  char *format;			/* Variable to hold sprintf format */

  va_start (pvar);
  format = va_arg (pvar, char*);
  (void) vsprintf (temp, format, pvar);
  va_end (pvar);

  return (temp);		/* Static: Do not overwrite! */
}

/***************************************************************************\
|*                                                                         *|
|*  hmstime() takes a number of seconds in a long integer and returns a    *|
|*  pointer to a static string which contains a textual description of     *|
|*  that length of time in days, hours, minutes and seconds.		   *|
|*                                                                         *|
\***************************************************************************/

char *hmstime (elapsed)
  long int elapsed;
{
  int days, hours, minutes, seconds;
  static char temp[LINELEN];

  days = elapsed/86400;
  elapsed -= days*86400;
  hours = elapsed/3600;
  elapsed -= hours*3600;
  minutes = elapsed/60;
  elapsed -= minutes*60;
  seconds = elapsed;

  *temp = '\0';
  if (days)
    (void) strcat (temp, form ("%d day%s, ", days, PLURAL (days)));
  if (hours)
    (void) strcat (temp, form ("%d hour%s, ", hours, PLURAL (hours)));
  if (minutes)
    (void) strcat (temp, form ("%d minute%s, ", minutes, PLURAL (minutes)));
  if (seconds)
    (void) strcat (temp, form ("%d second%s, ", seconds, PLURAL (seconds)));

  if (*temp)
    temp[strlen (temp)-2] = '\0';
  else
    (void) strcpy (temp, "no time at all");

  return (temp);
}

/*--------------------------------------------------------------------------*/

int similarto ( from, to )
    char *from;
    char *to;
{
    int lf = strlen ( from ),
        lt = strlen ( to );
    
    return ( ! strncmp ( from, to, min(lf,lt) ) );
}

/*-------------------------------------------------------------------------*/

char *thedate ()
{
    time_t thetime;
    (void) time(&thetime);
    return ( ctime ( &thetime ) );
}

/*--------------------------------------------------------------------------*/

int loadok ()
{
#if 1				/* 2001: load is always "sufficiently OK" */
    return 1;
#else
    int pid, wpid;
    union wait status;
    struct tm *tm_time;
    long      the_time;
				/* Ignore between 12pm and 6am */
    time(&the_time);
    tm_time = localtime(&the_time);

    if ( tm_time->tm_hour >= 0 && tm_time->tm_hour <= 6 )
	return ( 1 );

    if ( (pid = fork ()) == 0 ) {
	(void) execl ( GAMESOK, GAMESOK, "-s", (char *) 0 );
	perror ( "gamesok" );
	exit ( 1 );
    } else {
	while ( ( wpid = wait ( &status ) ) != pid )
	    if ( wpid == -1 ) {
		perror ( "wait" );
		return ( 0 );
	    }
	if ( WIFEXITED(status ) )
	    return ( status.w_T.w_Retcode == 0 );
	else
	    return ( 0 );
    }
#endif
}

/*--------------------------------------------------------------------------*/

int get_hour ()
{
    time_t t = time (0);
    struct tm *tim = localtime(&t);

    return ( tim->tm_hour );
}

int get_sec ()
{
    time_t t = time (0);
    struct tm *tim = localtime(&t);

    return ( tim->tm_sec );
}

int get_min ()
{
    time_t t = time (0);
    struct tm *tim = localtime(&t);

    return ( tim->tm_min );
}

/*--------------------------------------------------------------------------*/
				/* Day is Monday = 0, Sunday = 6 */

int get_day ()
{
    time_t t = time (0);
    struct tm *tim = localtime(&t);
    int day = tim->tm_wday - 1;

    return ( (day >= 0) ? day : (day + DAYSINWEEK) );
}

/*--------------------------------------------------------------------------*/

int compute_seconds_till ( days, hours )
    int days;
    int hours;
{
				/* Day : Monday - Sunday (0-7) */
				/* Hour: 00-23 */
    int d = days - get_day (),
        h = hours - get_hour ();

   
    if ( h < 0 ) {
	h += HOURSINDAY;
	d --;
    }
    if ( d < 0 )
	d += DAYSINWEEK;

				/* Return seconds... */
    return ( (d * 24 + h) * 3600 - (get_sec () + get_min () * 60));
}


