/***************************************************************************\
|*                                                                         *|
|*  Verbs.c:	Part of the server program for MUNDI			   *|
|*		This module will handle all the routines for verbs	   *|
|*		that are not magical in nature, and which are builtins,	   *|
|*		ie. are not defined in the ADL				   *|
|*  Started:	Sat Jun 17 16:01:10 BST 1989				   *|
|*                                                                         *|
\***************************************************************************/

#include <sys/types.h>
#include <strings.h>

#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "handle.h"
#include "look.h"
#include "utils.h"
#include "verbs.h"

/*-------------------------------------------------------------------------*/

extern char *sprintf ();
extern long random ();

/*--------------------------------------------------------------------------*/
				/* Returns 1 if a limit is hit! */

int change_purity ( playerno, delta )
    int playerno;
    int delta;
{
    int new = (  player[playerno].purity += delta );

    if ( new < pu_levels[0].minscore ) {
	player[playerno].purity = pu_levels[0].minscore;
	return 1;
    }
    if ( new > MAX_PURITY ) {
	player[playerno].purity = MAX_PURITY;
	return 1;
    }
    return 0;
}
	
/*-------------------------------------------------------------------------*/

void explode ( playerno, over )
    int playerno;
    int over;
{
    int i, dam = over, vomitf = 0;;

    for ( i = 0; i < MAXUSERS; i++ ) {
	if (   player[i].fd != -1 && ISACTIVE(i)  && i != playerno
	    && player[i].where == player[playerno].where ) {
	    if ( rollagainst ( i, -dam, player[i].luck ) && !ARCHMODEP(i) ) {
		if ( ( player[i].hp -= dam ) < 0 )
		    vomitf = 1;
		else {
	    	    safewrite ( i, form ( "\
%s has exploded - showering you with vast quantities of vomit.\n", 
					 player[playerno].name ) );
		}
	    } else
		vomitf = 1;

	    if ( vomitf ) {
		safewrite ( i, form ( "%s%s\n", VOMITTEDON,
				     player[playerno].name ) );
		speak ( form ( "%s is killed by the force of vomit impacting on his body!\n", player[i].name ),
		       i,
		       player[i].where );
		shout ( form ( "*** %s has suffocated!\n",
			      player[i].name ),
		       i, MESG_GLOBAL );
		killoff ( i );
	    }
	}
    }
    shout ( form ( "*** %s has exploded!\n", player[playerno].name ),
	   playerno, MESG_GLOBAL );
    say ( form ( "%s\n", EXPLODED ) );
    killoff ( playerno );
}

/*--------------------------------------------------------------------------*/

void implode ( playerno )
    int playerno;
{    
    say ( form ( "%s\n", IMPLODED ) );
    speak ( form ( "%s intestines have imploded, causing him to slump into a pile of\nputrid offal. Good eh! You didn't even have to move a finger, so your not\ngetting any points for it ;-P\n", player[playerno].name ),
	   playerno,
	   player[playerno].where );
    shout ( form ( "*** %s has imploded!\n",
		  player[playerno].name ), NO_ONE, MESG_GLOBAL );
    killoff ( playerno );
}

/****************************************************************************\
  This function will change a persons intcont by delta, and will return
  true if they die!
\****************************************************************************/

int change_intcont ( playerno, delta )
    int playerno;
    int delta;
{
    int intper;		/* Intestinal content as a percentage */
    char *mesg = 0;

    player[playerno].intcont += delta;

    if ( player[playerno].intcap )
	intper  = player[playerno].intcont * 100 / player[playerno].intcap;
    else
	intper  = 0;

    if ( intper < 0 || player[playerno].intcont < 0 ) {
	implode ( playerno );
	return ( 1 );
    }

    else if ( intper >= 0 && intper < 1 )
	mesg = "\
If you don't eat soon, your stomach will surely implode, due to the vast\n\
atmospheric pressure bearing down upon it.\n";

    else if ( intper >= 1 && intper < 5 )
	mesg = "\
Your stomach is making horrible groaning, and splushing sounds, as your\n\
digestive acids eat away at your stomach lining.\n";

    else if ( intper >= 5 && intper < 10 )
	mesg = "You are so hungry you could eat at the Union Snack Bar and like it.\n";
    else if ( intper >= 10 && intper < 20 )
	mesg = "You feel like having some dinner, preferably a large pizza.\n";
    
    else if ( intper >= 20 && intper < 30 )
	mesg = "You feel a bit peckish.\n";
    
    else if ( intper > 100 && intper < 105 )
	mesg = "You feel a bit full!\n";

    else if ( intper >= 105 && intper < 110 )
	mesg = "You have eaten a bit too much!\n";

    else if ( intper >= 110 && intper < 120 )
	mesg = "Any more food and you would feel a bit ill!\n";

    else if ( intper >= 120 && intper < 130 )
	mesg = "You feel like you've eaten half of the London Philharmonic Orchestra!\n";

    else if ( intper >= 130 && intper < 150 ) 
	mesg = "\
Your stomach feels like the inside of a camel, after it has eaten a horribly\n\
big thing.\n";

    else if ( intper >= 150 && intper < 175 )
	mesg = "\
You feel *very* ill, like having been to Granny's and eaten sixteen main\n\
courses in one sitting.\n";

    else if ( intper >= 175 && intper < 200 )
	mesg = "\
If I didn't know better, I would say your pregnant! There is this HUGE mound\n\
of bubbling flesh parked right under your belly button.... If you eat any\n\
more - even one wafer thin mint - your bound to explode!\n";

    else if ( intper >= 200 ) {
	explode ( playerno, intper );
	return ( 1 );
    }

    if ( mesg )
        say ( mesg );

    return ( 0 );
}

/*--------------------------------------------------------------------------*/

void damage ( playerno, victim, dam, type, weapon )
int playerno;
int victim;
int dam;
char *type;
char *weapon;
{
    char *damstr;
    int   sign = (dam >= 0) ? 1 : -1;

    if ( strcmp ( type, GUN_DEATH ) )	/* Guns are deadly weapons! */
	dam = sign * ((int) (random() % dam));

    if ( dam < 0 ) {
	if ( ( damstr = heal_player ( victim, -dam, 1, 0 ) ) == 0 )
	    damstr = "no effect";
	say ( form ( "Your target feels %s!\n", damstr ) );
	safewrite ( victim, form ( "You feel %s!\n", damstr ) );
	return;
    } else if ( dam == 0 )
	damstr = "causes no damage at all";
    else if ( dam > 0 && dam <= 2 )
	damstr = "has little effect";
    else if ( dam > 2 && dam <= 4 )
	damstr = "produces a spot of blood";
    else if ( dam > 4 && dam <= 6 )
	damstr = "cuts through some clothing, and darkens it with blood";
    else if ( dam > 6 && dam <= 12 )
	damstr = "causes servere loss of blood";
    else if ( dam > 12 && dam <= 24 )
	damstr = "shreds clothing, and produces loadsablood";
    else if ( dam > 24 && dam <= 48 )
	damstr = "\
is froggin deadly, causing massive nauseous amounts of blood\nto pour forth";
    else if ( dam > 48 && dam <= 96 )
	damstr = "rips off skin, tears muscles and gouges intestines";
    else if ( dam > 96 && dam <= 192 )
	damstr = "pulverises bones and tissue, causing huge piles of gore to\n\
vomit forth from the wounds";
    else
	damstr = "is frogging deadly mate, loads of internal organs are splattered\nrandomly over a vast area (about the size of Basingstoke!)";
    
    say ( form ( "Your attack %s.\n", damstr ) );

    safewrite ( victim, form ( "The attack %s.\n", damstr ) );

    if ( dam > 0 && !ARCHMODEP(playerno) ) {
	int diff = ( player[victim].level - player[playerno].level );
	int mult = 0, pl = player[playerno].level;

	if ( change_purity ( playerno, - ( dam / 10 ) ) ) 
	    say ( "Wow... You're WELL evil. STOP that!\n" ); 

	if ( player[victim].level > 0 ) {
	    if ( diff > -5 )
		mult = (diff + 5) * (diff + 5);
	    else
		mult = 0;

	    if ( diff > pl/4 )	
		mult += 2;

	    if ( diff > pl/2 )
		mult += 5;
				/* Extra points for double levels */
	    if ( diff > pl )
		mult += 5	;
				/* Extra points for triple levels */
	    if ( diff > (pl + pl/2) )
		mult += 10;
				/* Extra points for quadruple levels */
	    if ( diff > (pl*2) )
		mult += 20;
				/* Extra points for non-wiz killing wiz */
	    if (   (pl < MIN_WIZLEVEL)
		&& (player[victim].level >= MIN_WIZLEVEL) )
		mult += 2;
	}
	if ( mult ) {
	    scorepoints ( playerno, (dam * mult), 0, 1 );
	    scorepoints ( victim, -(dam * mult), 0, 3 );
	} else
	    say ( "You score ZERO points for killing people of more then 4 levels below you.\n" );
    } 

#if 0
    if ( dam > 0 ) {
	int diff = (player[victim].level * player[victim].level / 10 );
	if ( diff == 0 )
	    diff++;
	scorepoints ( playerno, (dam * diff), 0, 1 );
    }
#endif

    if ( ( player[victim].hp -= dam ) < 0 ) {
	safewrite ( victim, form ( "%s%s\n", type, weapon ) );
	speak ( form ( "%s dies from his wounds!\n", player[victim].name ),
	       victim, player[victim].where );
	shout ( form ( "*** %s has been killed by %s!\n",
		      player[victim].name,
		      player[playerno].name ), NO_ONE, MESG_GLOBAL );
	killoff ( victim );
    }
}

/*-------------------------------------------------------------------------*/

void getitem (playerno, ind)
    int playerno;
    int ind;
{
    char tempout[LINELEN];
    int oldloc;

    if (obj[ind].location == player[playerno].invloc) {
	say ( form ( "You're already carrying the %s!\n",
		    obj[ind].desc ) );
	return;
    }
    if (!visible (ind, playerno)) {
	say ( form ( "There's no %s here.\n", obj[ind].desc ) );
	return;
    }
    if (!moveablep(ind)) {
	say ( form ( "It's way beyond your power to take the %s!\n",
		    obj[ind].desc ) );
	return;
    }
    if (!accessible (ind, playerno)) {
	say ( form ( "You can't get at the %s right now.\n",
		    obj[ind].desc ) );
	return;
    }
				/* Check that we can carry */
				/* This must be done since person */
				/* may be transferring item on his person */
    oldloc = obj[ind].location;
    obj[ind].location = player[playerno].invloc;

    if (!ARCHMODEP(playerno)) {
	if ( totallocweight(player[playerno].invloc) > 
	    weight_limit(playerno) ) {
	    say ( "You can't carry as much as that!\n" );
	    obj[ind].location = oldloc;
	    return;
	}
    }

    if ( oldloc >= 0 ) {
	say ( form ( "You pick up the %s.\n",
		    obj[ind].desc ) );

	speak ( form ( "%s picks up the %s.\n",
		      player[playerno].name, obj[ind].desc ), 
	       playerno, player[playerno].where );
    } else {
/*	say ( form ( "You take the %s out of the %s.\n", 
		    obj[ind].desc, obj[-oldloc].desc ) ); */ /*MIDAS*/
	say ( form ( "You take the %s %s the %s.\n",
		    obj[ind].desc, surfacep( obj[ind].container ) ? "off" : "out of", obj[-oldloc].desc ) );
    }
}

/****************************************************************************\
  Rollagainst checks if a person succeeds on a roll against a particular
  stat (percentile) which is passed as check.

  The person also recieves a bonus (higher for a better chance of success),
  and automatically recieves some form of luck bonus.

  This means; the bigger check + bonus + level the better chance of suceeding

\****************************************************************************/

int rollagainst ( playerno, bonus, check )
int playerno;
int bonus;
int check;
{
    int add = ( player[playerno].luck / 5 );

    if ( WIZMODEP ( playerno ) )
	add += 10;
    if ( ARCHMODEP ( playerno ) )
	add += 20;
    if ( METAMODEP ( playerno ) )
	add += 30;

    add += ( player[playerno].level * 5 );

    if ( ( random () % 100 ) > ( check + bonus + add ) )
	return ( 0 );
    return ( 1 );
}

/*--------------------------------------------------------------------------*/

int nuclear_joker ( playerno )
    int playerno;
{
    int i;

    if ( ARCHMODEP(playerno ) )
	return (1);

    for ( i = 0; i < no_objects; i++ )
	if ( nuclearsavep(i) )
	    if ( reallyheld ( i, playerno ) )
		return(1);
    return(0);
}
	
/*--------------------------------------------------------------------------*/

int saved_from_nuke(playerno, here, dropper)
int playerno;
int here;
int dropper;
{
    int delta = player[playerno].level * 2 - 150;

    if ( WIZMODEP ( playerno ) )
	delta += 20;

    if ( rollagainst ( playerno, delta, player[playerno].intelligence ) ) {
	say ( "The blast of the explosion activates your protection spell saving you from\n");
	say ( "instant death!\n" );
	speak ( form("%s is saved from death by a protection spell.\n",
		     player[playerno].name ), playerno, here);
	return ( 1 );
    } else if ( nuclear_joker(playerno) ) {
	say ( "Amazingly you are saved from the blasts effect!\n" );
	speak ( form("%s is saved from death by some strange means!\n",
		     player[playerno].name ), playerno, here);
	return ( 1 );
    } else {
	safewrite ( playerno, form("%s%s\n", NUCLEAR_DEATH,  player[dropper].name ));
	speak ( form("%s is vapourized instantly!\n", player[playerno].name ), playerno, here );
	shout ( form("*** %s has been vapourized!\n",
		     player[playerno].name ),
	       NO_ONE, MESG_GLOBAL );
	killoff(playerno);
	return ( 0 );
    }
}

/*--------------------------------------------------------------------------*/

void dropitem (playerno, ind, verb, newloc, dirno )
    int   playerno;		/* player number */
    int   ind;			/* Item number */
    char *verb;			/* Command used */
    int   newloc;		/* location number */
    int   dirno;		/* direction number */
{
    char temp[LINELEN];
    char tempout[LINELEN];
    char *dirstr = "";
    int  throwp = ( player[playerno].where != newloc );
    
    if (!reallyheld (ind, playerno)) {
	say ( form ( "You don't have the %s!\n", obj[ind].desc ) );
	return;
    }

    if (!accessheld (ind, playerno)) {
	say ( form ( "You can't get at the %s right now.\n",
		    obj[ind].desc ) );
	return;
    }

    if ( !moveablep(ind) ) {
	say ( form ( "You can't %s the %s, you peasant!\n", verb,
		    obj[ind].desc ) );
	return;
    }
    
    if ( dirno != -1 )
	dirstr = dir[dirno].desc;
    
    say ( form ( "You %s the %s%s%s.\n", verb, obj[ind].desc,
		((dirno == -1) ? "" : " "), dirstr ) );
    /* 2001: Added ".name" in next line -- how did _that_ ever work?! */
    speak ( form ( "%s %ss the %s%s%s.\n", player[playerno].name, verb,
		  obj[ind].desc, ((dirno == -1) ? "" : " "), dirstr ),
	   playerno, player[playerno].where );
    if ( throwp ) {
	speak ( form ( "The %s flies in, and bounces at your feet.\n",
		      obj[ind].desc ),
	       NO_ONE,
	       newloc );
    }
    obj[ind].location = newloc;
    if ( loc_pitp(newloc) || newloc == UPITLOC ) {
	obj[ind].location = UPITLOC;
	if ( nuclearp(ind) ) {
	    shout ( "*** KER-BOOOOOOOOOM!!! ***\n", NO_ONE, MESG_GLOBAL );
	    for (ind = 0; ind < MAXUSERS; ind++)
		if ((player[ind].fd != -1) && (player[ind].where == newloc ))
		    (void) saved_from_nuke(ind, newloc, playerno);
	    for ( ind = 0; ind < no_objects; ind++ )
		if ( obj[ind].location == UPITLOC ) {
		    newloc = obj[ind].location = (random() % no_locations);
		    speak ( form ( "The %s flies in, and bounces at your feet.\n",
				  obj[ind].desc ),
			   NO_ONE,
			   newloc );
		}
	} else {
	    speak ("It sinks into the pit without trace.\n", NO_ONE, newloc );
	    obj[ind].location = UPITLOC;
	    if (obj[ind].value) {
/*		shout ("*** Ping!\n", NO_ONE, MESG_GLOBAL); */
		scorepoints (playerno, obj[ind].value, 1, 1);
	    }
	}
	return;
    }
    if (nuclearp(ind) && newloc != 0) {
	obj[ind].location = UPITLOC;
	(void) strcpy (loc[newloc].desc,
		       loc[WASTELOC].desc);
	shout ("*** BOOOOOOOOOOM!!! ***\n", NO_ONE, MESG_GLOBAL);
	for (ind = 0; ind < MAXUSERS; ind++)
	    if ((player[ind].fd != -1) && (player[ind].where == newloc ))
		(void) saved_from_nuke(ind, newloc, playerno);
	for (ind = 0; ind < MAXUSERS; ind++)
	    if ((player[ind].fd != -1) && (player[ind].where == newloc)) {
		safewrite (ind, "\n");
		describe (ind, player[ind].where);
	    }
    }
}


/*--------------------------------------------------------------------------*/

void move (playerno, direction)
  int playerno;
  int direction;
{
    int newloc, cw;

    switch ( newloc = getnewloc ( playerno, &direction ) ) {
    case DIR_NOEXIT:
	say ( form ( "You can't go %s from here.\n", dir[direction].desc ) );
	return;
    case DIR_ATPIT:
	say ( "You are already at the pit!\n" );
	return;
    case DIR_NOPITDIR:
	say ( "Naah... that would be too easy!\n" );
	return;
    default:
	speak ( form ( "%s has gone %s.\n", player[playerno].name,
		      dir[direction].desc),
	       playerno, player[playerno].where );
	speak ( form ( "%s just arrived.\n",
			player[playerno].name ), playerno, newloc );
	say ( form ( "You go %s ...\n", dir[direction].desc ) );

	cw = (totallocweight(player[playerno].invloc) / 20) + 1;

	moveto (playerno, newloc);

	if ( player[playerno].newmoves % 5 == 0 )
	    change_intcont ( playerno, -cw );
    }
}

/*--------------------------------------------------------------------------*/

void who (playerno)
  int playerno;
{
  int i;
  static char temp[LINELEN];

  say ("Currently playing:\n");
  for (i = 0; i < MAXUSERS; i++)
      if ( ISACTIVE(i) && (player[i].fd != -1) )
	  say (person_sd (player[i]));
}

/*--------------------------------------------------------------------------*/

void kiss (playerno, string, action, actionpresent)
  int playerno;
  char *string;
  char *action;
  char *actionpresent;
{
  char outmessage[LINELEN];
  char victimstr[LINELEN];
  char extra[LINELEN];
  int i, victim;

  extra[0] = 0;

  getnextword (victimstr, string);
  if (*victimstr == '\0')
    say (form ("Who do you want to %s?\n", action));
  else {
    if ((victim = strtoplayer (victimstr)) == -1) {
      (void) sprintf (outmessage,
		      "There is no-one called \"%s\" logged in.\n",
		      victimstr);
      say (outmessage);
    }
    else {
      if (playerno == victim)
	say ("Sorry, that's too narcissistic and I won't allow it!\n");
      else {
	if ((player[playerno].where != player[victim].where) &&
	    (!WIZMODEP (playerno))) {
	  (void) sprintf (outmessage, "%s isn't here!\n",
			  player[victim].name);
	  say (outmessage);
	}
	else {
	  if (*string == 0)
	    (void) strcpy (extra, "\b");
	  else
	    (void) strcat (extra, string);

	  if (player[playerno].where != player[victim].where)
	    (void) strcat (extra, " (even though he's not here!)");
	  (void) sprintf (outmessage, "You %s %s %s!\n", action,
			  player[victim].name, extra);
	  say (outmessage);
	  (void) sprintf (outmessage, "%s %s you %s!\n",
			  player[playerno].name, actionpresent, extra);
	  safewrite (victim, outmessage);
	  (void) sprintf (outmessage, "%s %s %s %s!\n",
			  player[playerno].name, actionpresent,
			  player[victim].name, extra);
	  for (i = 0; i < MAXUSERS; i++)
	    if ((player[i].fd != -1) && (i != playerno) && (i != victim) &&
		(player[i].where == player[playerno].where))
	      safewrite (i, outmessage);
	}
      }
    }
  }
}

/*--------------------------------------------------------------------------*/

void jump (playerno, how, action, description)
  int playerno;
  char *how;
  char *action;
  char *description;
{
  char temp[LINELEN];
  int vomit;

  if (*how == '\0')
    (void) strcpy (how, "\b");
  (void) sprintf (temp, "%s %ss %s.\n", player[playerno].name, action, how);
  speak (temp, playerno, player[playerno].where);
  say (description);
  if ( samenstring ( action, "vomit" ) ) {
      int intc = player[playerno].intcont / 10;
      vomit = random () % (intc ? intc : 1);
      change_intcont ( playerno, -vomit );
  }
  if ( samenstring ( action, "pray" ) ) {
      if ( change_purity ( playerno, ( random () % 20 ) ) ) 
	  say ( "S'cool.\n" );
  }
}

/*--------------------------------------------------------------------------*/

void wizact (playerno, action)
  int playerno;
  char *action;
{
  if (!WIZMODEP (playerno))
    say ("That's W-I-Z-act, for WIZzes.  I'm sorry if it's confusing.\n");
  else
    if (*action == '\0')
      say ("Wizact in what way?\n");
    else {
      speak (form ("%s %s\n", player[playerno].name, action), playerno,
	     player[playerno].where);
      say ("OK\n");
    }
}

/*--------------------------------------------------------------------------*/

void archact (playerno, action)
    int playerno;
    char *action;
{
    if (!ARCHMODEP (playerno))
	say ("That's A-R-C-H-act, for ARCHes. I'm sorry if it's confusing.\n");
    else
	if (*action == '\0')
	    say ("Archact in what way?\n");
	else {
	    speak ( form ( "%s\n", action ), playerno, player[playerno].where );
	    say ("OK\n");
	}
}

/*--------------------------------------------------------------------------*/

void look (playerno)
  int playerno;
{
  describe (playerno, player[playerno].where);
}

/*--------------------------------------------------------------------------*/

void shoutmess (playerno, string)
  int playerno;
  char *string;
{
    say ( form ( "You shout \"%s\".\n", string ) );

    if ( WIZMODEP ( playerno ) )
	shmortal ( form ( "%s shouts \"%s\".\n",
			 player[playerno].name,
			 string ),
		  playerno, MESG_SHOUT );
    else
	shmortal ( form ( "Someone somewhere shouts \"%s\".\n", string ),
		  playerno, MESG_SHOUT );
    shwiz ( form ( "%s (in %s (%s)) shouts \"%s\".\n",
		  player[playerno].name,
		  loc[player[playerno].where].name,
		  loc[player[playerno].where].code,
		  string ),
	   playerno, MESG_SHOUT );
}

/*--------------------------------------------------------------------------*/

void getstring (playerno, string)
  int playerno;
  char *string;
{
  int number = 0;
  int ind = -1;
  int gotone = 0;
  char item[LINELEN];
  char temp[LINELEN];

  while (*string != '\0') {
    number++;
    getnextobject (item, string);
    if (samestring ("all", item)) {
      for (ind = 0; ind < no_objects; ind++)
	if (obj[ind].location == player[playerno].where) {
	  getitem (playerno, ind);
	  gotone = 1;
	}
      if (!gotone)
	say ("There is nothing here to take.\n");
    }
    else
      if (samestring ("treasure", item)) {
	for (ind = 0; ind < no_objects; ind++)
	  if ((obj[ind].location == player[playerno].where) &&
	      (obj[ind].value)) {
	    getitem (playerno, ind);
	    gotone = 1;
	  }
	if (!gotone)
	  say ("There is no treasure here.\n");
      }
      else {
	ind = strtonoun (item, player[playerno].where);
	if (ind == -1) {
	  (void) sprintf (temp, "I don't know what a \"%s\" is.\n", item);
	  say (temp);
	}
	else
	  getitem (playerno, ind);
      }
  }
  if (number == 0)
    say ("Where are the nouns, you item of aquatic crustacea?\n");
}

/*--------------------------------------------------------------------------*/

void dropstring (playerno, string)
  int playerno;
  char *string;
{
  int number = 0;
  int ind = -1;
  int gotone = 0;
  char item[LINELEN];
  char temp[LINELEN];

  while (*string != '\0') {
    number++;
    getnextobject (item, string);
    if (samestring ("all", item)) {
      for (ind = 0; ind < no_objects; ind++)
	if (obj[ind].location == player[playerno].invloc) {
	  dropitem (playerno, ind, "drop", player[playerno].where, -1 );
	  gotone = 1;
	}
      if (!gotone)
	say ("You're not carrying anything to drop.\n");
    }
    else
      if (samestring ("treasure", item)) {
	for (ind = 0; ind < no_objects; ind++)
	  if ((obj[ind].location == player[playerno].invloc) &&
	      (obj[ind].value)) {
	    dropitem (playerno, ind, "drop", player[playerno].where, -1 );
	    gotone = 1;
	  }
	if (!gotone)
	  say ("You're not carrying any treasure.\n");
      }
      else {
	ind = strtonoun (item, player[playerno].invloc);
	if (ind == -1) {
	  (void) sprintf (temp, "I don't know what a \"%s\" is.\n", item);
	  say (temp);
	}
	else
	  dropitem (playerno, ind, "drop", player[playerno].where, -1 );
      }
  }
  if (number == 0)
    say ("Where are the nouns, you great heap of antelope's cataarh?\n");
}

/*--------------------------------------------------------------------------*/

void openstring (playerno, string)
    int playerno;
    char *string;
{
    int number = 0;
    int ind = -1;
    char item[LINELEN];
    
    while (*string != '\0') {
	number++;
	getnextobject (item, string);
	if ((ind = strtonoun (item, player[playerno].where)) == -1) {
	    say ( form ( "I don't know what a \"%s\" is.\n", item ) );
	    continue;
	}
	if (!visible (ind, playerno)) {
	    say ( form ( "You can't see any %s.\n", obj[ind].desc ) );
	    continue;
	}
	if (!accessible (ind, playerno)) {
	    say ( form ( "You can't get at the %s.\n", obj[ind].desc) );
	    continue;
	}
	if (!containerp(ind)) {
	    say ( form ( "The %s isn't a container, you gonzo!\n",
			    obj[ind].desc ) );
	    continue;
	}
	if (openp(ind)) {
	    say ( form ( "The %s is already open!\n",
			obj[ind].desc ) );
	    continue;
	}
				/* From the above we can deduce; */
				/* 1) Its closed */
				/* 2) It is openable */
				/* 3) It is a container */
				/* 4) There is not fourth thing (ha) */
				/* 5) It is unlockable (hopefully) */
				/*    (Cheese - well we'll check it anyway) */
	if ( lockedp(ind) ) {
	    if ( !lockablep(ind) ) {
		say ( form ( "Fish, the %s seems to be impossible to open!\n",
			    obj[ind].desc ) );
		continue;
	    }
	    say ( form ( "The %s is locked, you need a key.\n",
			obj[ind].desc ) );
	    continue;
	}
				/* Phew - that was horrible! */
	say ( form ( "You open the %s!\n", obj[ind].desc ) );
	obj[ind].container |= 2;
    }
    
    if (number == 0)
	say ( "There should a noun list after that verb.\n");
}

/*--------------------------------------------------------------------------*/

void closestring (playerno, string)
    int playerno;
    char *string;
{
    int number = 0;
    int ind = -1;
    char item[LINELEN];

    while (*string != '\0') {
	number++;
	getnextobject (item, string);
	if ((ind = strtonoun (item, player[playerno].where)) == -1) {
	    say ( form ( "I don't know what a \"%s\" is.\n", item ) );
	    continue;
	}
	if (!visible (ind, playerno)) {
	    say ( form ( "You can't see any %s.\n", obj[ind].desc ) );
	    continue;
	}
	if (!accessible (ind, playerno)) {
	    say ( form ( "You can't get at the %s.\n", obj[ind].desc) );
	    continue;
	}
	if (!containerp(ind)) {
	    say ( form ( "The %s ain't a container, ya gonzo!\n",
			obj[ind].desc) );
	    continue;
	}
	if (!openp(ind)) {
	    say ( form ( "The %s is already closed!\n",
			obj[ind].desc ) );
	    continue;
	}
	if (!openablep(ind)) {
	    say ( form ( "You can't close the %s cheese head!\n",
			obj[ind].desc ) );
	    continue;
	}
				/* We can deduce */
				/* 1) It is open */
				/* 2) It is a container */
				/* 3) It is closeable */
				/* Check if its locked */
	if ( lockedp(ind) ) {
	    if ( !lockablep(ind) ) {
		say ( form ( "Cheese the %s seems to be impossibly stuck\n",
			    obj[ind].desc ) );
		continue;
	    }
	    say ( form ( "The %s is locked open - you'll need a key!\n",
			obj[ind].desc ) );
	    continue;
	}
	say ( form ( "You close the %s!\n", obj[ind].desc) );
	obj[ind].container &= ~2;
    }

    if (number == 0)
	say ( "There should a noun list after that verb.\n");
}

/****************************************************************************\
  This was written by Kenton Oatley - who doesn't know *any* C, and just
  copied and modified a version of get and extrapulated the syntax of C
  from it. I will label any changes I made - but you gotta admit
  the lad had guts....

  Syntax: [put|place] <item> in
\****************************************************************************/

void putstring (playerno, string)
  int playerno;
  char *string;
{
    int ind, ind1, ind2, i;
    int contained = 0;

    char item[LINELEN];	
				/* I got rid of some orrible loop... */
    getnextword (item, string);
    if ((ind1 = strtonoun (item, player[playerno].invloc)) == -1 ) {
	say ( form ( "I don't know what a \"%s\" is.\n", item ) );
	return;
    }

    getnextword (item, string);
    if (samenstring ("in", item) || samenstring ("into", item) ||
	samenstring ("on", item) || samenstring ("onto", item))
	getnextword (item, string);

    if ((ind2 = strtonoun (item, player[playerno].invloc)) == -1 ) {
	say ( form ( "I don't know what a \"%s\" is.\n", item ) );
	return;
    }

    if (!reallyheld (ind1, playerno)) {
	say ( form ( "You don't have the %s.\n", obj[ind1].desc) );
	return;
    }
    if (!visible (ind2, playerno)) {
	say ( form ( "You can't see any %s.\n", obj[ind2].desc) );
	return;
    }
    if (!accessheld (ind1, playerno)) {
	say ( form ( "You can't get at the %s.\n", obj[ind1].desc) );
	return;
    }
    if (!accessible (ind2, playerno)) {
	say ( form ( "You can't get at the %s.\n", obj[ind2].desc) );
	return;
    }
    if (!moveablep (ind1)) {
	say ( form ( "You can't seem to move the %s.\n",
		    obj[ind2].desc ) );
	return;
    }
				/* No weight check - you must be holding */
				/* the first item. */

				/* Is ind2 a surface... */
    if ( surfacep(ind2) ) {
	say ( form ( "You put the %s on the %s.\n", obj[ind1].desc,
		    obj[ind2].desc ) );
	obj[ind1].location = -ind2;
	return;
    }
				/* If it is unopenable and closed it is
				   deffinately not a container */
    if ( ! containerp(ind2) ) {
	say ( form ( "The %s isn't a container you gonzo!\n",
		    obj[ind2].desc ) );
	return;
    }
				/* Otherwise it must be a container... */
    if ( ! openp ( ind2 ) ) {
	say ( form ( "The %s isn't open!\n", obj[ind2].desc ) );
	return;
    }
				/* Count objects contained in */
				/* OR objects on objects which are contained 
				   in */

    for (ind = 0; ind < no_objects; ind++) {
				/* Count things in the object */
	if (obj[ind].location == -ind2) {
	    contained += obj[ind].size;
				/* Count things on objects within it? */
	    if (surfacep(ind)) {
				/* yep - search through for them! */
		for ( i = 0; i < no_objects; i++ ) {
		    if ( obj[i].location == -ind )
			contained += obj[i].size;
		}
	    }
	}
    }
				/* Have we room? */
    if ((obj[ind2].size - contained) > obj[ind1].size) {
				/* Yeah! */
	say ( form ( "You put the %s in the %s.\n", obj[ind1].desc,
			obj[ind2].desc) );
	obj[ind1].location = -ind2;
    } else
	say ( form ( "There isn't enough room for the %s in the %s.\n", 
			obj[ind1].desc, obj[ind2].desc) );
}
